/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "canvas.h"

void
on_new_game_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_new_game(g_canvas);
}


void
on_giveup_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_give_up_activate(g_canvas);
}


void
on_stop_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_stop_activate(g_canvas);
}


void
on_restart_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  printf("on_restart_activet()\n");

  daemon_canvas_restart_activate(g_canvas);
}


void
on_first_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_kif_first(g_canvas);
}


void
on_back_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_kif_back(g_canvas);
}


void
on_next_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_kif_next(g_canvas);
}


void
on_last_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_kif_last(g_canvas);
}


void
on_hirate_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_mate_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_fileopen_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* fileselection;

  fileselection = daemon_create_fileselection ();
  gtk_widget_show(fileselection);

}


void
on_exit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_exit( 0 );
}


void
on_about_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* about;
  
  about = create_about();
  gtk_widget_show( about );
}

gboolean
on_window_destroy_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_exit(0);

  return FALSE;
}


void
on_window_destroy                      (GtkObject       *object,
                                        gpointer         user_data)
{
  gtk_exit(0);
}


gboolean
on_drawingarea_motion_notify_event     (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;
  
  daemon_canvas_motion_notify(g_canvas, event);

  return FALSE;
}


gboolean
on_drawingarea_button_press_event      (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;
  
  daemon_canvas_button_press(g_canvas, event);

  return FALSE;
}


gboolean
on_drawingarea_button_release_event    (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;
  
  daemon_canvas_button_release(g_canvas, event);

  return FALSE;
}


void
on_retract_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}



void
on_filesave_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* fileselection;

  fileselection = daemon_create_fileselection_save();
  gtk_widget_show(fileselection);
}


void
on_filesaveas_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* fileselection;

  fileselection = daemon_create_fileselection_save();
  gtk_widget_show(fileselection);
}


void
on_playgame_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_startedit_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;
  D_CANVAS_MODE mode;

  mode = daemon_canvas_get_mode(g_canvas);

  if (mode != D_CANVAS_MODE_EDIT) {
    int ret = daemon_canvas_ok_cancel_dialog(g_canvas, _("Start edit mode?"));
    if (ret  == GNOME_OK) {
      daemon_canvas_change_mode(g_canvas, D_CANVAS_MODE_EDIT);
    }
  }
}




void
on_small_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  if (g_canvas != NULL) {
    daemon_canvas_change_size(g_canvas, D_CANVAS_SIZE_SMALL);
  }
}


void
on_middle_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_change_size(g_canvas, D_CANVAS_SIZE_MIDDLE);
}


void
on_big_activate                        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  if (g_canvas != NULL) {
    daemon_canvas_change_size(g_canvas, D_CANVAS_SIZE_BIG);
  }
}

void
on_bookwindow_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_bookwindow_activate(g_canvas, menuitem, user_data);
}

void
on_icon_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_text_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_none_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


gboolean
on_booklist_button_press_event         (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{

  return FALSE;
}


void
on_toolbarshow_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_toolbarhide_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  if (g_canvas != NULL) {
    daemon_canvas_change_toolbar(g_canvas, D_CANVAS_TOOLBAR_NONE);
  }
}


gboolean
on_drawingarea_expose_event            (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_expose(g_canvas, widget, event, user_data);

  return FALSE;
}


void
on_bookwindow_destroy                  (GtkObject       *object,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_bookwindow_destroy(g_canvas, object, user_data);
}

void
on_menu_icon_and_text_activate         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  if (g_canvas != NULL) {
    daemon_canvas_change_toolbar(g_canvas, D_CANVAS_TOOLBAR_BOTH);
  }
}


void
on_menu_icon_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_change_toolbar(g_canvas, D_CANVAS_TOOLBAR_ICON);
}


void
on_menu_text_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_change_toolbar(g_canvas, D_CANVAS_TOOLBAR_TEXT);
}


void
on_ok_button_pressed                   (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_cancel_button_pressed               (GtkButton       *button,
                                        gpointer         user_data)
{

}


gboolean
on_drawingarea_leave_notify_event      (GtkWidget       *widget,
                                        GdkEventCrossing *event,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_leave_notify(g_canvas, widget, event, user_data);

  return FALSE;
}


void
on_flip_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_flip_activate(g_canvas);
}


void
on_button_mate_start_pressed           (GtkButton       *button,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_mate_start_pressed(g_canvas);
}


void
on_button_mate_cancel_pressed          (GtkButton       *button,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_mate_cancel_pressed(g_canvas);
}


void
on_check_mate_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_mate_activate(g_canvas);
}


void
on_set_hirate_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_set_hirate_activate(g_canvas);
}

void
on_set_mate_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_set_mate_activate(g_canvas);
}


void
on_rl_reversal_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_set_rl_reversal_activate(g_canvas);
}


void
on_order_reversal_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_set_order_reversal_activate(g_canvas);
}


void
on_edit_flip_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_set_edit_flip_activate(g_canvas);
}


void
on_all_koma_to_pbox_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_all_koma_to_pbox_activate(g_canvas);
}

void
on_mate_thik_cancel                    (GtkButton       *button,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_mate_think_cancel(g_canvas);
}


void
on_button_game_ok_pressed              (GtkButton       *button,
                                        gpointer         user_data)
{
}


void
on_button_game_cancel_pressed          (GtkButton       *button,
                                        gpointer         user_data)
{
}

void
on_game_optionmenu_pressed             (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_give_up_activate                    (GtkButton       *button,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_give_up_activate(g_canvas);
}


void
on_matta_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  daemon_canvas_restart_activate(g_canvas);
}


void
on_statusbar_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  extern Canvas *g_canvas;

  if (g_canvas != NULL) {
    daemon_canvas_statusbar_activate(g_canvas);
  }
}

