/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * Canvas class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/canvas.c,v $
 * $Id: canvas.c,v 1.2 2005/12/09 10:06:25 tokita Exp $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>

#include "interface.h"
#include "support.h"
#include "canvas.h"
#include "misc.h"
#include "si/si.h"
#include "si/ui.h"

#include "back.xpm"
#include "koma.xpm"

/** 󥹥ȥ饯 */
Canvas* daemon_canvas_new(GtkWidget *window) {
  Canvas *canvas;

  g_assert(window != NULL); 

  canvas = (Canvas *)malloc(sizeof(Canvas));
  if (canvas == NULL) {
    daemon_abort("No enough memory in daemon_canvas_new().");
  }

  daemon_canvas_init(canvas, window);

  /* ط */
  daemon_canvas_draw_back(canvas);
  /* ﹹ */
  daemon_canvas_draw_time(canvas);
  /* ץ饤 */
  daemon_canvas_draw_sprite_all(canvas);
  /*  */
  daemon_canvas_draw_all(canvas);
  /* ơС */
  daemon_canvas_update_statusbar(canvas);

  return canvas;
}

/**  */
void daemon_canvas_init(Canvas *canvas, GtkWidget *window) {
  GtkWidget *widget;
  g_assert(window != NULL);
  g_assert(canvas != NULL);

  canvas->window      = window;
  daemon_canvas_set_size(canvas, D_CANVAS_SIZE_SMALL);
  canvas->pixmap[0]   = gdk_pixmap_new(window->window,
				       D_CANCAS_SIZE_SMALL_X,
				       D_CANCAS_SIZE_SMALL_Y,
				       -1);
  canvas->pixmap[1]   = gdk_pixmap_new(window->window,
				       D_CANCAS_SIZE_MIDDLE_X,
				       D_CANCAS_SIZE_MIDDLE_Y,
				       -1);
  canvas->pixmap[2]   = gdk_pixmap_new(window->window,
				       D_CANCAS_SIZE_BIG_X,
				       D_CANCAS_SIZE_BIG_Y,
				       -1);
  canvas->gc          = gdk_gc_new(window->window);
  canvas->drawingarea = lookup_widget(GTK_WIDGET(window), "drawingarea");
  canvas->drag        = FALSE;
  canvas->mode        = D_CANVAS_MODE_EDIT;
  canvas->back_src    = gdk_imlib_create_image_from_xpm_data(back_xpm);
  if (canvas->back_src == NULL) {
    daemon_abort("No enough memory in daemon_canvas_init(). back_src is NULL\n");
  }
  canvas->koma_src    = gdk_imlib_create_image_from_xpm_data(koma_xpm);
  if (canvas->back_src == NULL) {
    daemon_abort("No enough memory in daemon_canvas_init(). koma_src is NULL\n");
  }
  canvas->back[0] = NULL;
  canvas->back[1] = NULL;
  canvas->back[2] = NULL;
  daemon_canvas_create_back_pixmap(canvas);
  daemon_dboard_set_hirate(&(canvas->board));
  canvas->bookwindow  = NULL;
  canvas->flg_bookwindow  = FALSE;
  canvas->tmp_pixmap = NULL;

  {
    /* ġСθܤ򥻥å */
    GtkWidget *toolbar;

    toolbar = lookup_widget(canvas->window, "toolbar");
    gtk_toolbar_set_button_relief(GTK_TOOLBAR(toolbar), GTK_RELIEF_NORMAL);
  }

  /* ץ饤Ƚ */
  daemon_canvas_init_sprite(canvas);
  /* ץ饤 */
  daemon_canvas_dispose_sprite(canvas);
  /* ֤ɽ */
  daemon_canvas_draw_time(canvas);

  /* ॢȤ򥻥å */
  /* gtk_timeout_add(1000, (GtkFunction)daemon_canvas_timeout, (gpointer)canvas); */

  canvas->record = NULL;
  canvas->front = D_SENTE;
  daemon_canvas_set_sensitive(canvas);
  canvas->flg_thinking = FALSE;

  /* ƳϻȤʤ */
  widget = lookup_widget(canvas->window, "restart");
  gtk_widget_hide(widget);

  canvas->statusbar = TRUE;
}

/** ǥȥ饯 */
void daemon_canvas_free(Canvas* canvas) {
  g_assert(canvas != NULL);

  free(canvas);
}

void daemon_canvas_init_sprite(Canvas *canvas) {
  GdkImlibImage *im;
  GdkImlibImage *im2;
  Sprite *sprite;
  gint num, i;

  for (i=0; i<SPRITEMAX; i++) {
    canvas->sprite[i] = NULL;
  }

  num = 0;
#ifdef NOREAD
  {
    /* ǡ󷯤Υץ饤 */
    GdkPixmap *pixmap;
    
    pixmap = load_pixmap_from_data(canvas->window, daemon_xpm);
    sprite = daemon_sprite_new(pixmap, NULL, 128, 128);
    daemon_sprite_set_type(sprite, ETC_SPRITE);
    canvas->sprite[num++] = sprite;
  }
  {
    /* Υץ饤 */
    GdkPixmap *pixmap;
    
    pixmap = load_pixmap_from_data(canvas->window, logo2_xpm);
    sprite = daemon_sprite_new(pixmap, NULL, 128, 128);
    daemon_sprite_set_type(sprite, ETC_SPRITE);
    canvas->sprite[num++] = sprite;
  }
#endif /* NOREAD */

  im = canvas->koma_src;
  gdk_imlib_render(im, im->rgb_width, im->rgb_height);
  canvas->koma[0] = gdk_imlib_move_image(im);
  canvas->mask[0] = gdk_imlib_move_mask(im);

  im2 = gdk_imlib_clone_scaled_image(im, 400, 200);
  if (im2 == NULL) {
    daemon_abort("No enough memory in daemon_canvas_init_sprite(). im2 is NULL\n");
  }
  gdk_imlib_render(im2, im2->rgb_width, im2->rgb_height);
  canvas->koma[1] = gdk_imlib_move_image(im2);
  canvas->mask[1] = gdk_imlib_move_mask(im2);
  gdk_imlib_destroy_image(im2);

  im2 = gdk_imlib_clone_scaled_image(im, 512, 256);
  if (im2 == NULL) {
    daemon_abort("No enough memory in daemon_canvas_init_sprite(). im2 is NULL\n");
  }
  gdk_imlib_render(im2, im2->rgb_width, im2->rgb_height);
  canvas->koma[2] = gdk_imlib_move_image(im2);
  canvas->mask[2] = gdk_imlib_move_mask(im2);
  gdk_imlib_destroy_image(im2);

  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_sprite_new(canvas->koma[0], canvas->mask[0], 40, 40);
    canvas->sprite[num] = sprite;
    daemon_sprite_set_type(sprite, KOMA_SPRITE);
    daemon_sprite_set_src_x(sprite, 0);
    daemon_sprite_set_src_y(sprite, 0);
    daemon_sprite_set_width(sprite, 40);
    daemon_sprite_set_height(sprite, 40);
    daemon_sprite_set_visible(sprite, FALSE);
    daemon_sprite_set_move(canvas->sprite[num], TRUE);
    num++;
  }
}

/** ץ饤Ȥ֤ */
void daemon_canvas_dispose_sprite(Canvas *canvas) {
  DBoard* board;
  Sprite* sprite;
  gint x, y, num;
  gint width, height;
  gint base_x, base_y;
  gint p;
  gint i, j;
  
  board = &(canvas->board);
  num = 0;

  base_x = 0;
  base_y = 0;

  if (canvas->size == D_CANVAS_SIZE_SMALL) {
    base_x = 140;
    base_y = 10;
  } else if (canvas->size == D_CANVAS_SIZE_MIDDLE) {
    base_x = 140 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X;
    base_y = 10 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X;
  } else if (canvas->size == D_CANVAS_SIZE_BIG) {
    base_x = 140 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X;
    base_y = 10 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X;
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }
  
  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_canvas_get_sprite(canvas, i);
    if (sprite == NULL) {
      continue;
    }
    if (daemon_sprite_get_type(sprite) != KOMA_SPRITE) {
      continue;
    }
    daemon_sprite_set_visible(sprite, FALSE);
  }

#ifdef NOREAD
  /* ǡ󷯤Υץ饤 */
  sprite = daemon_canvas_get_sprite(canvas, num++);
  daemon_sprite_set_visible(sprite, TRUE);
  daemon_sprite_set_move(sprite, TRUE);
  /* Υץ饤 */
  sprite = daemon_canvas_get_sprite(canvas, num++);
  daemon_sprite_set_visible(sprite, TRUE);
  daemon_sprite_set_move(sprite, TRUE);
  daemon_sprite_set_x(sprite, 640 - 129);
  daemon_sprite_set_y(sprite, 1);
#endif /* NOREAD */

  /* פ˶ */
  for (y=1; y <= 9; y++) {
    for (x=1; x <= 9; x++) {
      p = daemon_dboard_get_board(board, x, y);
      if (p == 0) {
	continue;
      }
      g_assert(1 <=p && p <= 0x1F);
      sprite = daemon_canvas_get_sprite(canvas, num++);
      g_assert(sprite != NULL);
      while (sprite->type != KOMA_SPRITE) {
	sprite = daemon_canvas_get_sprite(canvas, num++);
	g_assert(sprite != NULL);
      }
      width  = daemon_sprite_get_width(sprite);
      height = daemon_sprite_get_height(sprite);
      daemon_sprite_set_x(sprite, (10 - x) * width - width + base_x);
      daemon_sprite_set_y(sprite, y * height - height + base_y);
      daemon_sprite_set_koma(sprite, p);
      daemon_sprite_set_visible(sprite, TRUE);
    }
  }

  /* ֤ */
  {
    static gint base_y_value[9] = {
      0, 65, 65, 45, 45, 25, 5, 5, 5,
    };
    static gint base_y_value2[9] = {
      0, 65, 65, 85, 85, 105, 125, 125, 125,
    };
    gint count, p, flg;
    gint x, y, skip_x;

    /*  */
    count = count_piece(&(canvas->board), D_SENTE);
    base_x = 511;
    base_y = 210;
    flg = 0;
    skip_x = 20;

    for (i=7; 1 <= i; i--) {
      p = daemon_dboard_get_piece(board, D_SENTE, i);
      if (p == 0) {
	continue;
      }
      if (i == 1) {
	skip_x = 80;
	if (flg == 1) {
	  base_y += 40;
	  flg = 0;
	}
      } else {
	skip_x = 20;
      }
      for (j=0; j<p; j++) {
	x = 5 + (skip_x - skip_x / p * (j + 1)) + (flg ? 60 : 0);
	y = base_y_value[count];

	sprite = daemon_canvas_get_sprite(canvas, num++);
	g_assert(sprite != NULL);
	while (sprite->type != KOMA_SPRITE) {
	  sprite = daemon_canvas_get_sprite(canvas, num++);
	  g_assert(sprite != NULL);
	}

	daemon_sprite_set_x(sprite, daemon_canvas_rate_xy(canvas, base_x + x));
	daemon_sprite_set_y(sprite, daemon_canvas_rate_xy(canvas, base_y + y));
	daemon_sprite_set_koma(sprite, i);
	daemon_sprite_set_visible(sprite, TRUE);
      }

      if (flg == 1) {
	base_y += 40;
	flg = 0;
      } else {
	flg = 1;
      }
    }

    /*  */
    count = count_piece(&(canvas->board), D_GOTE);
    base_x = 1;
    base_y = 1;
    flg = 1;
    skip_x = 20;

    for (i=7; 1 <= i; i--) {
      p = daemon_dboard_get_piece(board, D_GOTE, i);
      if (p == 0) {
	continue;
      }
      if (i == 1) {
	skip_x = 80;
	if (flg == 0) {
	  base_y -= 40;
	}
	flg = 0;
      } else {
	skip_x = 20;
      }
      for (j=0; j<p; j++) {
	x = 5 + (skip_x - skip_x / p * (j + 1)) + (flg ? 60 : 0);
	y = base_y_value2[count];

	sprite = daemon_canvas_get_sprite(canvas, num++);
	g_assert(sprite != NULL);
	while (sprite->type != KOMA_SPRITE) {
	  sprite = daemon_canvas_get_sprite(canvas, num++);
	  g_assert(sprite != NULL);
	}

	daemon_sprite_set_x(sprite, daemon_canvas_rate_xy(canvas, base_x + x));
	daemon_sprite_set_y(sprite, daemon_canvas_rate_xy(canvas, base_y + y));
	daemon_sprite_set_koma(sprite, i + 0x10);
	daemon_sprite_set_visible(sprite, TRUE);
      }

      if (i == 1 || flg == 1) {
	flg = 0;
      } else {
	base_y -= 40;
	flg = 1;
      }
    }

    /* Ȣ */
    if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_EDIT) {
    count = count_pbox(&(canvas->board));
    base_x = 1;
    base_y = 210;
    flg = 0;
    skip_x = 20;

    for (i=8; 1 <= i; i--) {
      p = daemon_dboard_get_pbox(board, i);
      if (p == 0) {
	continue;
      }
      if (0 && i == 1) {
	skip_x = 80;
	flg = 0;
      } else {
	skip_x = 20;
      }
      for (j=0; j<p; j++) {
	x = 5 + (skip_x - skip_x / p * (j + 1)) + (flg ? 60 : 0);
	y = base_y_value[count];

	sprite = daemon_canvas_get_sprite(canvas, num++);
	g_assert(sprite != NULL);
	while (sprite->type != KOMA_SPRITE) {
	  sprite = daemon_canvas_get_sprite(canvas, num++);
	  g_assert(sprite != NULL);
	}

	daemon_sprite_set_x(sprite, daemon_canvas_rate_xy(canvas, base_x + x));
	daemon_sprite_set_y(sprite, daemon_canvas_rate_xy(canvas, base_y + y));
	daemon_sprite_set_koma(sprite, i);
	daemon_sprite_set_visible(sprite, TRUE);
      }

      if (flg == 1) {
	base_y += 40;
	flg = 0;
      } else {
	flg = 1;
      }
    }
    }
  }
}

/*************************************************************/
/* set_* , get_* */
/*************************************************************/

GtkWidget* daemon_canvas_get_window(Canvas *canvas) {
  return canvas->window;
}

GdkGC* daemon_canvas_get_gc(Canvas *canvas) {
  return canvas->gc;
}

GdkPixmap *daemon_canvas_get_pixmap(Canvas *canvas) {
  g_assert(canvas != NULL);
  if (canvas->size == D_CANVAS_SIZE_SMALL) {
    return canvas->pixmap[0];
  } else if (canvas->size == D_CANVAS_SIZE_MIDDLE) {
    return canvas->pixmap[1];
  } else if (canvas->size == D_CANVAS_SIZE_BIG) {
    return canvas->pixmap[2];
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }

  return NULL;
}

GdkPixmap *daemon_canvas_get_back(Canvas *canvas) {
  g_assert(canvas != NULL);
  if (canvas->size == D_CANVAS_SIZE_SMALL) {
    return canvas->back[0];
  } else if (canvas->size == D_CANVAS_SIZE_MIDDLE) {
    return canvas->back[1];
  } else if (canvas->size == D_CANVAS_SIZE_BIG) {
    return canvas->back[2];
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }

  return NULL;
}

GtkWidget* daemon_canvas_get_drawinarea(Canvas *canvas) {
  g_assert(canvas != NULL);
  return canvas->drawingarea;
}

gboolean daemon_canvas_isdrag(Canvas *canvas) {
  g_assert(canvas != NULL);
  return canvas->drag;
}

void daemon_canvas_set_drag(Canvas *canvas, gboolean drag) {
  g_assert(canvas != NULL);
  g_assert(drag == TRUE || drag == FALSE);
  canvas->drag = drag;
}

gint daemon_canvas_get_dragno(Canvas *canvas) {
  g_assert(canvas != NULL);
  return canvas->drag_no;
}

void daemon_canvas_set_dragno(Canvas *canvas, gint no) {
  g_assert(canvas != NULL);
  g_assert(0 <= no);
  canvas->drag_no = no;
}

GdkPoint daemon_canvas_get_diff(Canvas *canvas) {
  g_assert(canvas != NULL);
  return canvas->diff;
}

void daemon_canvas_set_diff(Canvas *canvas, gint x, gint y) {
  g_assert(canvas != NULL);
  g_assert(0 <= x);
  g_assert(0 <= y);
  canvas->diff.x = x;
  canvas->diff.y = y;
}

D_CANVAS_MODE daemon_canvas_get_mode(Canvas *canvas) {
  return canvas->mode;
}

void daemon_canvas_set_mode(Canvas *canvas, D_CANVAS_MODE mode) {
  canvas->mode = mode;
}

/** canvas Υ򥻥åȤ */
void daemon_canvas_set_size(Canvas *canvas, D_CANVAS_SIZE size) {
  g_assert(canvas != NULL);
  g_assert(size == D_CANVAS_SIZE_SMALL ||
	   size == D_CANVAS_SIZE_MIDDLE ||
	   size == D_CANVAS_SIZE_BIG);

  canvas->size = size;
  if (size == D_CANVAS_SIZE_SMALL) {
    canvas->width  = D_CANCAS_SIZE_SMALL_X;
    canvas->height = D_CANCAS_SIZE_SMALL_Y;
  } else if (size == D_CANVAS_SIZE_MIDDLE) {
    canvas->width  = D_CANCAS_SIZE_MIDDLE_X;
    canvas->height = D_CANCAS_SIZE_MIDDLE_Y;
  } else if (size == D_CANVAS_SIZE_BIG) {
    canvas->width  = D_CANCAS_SIZE_BIG_X;
    canvas->height = D_CANCAS_SIZE_BIG_Y;
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }
}

Sprite* daemon_canvas_get_sprite(Canvas* canvas, gint no) {
  g_assert(canvas != NULL);
  g_assert(0 <= no);
  g_assert(no < SPRITEMAX);

  return canvas->sprite[no];
}

void daemon_canvas_set_front(Canvas *canvas, D_TEBAN next) {
  canvas->front = next;
}

D_TEBAN daemon_canvas_get_front(Canvas *canvas) {
  return canvas->front;
}


/*************************************************************/
/* event */
/*************************************************************/

void daemon_canvas_expose(Canvas *canvas,
			  GtkWidget *widget,
			  GdkEventExpose *event,
			  gpointer user_data) {
  int width, height;
  int x, y;
  GdkGC* gc;
  GtkWidget* drawingarea;
  GdkPixmap* pixmap;
  GdkRectangle rect;

  rect        = event->area;
  width       = rect.width;
  height      = rect.height;
  x           = rect.x;
  y           = rect.y;
  gc          = canvas->gc;
  drawingarea = canvas->drawingarea;
  pixmap      = daemon_canvas_get_pixmap(canvas);
  
  gdk_draw_pixmap(drawingarea -> window, 
		  drawingarea -> style -> fg_gc[GTK_WIDGET_STATE(drawingarea)],
		  pixmap,
		  x, y,
		  x, y,
		  width, height);
}

void daemon_canvas_motion_notify(Canvas *canvas, GdkEventMotion  *event) {
  GdkModifierType mask;
  gint x, y;
  gint no;
  Sprite* sprite;
  GdkPoint diff;
  GdkRectangle rect;

  g_assert(canvas != NULL);
  g_assert(event != NULL);

  if (event->is_hint) {
    gdk_window_get_pointer (event->window, &x, &y, &mask);
  } else {
    x = event->x;
    y = event->y;
  }
  
  if (daemon_canvas_isdrag(canvas) == FALSE) {
    /* ɥåǤʤ */
    return;
  }

  if (canvas->tmp_pixmap == NULL) {
    daemon_canvas_create_tmp_pixmap(canvas);
  }

  /* ֥륯åȽѻ */
  canvas->tv.tv_sec  = 0;
  canvas->tv.tv_usec = 0;

  no     = daemon_canvas_get_dragno(canvas);
  sprite = daemon_canvas_get_sprite(canvas, no);
  diff   = daemon_canvas_get_diff(canvas);

  rect.x      = daemon_sprite_get_x(sprite);
  rect.y      = daemon_sprite_get_y(sprite);
  rect.width  = daemon_sprite_get_width(sprite);
  rect.height = daemon_sprite_get_height(sprite);

  daemon_sprite_set_x(sprite, event->x - diff.x);
  daemon_sprite_set_y(sprite, event->y - diff.y);

  /* طʤ */
  daemon_canvas_draw_back_update_from_tmp_pixmap(canvas, &rect);
  daemon_canvas_draw_sprite(canvas, no);
  daemon_canvas_draw_all_update(canvas, &rect);
}

void daemon_canvas_button_press(Canvas *canvas, GdkEventButton  *event) {
  gint     no;
  gboolean doubleclick = FALSE;
  Sprite* sprite;
  gint sprite_x, sprite_y;
  gint dest_x, dest_y;
  D_CANVAS_AREA area;
  SPRITE_TYPE type;
  D_CANVAS_MODE mode;
  gint p;

  g_assert(canvas != NULL);
  g_assert(event != NULL);

  if (event->type != GDK_BUTTON_PRESS) {
    return;
  }

  if (event->button != 1) {
    /* 1 ֤Υܥʳ줿 */
    return;
  }

  if (canvas->mode == D_CANVAS_MODE_BOOK) {
    /* ⡼ɤǤϥåνϤʤ */
    return;
  }

  if (daemon_canvas_isdrag(canvas) == TRUE) {
    /* ɥå˥å줿 */
    daemon_canvas_set_drag(canvas, FALSE);
    return;
  }

  no = daemon_canvas_on_sprite(canvas, event);

  if (no == -1) {
    /* å줿ץ饤Ȥʤ */
    return;
  }

  sprite = daemon_canvas_get_sprite(canvas, no);
  doubleclick = daemon_canvas_is_doubleclick(canvas);

  sprite_x = daemon_sprite_get_x(sprite);
  sprite_y = daemon_sprite_get_y(sprite);
  area = daemon_canvas_on_drag_area(canvas, sprite_x, sprite_y, &dest_x, &dest_y);
  type = daemon_sprite_get_type(sprite);
  mode = daemon_canvas_get_mode(canvas);

  if (mode == D_CANVAS_MODE_GAME &&
      type == KOMA_SPRITE) {
    p = daemon_sprite_get_koma(sprite);

    if (daemon_dboard_get_front(&(canvas->board)) == D_SENTE) {
      if (canvas->board.next == D_SENTE && 0x10 < p) {
	return;
      } else if (canvas->board.next == D_GOTE && p < 0x10) {
	return;
      }
    } else {
      if (canvas->board.next == D_SENTE && p < 0x10) {
	return;
      } else if (canvas->board.next == D_GOTE && 0x10 < p) {
	return;
      }
    }

#ifdef NOREAD
    if (daemon_dboard_get_front(&(canvas->board)) == D_SENTE && 0x10 < p) {
      return;
    } else if (daemon_dboard_get_front(&(canvas->board)) == D_GOTE && 0x10 < p) {
      return;
    }
#endif /* NOREAD */
  }

  if (mode == D_CANVAS_MODE_EDIT &&
      doubleclick && area == D_CANVAS_AREA_BOARD && type == KOMA_SPRITE) {
    /* ֥륯å줿 */
    gint koma;
    GdkRectangle rect;

    koma = daemon_sprite_get_koma(sprite);
    if ((koma & 0x0F) == 0x05 || (koma & 0x0F) == 0x08) {
      if (koma <= 0x10) {
	koma += 0x10;
      } else {
	koma -= 0x10;
      }
    } else if (koma <= 0x10) {
      if (koma & 0x08) {
	koma += 0x10;
      }
      koma ^= 0x08;
    } else {
      if (koma & 0x08) {
	koma -= 0x10;
      }
      koma ^= 0x08;
    }
    daemon_sprite_set_koma(sprite, koma);
    daemon_dboard_set_board(&(canvas->board), dest_x, dest_y, koma);
    
    rect.x      = daemon_sprite_get_x(sprite);
    rect.y      = daemon_sprite_get_y(sprite);
    rect.width  = daemon_sprite_get_width(sprite);
    rect.height = daemon_sprite_get_height(sprite);
    /* طʤ */
    daemon_canvas_draw_back_update(canvas, &rect);
    /*  */
    daemon_canvas_draw_sprite(canvas, no);
    daemon_canvas_draw_sprite_update_all(canvas, &rect);
    daemon_canvas_draw_sprite_all(canvas);
    
    daemon_canvas_draw_all(canvas);
    
    return;
  }

  {
    /* ץ饤Ⱦǥå줿 */
    gint x, y;

    if (daemon_sprite_ismove(sprite) == TRUE) {
      x = daemon_sprite_get_x(sprite);
      y = daemon_sprite_get_y(sprite);

      daemon_sprite_set_from_x(sprite, x);
      daemon_sprite_set_from_y(sprite, y);

      daemon_canvas_set_dragno(canvas, no);
      daemon_canvas_set_diff(canvas, event->x - x, event->y - y);
      daemon_canvas_set_drag(canvas, TRUE);
    }
  }
}

/**
 * (x, y) κɸ̤ΤɤΥꥢˤ뤫Ĵ٤롣<br>
 * ֤ͤ<UL>
 * <LI> 0 : ꥢʤ
 * <LI> 1 : 
 * <LI> 2 : 
 * <LI> 3 : Ȣ
 * <LI> 4 : פξ
 * </UL>
 * ֤ͤ 4 ξϡ׾κɸdest_x dest_y ˥åȤ
 *  (1, 1)  (9, 9) ͡
 */
D_CANVAS_AREA daemon_canvas_on_drag_area(Canvas *canvas, gint x, gint y,
			       gint *dest_x, gint *dest_y) {
  GdkPoint point;
  GdkRectangle left_koma, right_koma, komabako, board;

  point.x = daemon_canvas_rerate_xy(canvas, x) + D_SPRIRE_KOMA_WIDTH / 2;
  point.y = daemon_canvas_rerate_xy(canvas, y) + D_SPRIRE_KOMA_HEIGHT / 2;

  left_koma.x = 1;
  left_koma.y = 1;
  left_koma.width  = 129;
  left_koma.height = 169;
  if (daemon_on_rectangle(&left_koma, &point) == TRUE) {
    return D_CANVAS_AREA_LEFT_KOMADAI;
  }

  right_koma.x = 511;
  right_koma.y = 210;
  right_koma.width  = 129;
  right_koma.height = 169;
  if (daemon_on_rectangle(&right_koma, &point) == TRUE) {
    return D_CANVAS_AREA_RIGHT_KOMADAI;
  }

  komabako.x = 1;
  komabako.y = 210;
  komabako.width  = 129;
  komabako.height = 169;
  if (daemon_on_rectangle(&komabako, &point) == TRUE) {
    return D_CANVAS_AREA_KOMABAKO;
  }

  board.x = 140;
  board.y = 11;
  board.width  = 360;
  board.height = 360;
  if (daemon_on_rectangle(&board, &point) == TRUE) {
    /* 10 - ϶κɸαȿž뤿 */
    *dest_x = 10 - ((point.x - board.x) / D_SPRIRE_KOMA_WIDTH + 1);
    *dest_y = (point.y - board.y) / D_SPRIRE_KOMA_HEIGHT + 1;
    return D_CANVAS_AREA_BOARD;
  }

  return D_CANVAS_AREA_NONE;
}

void daemon_canvas_button_release(Canvas *canvas, GdkEventButton  *event) {
  g_assert(canvas != NULL);
  g_assert(event != NULL);

  if (event->type != GDK_BUTTON_RELEASE || event->button != 1) {
    return;
  }

  if (daemon_canvas_isdrag(canvas) == FALSE) {
    return;
  }

  if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_EDIT) {
    gint to_x, to_y;
    gint from_x, from_y;
    gint sprite_x, sprite_y;
    gint sprite_from_x, sprite_from_y;
    gint no;
    gint to_ret, from_ret;
    Sprite *sprite;
    gint p, p2;
    SPRITE_TYPE type;

    no            = daemon_canvas_get_dragno(canvas);
    sprite        = daemon_canvas_get_sprite(canvas, no);
    sprite_x      = daemon_sprite_get_x(sprite);
    sprite_y      = daemon_sprite_get_y(sprite);
    sprite_from_x = daemon_sprite_get_from_x(sprite);
    sprite_from_y = daemon_sprite_get_from_y(sprite);
    type          = daemon_sprite_get_type(sprite);

    to_x = to_y = 0;
    to_ret = daemon_canvas_on_drag_area(canvas, sprite_x, sprite_y,
					&to_x, &to_y);

    from_ret = daemon_canvas_on_drag_area(canvas, sprite_from_x, sprite_from_y,
					  &from_x, &from_y);

    p = daemon_sprite_get_koma(sprite);

    if (from_ret == D_CANVAS_AREA_NONE ||
	to_ret == D_CANVAS_AREA_NONE ||
	((p & 0xF) == 0x8 &&
	 (to_ret == D_CANVAS_AREA_LEFT_KOMADAI ||
	  to_ret == D_CANVAS_AREA_RIGHT_KOMADAI))) {
      /* 桢פγ˥ɥå褦Ȥ */

      from_x = daemon_sprite_get_from_x(sprite);
      from_y = daemon_sprite_get_from_y(sprite);
      daemon_sprite_set_x(sprite, from_x);
      daemon_sprite_set_y(sprite, from_y);

      /* ץ饤Ⱥ */
      daemon_canvas_dispose_sprite(canvas);

      /* ̺ */
      daemon_canvas_draw_back(canvas);
      daemon_canvas_draw_sprite_all(canvas);
      daemon_canvas_draw_time(canvas);
      daemon_canvas_draw_all(canvas);
    } else if (to_x == from_x && to_y == from_y && from_ret == D_CANVAS_AREA_BOARD) {
      daemon_dboard_set_board(&(canvas->board), from_x, from_y, p);

      /* ץ饤Ⱥ */
      daemon_canvas_dispose_sprite(canvas);

      /* ̺ */
      daemon_canvas_draw_back(canvas);
      daemon_canvas_draw_sprite_all(canvas);
      daemon_canvas_draw_time(canvas);
      daemon_canvas_draw_all(canvas);
    } else if (type == KOMA_SPRITE) {
      if (from_ret == D_CANVAS_AREA_BOARD) {
	daemon_dboard_set_board(&(canvas->board), from_x, from_y, 0);
      } else if (from_ret == D_CANVAS_AREA_LEFT_KOMADAI) {
	daemon_dboard_dec_piece(&(canvas->board),
				daemon_canvas_get_front(canvas) ^ 1,
				p & 7);
      } else if (from_ret == D_CANVAS_AREA_RIGHT_KOMADAI) {
	daemon_dboard_dec_piece(&(canvas->board),
				daemon_canvas_get_front(canvas),
				p & 7);
      } else {
	p &= 0xF;
	p = (p != 8) ? p & 7 : p;
	daemon_dboard_dec_pbox(&(canvas->board), p);
      }
      if (to_ret == D_CANVAS_AREA_BOARD) {
	p2 = daemon_dboard_get_board(&(canvas->board), to_x, to_y);
	if ((p2 & 0xF) == 8) {
	  p2 &= 0xF;
	  p2 = (p2 != 8) ? p2 & 7 : p2;
	  daemon_dboard_add_pbox(&(canvas->board), p2);
	} else if (p2 != 0) {
	  if (daemon_canvas_get_front(canvas) == D_SENTE) {
	    daemon_dboard_add_piece(&(canvas->board),
				    (p < 0x10) ? D_SENTE : D_GOTE,
				    p2 & 7);
	  } else {
	    daemon_dboard_add_piece(&(canvas->board),
				    (p < 0x10) ? D_GOTE : D_SENTE,
				    p2 & 7);
	  }
	}
	daemon_dboard_set_board(&(canvas->board), to_x, to_y, p);
      } else if (to_ret == D_CANVAS_AREA_LEFT_KOMADAI) {
	daemon_dboard_add_piece(&(canvas->board),
				daemon_canvas_get_front(canvas) ^ 1,
				p & 7);
      } else if (to_ret == D_CANVAS_AREA_RIGHT_KOMADAI) {
	daemon_dboard_add_piece(&(canvas->board),
				daemon_canvas_get_front(canvas),
				p & 7);
      } else {
	p &= 0xF;
	p = (p != 8) ? p & 7 : p;
	daemon_dboard_add_pbox(&(canvas->board), p);
      }

      /* ץ饤Ⱥ */
      daemon_canvas_dispose_sprite(canvas);

      /* ̺ */
      daemon_canvas_draw_back(canvas);
      daemon_canvas_draw_sprite_all(canvas);
      daemon_canvas_draw_time(canvas);
      daemon_canvas_draw_all(canvas);
    }
  } else if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME) {
    gint to_x, to_y;
    gint from_x, from_y;
    gint sprite_x, sprite_y;
    gint sprite_from_x, sprite_from_y;
    gint no;
    gint to_ret, from_ret;
    gint p;
    Sprite *sprite;
    SPRITE_TYPE type;

    no            = daemon_canvas_get_dragno(canvas);
    sprite        = daemon_canvas_get_sprite(canvas, no);
    sprite_x      = daemon_sprite_get_x(sprite);
    sprite_y      = daemon_sprite_get_y(sprite);
    sprite_from_x = daemon_sprite_get_from_x(sprite);
    sprite_from_y = daemon_sprite_get_from_y(sprite);
    type          = daemon_sprite_get_type(sprite);

    to_x = to_y = 0;
    to_ret = daemon_canvas_on_drag_area(canvas, sprite_x, sprite_y,
					&to_x, &to_y);

    from_ret = daemon_canvas_on_drag_area(canvas, sprite_from_x, sprite_from_y,
					  &from_x, &from_y);

    p = daemon_sprite_get_koma(sprite);

    if (from_ret == D_CANVAS_AREA_NONE ||
	to_ret == D_CANVAS_AREA_NONE ||
	to_ret != D_CANVAS_AREA_BOARD) {
      /* 桢פγ˥ɥå褦Ȥ */

      from_x = daemon_sprite_get_from_x(sprite);
      from_y = daemon_sprite_get_from_y(sprite);
      daemon_sprite_set_x(sprite, from_x);
      daemon_sprite_set_y(sprite, from_y);

      /* ץ饤Ⱥ */
      daemon_canvas_dispose_sprite(canvas);

      /* ̺ */
      daemon_canvas_draw_back(canvas);
      daemon_canvas_draw_sprite_all(canvas);
      daemon_canvas_draw_time(canvas);
      daemon_canvas_draw_all(canvas);
    } else if (type == KOMA_SPRITE) {
      extern BOARD g_board;
      TE te, te2;
      gboolean just_te;
      gboolean just_nari;

      if (from_ret == D_CANVAS_AREA_LEFT_KOMADAI) {
	te.fm = 0;
	te.to = (to_y << 4) + to_x;
	te.nari = 0;
	te.uti = p & 7;
      } else if (from_ret == D_CANVAS_AREA_RIGHT_KOMADAI) {
	te.fm = 0;
	te.to = (to_y << 4) + to_x;
	te.nari = 0;
	te.uti = p & 7;
      } else {
	te.fm = (from_y << 4) + from_x;
	te.to = (to_y << 4) + to_x;
	te.nari = 0;
	te.uti = 0;
      }

      if (daemon_dboard_get_front(&(canvas->board)) == D_GOTE) {
	gint x, y;
	if (te.fm) {
	  x = te.fm & 0xF;
	  y = te.fm >> 4;
	  te.fm = ((10 - y) << 4) + (10 - x);
	}
	x = te.to & 0xF;
	y = te.to >> 4;
	te.to = ((10 - y) << 4) + (10 - x);
      }

      te2 = te;
      if (te.uti == 0) {
	te2.nari = 1;
      }
      
      just_te = FALSE;
      just_nari = FALSE;

      if (boCheckTE(&g_board, &te)) {
	just_te = TRUE;
      } else if (boCheckTE(&g_board, &te2)) {
	just_te = TRUE;
	te.nari = 1;
      }

      if (te.nari == 0 && (p & 0xF) != OH && (p & 0xF) != KIN && (p & 8) == 0) {
	just_nari = TRUE;
      }

      if (just_te && just_nari && boCanNari(&g_board, &te)) {
	if (daemon_canvas_yes_no_dialog(canvas, _("Evolution ?")) == GNOME_YES) {
	  te.nari = 1;
	}
      }
      
      if (just_te) {
	canvas->gui_te = te;
#ifdef NOREAD
	/* ץ饤Ⱥ */
	daemon_canvas_dispose_sprite(canvas);

	/* ̺ */
	daemon_canvas_draw_back(canvas);
	daemon_canvas_draw_sprite_all(canvas);
	daemon_canvas_draw_time(canvas);
#endif /* NOREAD */
	daemon_canvas_draw_all(canvas);
      } else {
	from_x = daemon_sprite_get_from_x(sprite);
	from_y = daemon_sprite_get_from_y(sprite);
	daemon_sprite_set_x(sprite, from_x);
	daemon_sprite_set_y(sprite, from_y);

	/* ץ饤Ⱥ */
	daemon_canvas_dispose_sprite(canvas);

	/* ̺ */
	daemon_canvas_draw_back(canvas);
	daemon_canvas_draw_sprite_all(canvas);
	daemon_canvas_draw_time(canvas);
	daemon_canvas_draw_all(canvas);
      }
    }
  }

  daemon_canvas_set_drag(canvas, FALSE);
  gdk_imlib_free_pixmap(canvas->tmp_pixmap);
  canvas->tmp_pixmap = NULL;
}

/** 襦ɥƤϿ */
GtkWidget *daemon_canvas_create_bookwindow(Canvas *canvas) {
  GtkWidget *bookwindow;
  GtkWidget *booklist;
  Record *record;
  DTe *te;
  DBoard *board;
  DMoveInfo *mi;
  GtkWidget *item;
  char buf[BUFSIZ];
  int i;

  bookwindow = create_bookwindow();

  booklist = lookup_widget(bookwindow, "booklist");
  g_assert(booklist != NULL);

  if (canvas->record != NULL) {
    record = canvas->record;
    
    board = daemon_dboard_new();
    daemon_dboard_copy(&(record->first_board), board);
    mi = &(record->mi);

    for (i=0; i<=mi->count; i++) {
      if (0 < i) {
	te = daemon_dmoveinfo_get(mi, i - 1);
      } else {
	te = NULL;
      }
      create_te_string(board, i, te, buf);
      
      item = gtk_list_item_new_with_label(buf);
      gtk_container_add(GTK_CONTAINER(booklist), item);
      gtk_widget_show(item);

      gtk_signal_connect (GTK_OBJECT (item), "button_press_event",
			  GTK_SIGNAL_FUNC (daemon_canvas_bookwindow_click),
			  (gpointer)i);

      if (te != NULL) {
	daemon_dboard_move(board, te);
      }
    }
    daemon_dboard_free(board);
  }

  return bookwindow;
}

void daemon_canvas_bookwindow_add_dte(Canvas *canvas, DTe *te) {
  GtkWidget *booklist;
  GtkWidget *item;
  char buf[BUFSIZ];
  gint count;

  g_assert(canvas != NULL);
  g_assert(te != NULL);

  if (canvas->bookwindow == NULL) {
    return;
  }

  booklist = lookup_widget(canvas->bookwindow, "booklist");
  g_assert(booklist != NULL);

  count = canvas->record->mi.count;

  create_te_string(&(canvas->board), count, te, buf);
  item = gtk_list_item_new_with_label(buf);
  gtk_container_add(GTK_CONTAINER(booklist), item);
  gtk_widget_show(item);

  gtk_signal_connect (GTK_OBJECT (item), "button_press_event",
		      GTK_SIGNAL_FUNC (daemon_canvas_bookwindow_click),
		      (gpointer)count);
}

void daemon_canvas_bookwindow_activate(Canvas *canvas,
				       GtkMenuItem *menuitem,
				       gpointer user_data) {
  GtkCheckMenuItem *checkmenuitem;

  checkmenuitem = GTK_CHECK_MENU_ITEM(menuitem);

  if (canvas->flg_bookwindow == TRUE) {
    /* ɽ */
    canvas->flg_bookwindow = FALSE;
    checkmenuitem->active = 0;
    gtk_widget_show(GTK_WIDGET(checkmenuitem));
    gtk_widget_hide(canvas->bookwindow);
    return;
  }

  canvas->flg_bookwindow = TRUE;
  checkmenuitem->active = 1;
  gtk_widget_show(GTK_WIDGET(checkmenuitem));

  if (canvas->bookwindow == NULL) {
    canvas->bookwindow = daemon_canvas_create_bookwindow(canvas);
  }

  gtk_widget_show(canvas->bookwindow);  
}

void daemon_canvas_bookwindow_click(GtkWidget       *widget,
				    GdkEventButton  *event,
				    gpointer         user_data) {
  DTe *te;
  gint i;
  extern Canvas *g_canvas;

  g_assert(g_canvas != NULL);

  if (g_canvas->mode != D_CANVAS_MODE_BOOK) {
    return;
  }

  while (daemon_dboard_can_back(&(g_canvas->board))) {
      daemon_dboard_back(&(g_canvas->board));
      g_canvas->count--;
  }

  for (i=0; i < (gint)user_data; i++) {
    te = daemon_dmoveinfo_get(&(g_canvas->record->mi), g_canvas->count++);
    daemon_dboard_move(&(g_canvas->board), te);
  }

  daemon_canvas_set_kif_sensitive(g_canvas);

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(g_canvas);

  /* ̺ */
  daemon_canvas_draw_back(g_canvas);
  daemon_canvas_draw_sprite_all(g_canvas);
  daemon_canvas_draw_time(g_canvas);
  daemon_canvas_draw_all(g_canvas);
}

void daemon_canvas_bookwindow_destroy(Canvas *canvas,
				      GtkObject *object,
				      gpointer user_data) {
  gtk_widget_destroy(canvas->bookwindow);
  canvas->bookwindow = NULL;
  canvas->flg_bookwindow = FALSE;
}

void daemon_canvas_change_size(Canvas *canvas, D_CANVAS_SIZE size) {
  g_assert(canvas != NULL);
  g_assert(size == D_CANVAS_SIZE_SMALL ||
	   size == D_CANVAS_SIZE_MIDDLE ||
	   size == D_CANVAS_SIZE_BIG);
  
  daemon_canvas_set_size(canvas, size);

  gtk_widget_set_usize(canvas->drawingarea, canvas->width, canvas->height);

  daemon_canvas_change_size_sprite(canvas, size);
  daemon_canvas_dispose_sprite(canvas);

  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_all(canvas);
  gtk_widget_show(canvas->drawingarea);
}

void daemon_canvas_change_toolbar(Canvas *canvas,
				  D_CANVAS_TOOLBAR toolbar_size) {
  GtkHandleBox *handlebox;
  GtkToolbar *toolbar;

  handlebox = (GtkHandleBox *)
    lookup_widget(canvas->window, "handlebox_toolbar");
  toolbar = (GtkToolbar *)lookup_widget(canvas->window, "toolbar");

  switch (toolbar_size) {
  case D_CANVAS_TOOLBAR_NONE:
    gtk_widget_hide(GTK_WIDGET(handlebox));
    break;
  case D_CANVAS_TOOLBAR_ICON:
    gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
    break;
  case D_CANVAS_TOOLBAR_TEXT:
    gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_TEXT);
    break;
  case D_CANVAS_TOOLBAR_BOTH:
    gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_BOTH);
    break;
  default:
    break;
  }

  if (toolbar_size != D_CANVAS_TOOLBAR_NONE) {
    gtk_widget_show(GTK_WIDGET(handlebox));
    gtk_widget_queue_resize(GTK_WIDGET(handlebox));
  }
}

void daemon_canvas_change_mode(Canvas *canvas, D_CANVAS_MODE mode) {
  D_CANVAS_MODE before_mode;

  before_mode = daemon_canvas_get_mode(canvas);

  if (mode == D_CANVAS_MODE_EDIT) {
    daemon_dboard_set_for_edit(&(canvas->board));
    if (canvas->flg_bookwindow == TRUE) {
      gtk_widget_destroy(canvas->bookwindow);
      canvas->flg_bookwindow = TRUE;
    }
    if (canvas->record != NULL) {
      canvas->record->mi.count = 0;
    }
    canvas->board.count = 0;
  }

  daemon_canvas_set_mode(canvas, mode);

  /* طʤβԽ⡼Ѥˤ */
  daemon_canvas_create_back_pixmap(canvas);

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  /* ܥ󡢥˥塼λѲġԲĤꤹ */
  daemon_canvas_set_sensitive(canvas);
  
  if (mode == D_CANVAS_MODE_BOOK) {
    daemon_canvas_set_kif_sensitive(canvas);
    if (canvas->flg_bookwindow == TRUE) {
      gtk_widget_destroy(canvas->bookwindow);
      canvas->bookwindow = daemon_canvas_create_bookwindow(canvas);
      gtk_widget_show(canvas->bookwindow);  
      canvas->flg_bookwindow = TRUE;
    }
  }

  daemon_canvas_update_statusbar(canvas);
}

/** ̤礭ѹ˹碌ƥץ饤Ȥ礭ѹ */
void daemon_canvas_change_size_sprite(Canvas *canvas, D_CANVAS_SIZE size) {
  Sprite *sprite;
  GdkPixmap *koma;
  GdkBitmap *mask;
  gint width, height;
  gint i;

  koma   = NULL;
  mask   = NULL;
  width  = 0;
  height = 0;

  if (size == D_CANVAS_SIZE_SMALL) {
    koma = canvas->koma[0];
    mask = canvas->mask[0];
    width = 40;
    height = 40;
  } else if (size == D_CANVAS_SIZE_MIDDLE) {
    koma = canvas->koma[1];
    mask = canvas->mask[1];
    width = 50;
    height = 50;
  } else if (size == D_CANVAS_SIZE_BIG) {
    koma = canvas->koma[2];
    mask = canvas->mask[2];
    width = 64;
    height = 64;
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }

  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_canvas_get_sprite(canvas, i);

    if (daemon_sprite_get_type(sprite) == ETC_SPRITE) {
      continue;
    }

    daemon_sprite_set_pixmap(sprite, koma);
    daemon_sprite_set_mask(sprite, mask);
    daemon_sprite_set_width(sprite, width);
    daemon_sprite_set_height(sprite, height);
  }
}

void daemon_canvas_timeout(gpointer data) {
  Canvas *canvas;

  canvas = (Canvas *)data;
  daemon_canvas_draw_time(canvas);
}

void daemon_kiffile_load_ok_button_pressed(GtkButton *button,
					   gpointer user_data) {
  GtkFileSelection *fileselection;
  gchar *buf;
  char filename[BUFSIZ];
  Record *record;
  D_LOADSTAT stat;
  GtkWidget *optionmenu;
  GtkWidget *widget;
  GtkWidget *menuitem;
  gchar *menuitem_name;
  extern Canvas *g_canvas;

  const gchar *optionmenu_load = "optionmenu_load";

  fileselection = (GtkFileSelection *)user_data;

  buf = gtk_file_selection_get_filename(fileselection);
  strncpy(filename, buf, BUFSIZ);

  optionmenu = (GtkWidget *)GTK_OPTION_MENU(lookup_widget(GTK_WIDGET(fileselection),
					      optionmenu_load));
  g_assert(optionmenu != NULL);
  widget = gtk_option_menu_get_menu(GTK_OPTION_MENU(optionmenu));
  
  menuitem = gtk_menu_get_active(GTK_MENU(widget));
  menuitem_name = gtk_widget_get_name(GTK_WIDGET(menuitem));

  record = NULL;

  if (strcmp(menuitem_name, "menu_item_auto") == 0) {
    record = daemon_record_load(filename);
  } else if (strcmp(menuitem_name, "menu_item_csa_sjis") == 0) {
    record = daemon_record_load_csa_code(filename, "shift_jis");
  } else if (strcmp(menuitem_name, "menu_item_csa_euc") == 0) {
    record = daemon_record_load_csa_code(filename, "euc-jp");
  } else if (strcmp(menuitem_name, "menu_item_kif_sjis") == 0) {
    record = daemon_record_load_kif_code(filename, "shift_jis");
  } else if (strcmp(menuitem_name, "menu_item_kif_euc") == 0) {
    record = daemon_record_load_kif_code(filename, "euc-jp");
  } else {
    g_assert(1);
  }

  if (record == NULL) {
    /* եɤ߹߼ */
    gtk_widget_destroy(GTK_WIDGET(fileselection));
    daemon_messagebox(_("A file did not open."));
    return;
  }

  stat = daemon_record_get_loadstat(record);

  if (stat == D_LOAD_SUCCESS) {
    /* եɤ߹ */

    g_canvas->record = record;
    g_canvas->board.front = D_SENTE;
    daemon_dboard_copy(&(record->first_board), &(g_canvas->board));

    daemon_canvas_change_mode(g_canvas, D_CANVAS_MODE_BOOK);
    g_canvas->count = 0;

    if (g_canvas->bookwindow != NULL) {
      gtk_widget_destroy(g_canvas->bookwindow);
    }

    /* 襦ɥɽä */
    if (g_canvas->flg_bookwindow == TRUE) {
      g_canvas->bookwindow = daemon_canvas_create_bookwindow(g_canvas);
      gtk_widget_show(g_canvas->bookwindow);  
      g_canvas->flg_bookwindow = TRUE;
    }
  } else {
    /* եɤ߹߼ */
    daemon_record_free(record);
    daemon_messagebox(_("A file did not open."));
  }

  gtk_widget_destroy(GTK_WIDGET(fileselection));
}

void daemon_kiffile_load_cancel_button_pressed(GtkButton*button,
					       gpointer user_data) {
  GtkFileSelection *fileselection;
  fileselection = (GtkFileSelection *)user_data;
  gtk_widget_destroy(GTK_WIDGET(fileselection));
}

void daemon_kiffile_save_ok_button_pressed(GtkButton *button,
					   gpointer user_data) {
  GtkFileSelection *fileselection;
  gchar *buf;
  char filename[BUFSIZ];
  extern Canvas *g_canvas;
  int ret;
  GtkWidget *optionmenu;
  GtkWidget *widget;
  GtkWidget *menuitem;
  gchar *menuitem_name;

  fileselection = (GtkFileSelection *)user_data;
  g_assert(fileselection != NULL);

  buf = gtk_file_selection_get_filename(fileselection);
  strncpy(filename, buf, BUFSIZ - 1);
  filename[BUFSIZ - 1] = '\0';

  if (g_canvas->record == NULL) {
    g_canvas->record = daemon_record_new();
    daemon_dboard_copy(&(g_canvas->board), &(g_canvas->record->first_board));
  }

  optionmenu = (GtkWidget *)lookup_widget(GTK_WIDGET(fileselection),
					      "optionmenu_save");
  g_assert(optionmenu != NULL);
  widget = gtk_option_menu_get_menu(GTK_OPTION_MENU(optionmenu));
  
  menuitem = gtk_menu_get_active(GTK_MENU(widget));
  menuitem_name = gtk_widget_get_name(GTK_WIDGET(menuitem));

  ret = -1;
  if (strcmp(menuitem_name, "menu_item_csa_sjis") == 0) {
    ret = daemon_record_output_csa(g_canvas->record, filename); 
  } else if (strcmp(menuitem_name, "menu_item_csa_euc") == 0) {
    ret = daemon_record_output_csa_code(g_canvas->record, filename, "euc-jp"); 
  } else if (strcmp(menuitem_name, "menu_item_kif_sjis") == 0) {
    ret = daemon_record_output_kif(g_canvas->record, filename); 
  } else if (strcmp(menuitem_name, "menu_item_kif_euc") == 0) {
    ret = daemon_record_output_kif_code(g_canvas->record, filename, "euc-jp"); 
  } else {
    /* ꤨʤʬ */
    g_assert(1);
  }

  gtk_widget_destroy(GTK_WIDGET(fileselection));

  if (ret != 0) {
    /* ե񤭹߼ */
    daemon_messagebox(_("The writing of a file went wrong."));
    return;
  }
}

void daemon_kiffile_save_cancel_button_pressed(GtkButton*button,
					       gpointer user_data) {
  GtkFileSelection *fileselection;
  fileselection = (GtkFileSelection *)user_data;
  gtk_widget_destroy(GTK_WIDGET(fileselection));
}

/**
 * ޥ뤬canvasФȤ˸ƤФ롣
 * ɥåʤХץ饤Ȥ᤹
 */
void daemon_canvas_leave_notify(Canvas *canvas,
				GtkWidget *widget,
				GdkEventCrossing *event,
				gpointer user_data) {
  gint no;
  Sprite *sprite;
  gint x, y;
  GdkPoint diff;
  GdkRectangle rect;

  if (daemon_canvas_isdrag(canvas) == FALSE) {
    return;
  }

  daemon_canvas_set_drag(canvas, FALSE);
  no     = daemon_canvas_get_dragno(canvas);
  sprite = daemon_canvas_get_sprite(canvas, no);
  diff   = daemon_canvas_get_diff(canvas);
  x      = daemon_sprite_get_from_x(sprite);
  y      = daemon_sprite_get_from_y(sprite);

  rect.x      = daemon_sprite_get_x(sprite);
  rect.y      = daemon_sprite_get_y(sprite);
  rect.width  = daemon_sprite_get_width(sprite);
  rect.height = daemon_sprite_get_height(sprite);

  daemon_sprite_set_x(sprite, x);
  daemon_sprite_set_y(sprite, y);

  /*  */
  daemon_canvas_draw_back_update_from_tmp_pixmap(canvas, &rect);
  daemon_canvas_draw_sprite(canvas, no);
  daemon_canvas_draw_all(canvas);
}

/** ޤ᤹ */
void daemon_canvas_kif_first(Canvas *canvas) {
  g_assert(canvas != NULL);

  if (canvas->mode != D_CANVAS_MODE_BOOK) {
    return;
  }

  while (daemon_dboard_can_back(&(canvas->board))) {
      daemon_dboard_back(&(canvas->board));
      canvas->count--;
  }

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  daemon_canvas_set_kif_sensitive(canvas);
}

void daemon_canvas_kif_back(Canvas *canvas) {
  g_assert(canvas != NULL);

  if (canvas->mode != D_CANVAS_MODE_BOOK) {
    return;
  }

  if (daemon_dboard_can_back(&(canvas->board)) == 0) {
    return;
  }

  daemon_dboard_back(&(canvas->board));
  canvas->count--;

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  daemon_canvas_set_kif_sensitive(canvas);
}

void daemon_canvas_kif_next(Canvas *canvas) {
  DTe *te;

  g_assert(canvas != NULL);

  if (canvas->mode != D_CANVAS_MODE_BOOK) {
    return;
  }

  if (daemon_dmoveinfo_size(&(canvas->record->mi)) <= canvas->count) {
    return;
  }

  te = daemon_dmoveinfo_get(&(canvas->record->mi), canvas->count++);
  daemon_dboard_move(&(canvas->board), te);

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  daemon_canvas_set_kif_sensitive(canvas);
}

void daemon_canvas_kif_last(Canvas *canvas) {
  DTe *te;

  g_assert(canvas != NULL);

  if (canvas->mode != D_CANVAS_MODE_BOOK) {
    return;
  }

  while (canvas->count < daemon_dmoveinfo_size(&(canvas->record->mi))) {
    te = daemon_dmoveinfo_get(&(canvas->record->mi), canvas->count++);
    daemon_dboard_move(&(canvas->board), te);
  }

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  daemon_canvas_set_kif_sensitive(canvas);
}

/** ⡼ɤǥܥ򲡤ȤˤФ */
void daemon_canvas_set_kif_sensitive(Canvas *canvas) {
  GtkWidget *widget;

  g_assert(canvas->mode == D_CANVAS_MODE_BOOK);

  if (canvas->count == 0) {
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, FALSE);
  } else if (daemon_dmoveinfo_size(&(canvas->record->mi)) == canvas->count) {
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, FALSE);
  } else {
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, FALSE);
  }
}

void daemon_canvas_set_sensitive(Canvas *canvas) {
  GtkWidget *widget;

  switch (daemon_canvas_get_mode(canvas)) {
  case D_CANVAS_MODE_WAIT:
    break;
  case D_CANVAS_MODE_BOOK:
    widget = lookup_widget(canvas->window, "button_stop");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "stop");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "startedit");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "set_hirate");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "set_mate");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "rl_reversal");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "order_reversal");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "edit_flip");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "all_koma_to_pbox");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "button_new_game");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "playgame");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "check_mate");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "fileopen");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "filesave");
    gtk_widget_set_sensitive(widget, TRUE);
    break;
  case D_CANVAS_MODE_EDIT:
    widget = lookup_widget(canvas->window, "button_stop");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "stop");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "startedit");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "set_hirate");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "set_mate");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "rl_reversal");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "order_reversal");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "edit_flip");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "all_koma_to_pbox");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "button_new_game");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "playgame");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "check_mate");
    gtk_widget_set_sensitive(widget, TRUE);

    widget = lookup_widget(canvas->window, "fileopen");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "filesave");
    gtk_widget_set_sensitive(widget, TRUE);
    break;
  case D_CANVAS_MODE_GAME:
    widget = lookup_widget(canvas->window, "button_stop");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_give_up");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "retract");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "stop");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "restart");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "give_up");
    gtk_widget_set_sensitive(widget, TRUE);
    widget = lookup_widget(canvas->window, "first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "last");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "startedit");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "set_hirate");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "set_mate");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "rl_reversal");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "order_reversal");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "edit_flip");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "all_koma_to_pbox");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "button_new_game");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "playgame");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "check_mate");
    gtk_widget_set_sensitive(widget, FALSE);

    widget = lookup_widget(canvas->window, "fileopen");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "filesave");
    gtk_widget_set_sensitive(widget, FALSE);
    break;
  case D_CANVAS_MODE_MATE:
    widget = lookup_widget(canvas->window, "button_stop");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_first");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_back");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_next");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_last");
    gtk_widget_set_sensitive(widget, FALSE);
    break;
  default:
    break;
  }
}

void daemon_canvas_flip_activate(Canvas *canvas) {
  daemon_messagebox(_("Conversely, it displays."));

  daemon_dboard_toggle_front(&(canvas->board));

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);
}

void daemon_canvas_mate_activate(Canvas *canvas) {
  GtkWidget *dialog;

  dialog = create_dialog_mate();
  canvas->dialog_mate = dialog;

  gtk_widget_show(dialog);
}

void daemon_canvas_mate_start_pressed(Canvas *canvas) {
  GtkWidget *dialog;
  GtkWidget *sente;
  TREE root;
  int ret;
  int i;
  extern BOARD g_board;
  extern GAME g_game;
  extern int MATE_DEEP_MAX;
  MOVEINFO *mi;

  dialog = canvas->dialog_mate;
  sente  = lookup_widget(dialog, "radiobutton_mate_sente");
  mi = NULL;
  
  daemon_dboard_copy_board(&(canvas->board), &g_board);
  boPlayInit(&g_board);
  boSetTsumeShogiPiece(&g_board);

  daemon_dboard_copy_from_board(&g_board, &(canvas->board));

  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(sente)) == TRUE) {
    g_board.next = SENTE;
    canvas->board.next = D_SENTE;
  } else {
    g_board.next = GOTE;
    canvas->board.next = D_GOTE;
  }

  gtk_widget_destroy(dialog);

  if (boCheckJustMate(&g_board) == 0) {
    daemon_messagebox(_("No good board."));
    return;
  }

  dialog = create_thinking_messagebox();
  gtk_widget_show(dialog);

  gmSetGameStatus(&g_game, SI_NORMAL);

#ifdef USE_HASH  
  /* ϥåơ֥ */
  newHASH();
#endif /* USE_HASH */

  initTREE(&root);

  ret = 0;

  for ( i=0; i<8; i++ ) {
    MATE_DEEP_MAX = i;
    if ( mate(&g_board, &root, 0) ) {
      ret = 1;
      /* ͤ */
      mi = newMOVEINFO();
      trSetMoveinfo(root.child, mi);
      break;
    }
  }

#ifdef USE_HASH  
  /* ϥåơ֥ */
  freeHASH();
#endif /* USE_HASH */

  gtk_widget_destroy(dialog);

  if (ret) {
    DTe te;

    daemon_messagebox(_("Check mate."));

    if (canvas->record != NULL) {
      daemon_record_free(canvas->record);
    }
    canvas->record = daemon_record_new();
    daemon_dboard_copy(&(canvas->board), &(canvas->record->first_board));
    for (i=0; i<mi->count; i++) {
      te.fm   = mi->te[i].fm;
      te.to   = mi->te[i].to;
      te.nari = mi->te[i].nari;
      te.uti  = mi->te[i].uti;
      daemon_record_add_dte(canvas->record, &te, 0);
    }
    freeMOVEINFO(mi);
    canvas->count = 0;
    canvas->board.count = 0;
    daemon_canvas_change_mode(canvas, D_CANVAS_MODE_BOOK);
  } else {
    daemon_messagebox(_("No check mate."));
  }
}

void daemon_canvas_mate_cancel_pressed(Canvas *canvas) {
  GtkWidget *dialog;

  dialog = canvas->dialog_mate;

  gtk_widget_destroy(dialog);
}

void daemon_canvas_set_hirate_activate(Canvas *canvas) {
  gint ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("Set hirate?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_set_hirate(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);
}

void daemon_canvas_set_mate_activate(Canvas *canvas) {
  int ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("Set for check mate?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_set_mate(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);

#ifdef NOREAD
  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);
#endif /* NOREAD */
}

void daemon_canvas_set_rl_reversal_activate(Canvas *canvas) {
  int ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("right/left reversal?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_set_rl_reversal(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);

#ifdef NOREAD
  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);
#endif /* NOREAD */
}

void daemon_canvas_set_order_reversal_activate(Canvas *canvas) {
  int ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("order reversal?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_set_order_reversal(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);
}

void daemon_canvas_set_edit_flip_activate(Canvas *canvas) {
  int ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("flip white/black board?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_flip(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);
}

void daemon_canvas_all_koma_to_pbox_activate(Canvas *canvas) {
  int ret;

  ret = daemon_canvas_ok_cancel_dialog(canvas, _("All piece to piecebox?"));

  if (GNOME_OK != ret) {
    return;
  }

  daemon_dboard_set_all_to_pbox(&(canvas->board));

  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_EDIT);
}

gint daemon_canvas_ok_cancel_dialog(Canvas *canvas, char *message) {
  GtkWidget *dialog;
  gint reply;

  dialog = 
  gnome_ok_cancel_dialog_modal_parented(message,
					daemon_canvas_ok_cancel_dialog_cb,
					&reply,
					GTK_WINDOW(canvas->window));

  gnome_dialog_run_and_close(GNOME_DIALOG(dialog));

  return reply;
}

void daemon_canvas_ok_cancel_dialog_cb(gint reply, gpointer data) {
  gint *r;

  r = (gint *)data;
  *r = reply;
}

void daemon_canvas_yes_no_dialog_cb(gint reply, gpointer data) {
  gint *r;

  r = (gint *)data;
  *r = reply;
}

int daemon_canvas_yes_no_dialog(Canvas *canvas, char *message) {
  GtkWidget *dialog;
  gint reply;

  dialog = 
  gnome_question_dialog_modal_parented(message,
				       daemon_canvas_yes_no_dialog_cb,
				       &reply,
				       GTK_WINDOW(canvas->window));

  gnome_dialog_run_and_close(GNOME_DIALOG(dialog));

  return reply;
}

void daemon_canvas_mate_think_cancel(Canvas *canvas) {
  extern GAME g_game;

  gmSetGameStatus(&g_game, SI_CHUDAN);
}

void daemon_canvas_new_game(Canvas *canvas) {
  canvas->dialog_game = daemon_create_dialog_game();

  gtk_widget_show(canvas->dialog_game);
}

void daemon_canvas_new_game_ok(GtkButton *button, gpointer user_data) {
  extern Canvas *g_canvas;
  extern BOARD g_board;
  PLAYERTYPE playertype[2];
  TEAI teai;
  char *player[2];
  char *t;
  char *playername[2];
  char *tmp;

  GtkWidget *menu_sente;
  GtkWidget *menu_gote;
  GtkWidget *menu_teai;
  GtkWidget *option_sente;
  GtkWidget *option_gote;
  GtkWidget *option_teai;
  GtkWidget *selected_sente;
  GtkWidget *selected_gote;
  GtkWidget *selected_teai;
  GtkWidget *entry_sente;
  GtkWidget *entry_gote;

  menu_sente = lookup_widget(g_canvas->dialog_game, "optionmenu_sente");
  menu_gote  = lookup_widget(g_canvas->dialog_game, "optionmenu_gote");
  menu_teai  = lookup_widget(g_canvas->dialog_game, "optionmenu_teai");
  option_sente = gtk_option_menu_get_menu(GTK_OPTION_MENU(menu_sente));
  option_gote  = gtk_option_menu_get_menu(GTK_OPTION_MENU(menu_gote));
  option_teai  = gtk_option_menu_get_menu(GTK_OPTION_MENU(menu_teai));
  selected_sente = gtk_menu_get_active(GTK_MENU(option_sente));
  selected_gote  = gtk_menu_get_active(GTK_MENU(option_gote));
  selected_teai  = gtk_menu_get_active(GTK_MENU(option_teai));

  player[0]     = gtk_widget_get_name(selected_sente);
  player[1]     = gtk_widget_get_name(selected_gote);
  playertype[0] = SI_GUI_HUMAN;
  playertype[1] = SI_COMPUTER_1;
  playername[0] = player[0];
  playername[1] = player[1];
  t             = gtk_widget_get_name(selected_teai);

  if (strncmp(player[0], "HUMAN", 5) == 0) {
    playertype[0] = SI_GUI_HUMAN;
  } else if (strncmp(player[0], "LEVEL1", 6) == 0) {
    playertype[0] = SI_COMPUTER_1;
  } else if (strncmp(player[0], "LEVEL2", 6) == 0) {
    playertype[0] = SI_COMPUTER_2;
  } else if (strncmp(player[0], "LEVEL3", 7) == 0) {
    playertype[0] = SI_COMPUTER_3;
  } else if (strncmp(player[0], "SERIALPORT1", 11) == 0) {
    playertype[0] = SI_SERIALPORT_1;
  } else if (strncmp(player[0], "SERIALPORT2", 11) == 0) {
    playertype[0] = SI_SERIALPORT_2;
  } else {
    g_assert(1);
  }
  if (strncmp(player[1], "HUMAN", 5) == 0) {
    playertype[1] = SI_GUI_HUMAN;
  } else if (strncmp(player[1], "LEVEL1", 6) == 0) {
    playertype[1] = SI_COMPUTER_1;
  } else if (strncmp(player[1], "LEVEL2", 6) == 0) {
    playertype[1] = SI_COMPUTER_2;
  } else if (strncmp(player[1], "LEVEL3", 6) == 0) {
    playertype[1] = SI_COMPUTER_3;
  } else if (strncmp(player[1], "SERIALPORT1", 11) == 0) {
    playertype[1] = SI_SERIALPORT_1;
  } else if (strncmp(player[1], "SERIALPORT2", 11) == 0) {
    playertype[1] = SI_SERIALPORT_2;
  } else {
    g_assert(1);
  }
  teai = 0;
  if (strncmp(t, "HIRATE", 6) == 0) {
    teai = HIRATE;
    daemon_dboard_set_hirate(&(g_canvas->board));
#ifdef NOREAD
  } else if (strncmp(t, "NIMAIOTI", 8) == 0) {
    teai = NIMAIOTI;
  } else if (strncmp(t, "YONMAIOTI", 9) == 0) {
    teai = YONMAIOTI;
#endif /* NOREAD */
  } else if (strncmp(t, "ETC", 3) == 0) {
    teai = TEAI_ETC;
  } else {
    g_assert(1);
  }

  if (teai == TEAI_ETC) {
    daemon_dboard_copy_board(&(g_canvas->board), &g_board);
    if (boCheckJustGame(&g_board) == 0) {
      gtk_widget_destroy(g_canvas->dialog_game);
      daemon_messagebox(_("No good board."));
      return;
    }
  }

  entry_sente = lookup_widget(g_canvas->dialog_game, "entry_sente");
  entry_gote  = lookup_widget(g_canvas->dialog_game, "entry_gote");

  tmp = gtk_entry_get_text(GTK_ENTRY(entry_sente));
  if (tmp[0] == '\0') {
    if (playertype[0] == SI_GUI_HUMAN) {
      uid_t uid;
      struct passwd *curr_passwd;
   
      uid = getuid();
      curr_passwd = getpwuid(uid);
      
      playername[0] = curr_passwd->pw_name;
    }
  } else {
    playername[0] = tmp;
  }
  tmp = gtk_entry_get_text(GTK_ENTRY(entry_gote));
  if (tmp[0] == '\0') {
    if (playertype[1] == SI_GUI_HUMAN) {
      uid_t uid;
      struct passwd *curr_passwd;
   
      uid = getuid();
      curr_passwd = getpwuid(uid);
      
      playername[1] = curr_passwd->pw_name;
    }
  } else {
    playername[1] = tmp;
  }

  if (g_canvas->record == NULL) {
    g_canvas->record = daemon_record_new();
  }

  daemon_record_set_player(g_canvas->record, D_SENTE, playername[0]);
  daemon_record_set_player(g_canvas->record, D_GOTE, playername[1]);

  gtk_widget_destroy(g_canvas->dialog_game);

  g_canvas->playertype[0] = playertype[0];
  g_canvas->playertype[1] = playertype[1];

  g_canvas->record->mi.count = 0;
  g_canvas->record->matchstat = D_UNKNOWN;
  daemon_canvas_change_mode(g_canvas, D_CANVAS_MODE_GAME);

  daemon_canvas_game(g_canvas);
}

void daemon_canvas_new_game_cancel(GtkButton *button, gpointer user_data) {
  extern Canvas *g_canvas;

  gtk_widget_destroy(g_canvas->dialog_game);
}

void daemon_canvas_new_game_player_cb(GtkButton *button, gpointer user_data) {
  gint player;
  extern Canvas *g_canvas;

  player = (gint)user_data;

  if (player < 100) {
    g_canvas->playertype[D_SENTE] = player;
  } else {
    g_canvas->playertype[D_GOTE] = player;
  }
}

void daemon_canvas_new_game_teai_cb(GtkButton *button, gpointer user_data) {
  gint teai;
  extern Canvas *g_canvas;

  teai = (gint)user_data;

  g_canvas->teai = teai;
}

void daemon_canvas_stop_activate(Canvas *canvas) {
  extern GAME g_game;

  g_assert(canvas != NULL);

  if (daemon_canvas_get_mode(canvas) != D_CANVAS_MODE_GAME) {
    return;
  }

  gmSetGameStatus(&g_game, SI_CHUDAN);
}

void daemon_canvas_restart_activate(Canvas *canvas) {
  extern BOARD g_board;

  g_assert(canvas != NULL);

  if (daemon_canvas_get_mode(canvas) != D_CANVAS_MODE_GAME) {
    return;
  }

  if (g_board.mi.count < 1) {
    return;
  }

  boToggleNext( &g_board );
  boBack( &g_board, g_board.next );
  boToggleNext( &g_board );
  boBack( &g_board, g_board.next );

  daemon_dboard_back(&(canvas->board));
  daemon_dboard_back(&(canvas->board));

  canvas->record->mi.count -= 2;

  if (canvas->flg_bookwindow == TRUE) {
    gtk_widget_destroy(canvas->bookwindow);
    canvas->bookwindow = daemon_canvas_create_bookwindow(canvas);
    canvas->flg_bookwindow = TRUE;
    gtk_widget_show(canvas->bookwindow);
  }
	
  /* ץ饤Ⱥ */
  daemon_canvas_dispose_sprite(canvas);

  /* ̺ */
  daemon_canvas_draw_back(canvas);
  daemon_canvas_draw_sprite_all(canvas);
  daemon_canvas_draw_time(canvas);
  daemon_canvas_draw_all(canvas);
}

void daemon_canvas_give_up_activate(Canvas *canvas) {
  extern GAME g_game;

  g_assert(canvas != NULL);

  if (daemon_canvas_get_mode(canvas) != D_CANVAS_MODE_GAME) {
    return;
  }

  gmSetGameStatus(&g_game, SI_TORYO);
}

void daemon_canvas_statusbar_activate(Canvas *canvas) {
  GtkWidget *widget;

  g_assert(canvas != NULL);

  widget = lookup_widget(canvas->window, "statusbar");

  if (canvas->statusbar == TRUE) {
    gtk_widget_hide(widget);
    canvas->statusbar = FALSE;
  } else {
    gtk_widget_show(widget);
    canvas->statusbar = TRUE;
  }
}

/*************************************************************/
/* draw */
/*************************************************************/

void daemon_canvas_draw_all(Canvas *canvas) {
  int width, height;
  GdkGC* gc;
  GtkWidget* drawingarea;
  GdkPixmap* pixmap;

  width       = canvas->width;
  height      = canvas->height;
  gc          = canvas->gc;
  drawingarea = canvas->drawingarea;
  pixmap      = daemon_canvas_get_pixmap(canvas);

  gdk_draw_pixmap(drawingarea -> window,
		  drawingarea -> style -> fg_gc[GTK_WIDGET_STATE(drawingarea)],
		  pixmap,
		  0, 0,
		  0, 0,
		  width, height);
}

void daemon_canvas_draw_all_update(Canvas *canvas, GdkRectangle *rect) {
  GdkGC* gc;
  GtkWidget* drawingarea;
  GdkPixmap* pixmap;

  gc          = canvas->gc;
  drawingarea = canvas->drawingarea;
  pixmap      = daemon_canvas_get_pixmap(canvas);

  gdk_draw_pixmap(drawingarea -> window,
		  drawingarea -> style -> fg_gc[GTK_WIDGET_STATE(drawingarea)],
		  pixmap,
		  rect->x, rect->y,
		  rect->x, rect->y,
		  rect->width, rect->height);
}

void daemon_canvas_draw_back(Canvas *canvas) {
  GdkRectangle rect;
  
  rect.x = 0;
  rect.y = 0;
  rect.width  = canvas->width;
  rect.height = canvas->height;

  daemon_canvas_draw_back_update(canvas, &rect);
}

void daemon_canvas_draw_back_update(Canvas *canvas, GdkRectangle* rect) {
  GdkGC* gc;
  GdkPixmap* pixmap;
  GdkPixmap* back;
  
  gc     = canvas->gc;
  pixmap = daemon_canvas_get_pixmap(canvas);
  back   = daemon_canvas_get_back(canvas);
  
  set_color(gc, 0, 0, 0);
  gdk_gc_set_clip_rectangle(gc, rect);
  gdk_draw_pixmap(pixmap, gc, back, rect->x, rect->y, rect->x, rect->y,
		  rect->width, rect->height);
}

void daemon_canvas_draw_back_update_from_tmp_pixmap(Canvas *canvas, GdkRectangle *rect) {
  GdkGC* gc;
  GdkPixmap* pixmap;
  GdkPixmap* tmp_pixmap;
  GdkPixmap* back;
  
  gc         = canvas->gc;
  pixmap     = daemon_canvas_get_pixmap(canvas);
  tmp_pixmap = canvas->tmp_pixmap;
  back       = daemon_canvas_get_back(canvas);
  
  gdk_gc_set_clip_rectangle(gc, rect);
  gdk_draw_pixmap(pixmap, gc, tmp_pixmap, rect->x, rect->y, rect->x, rect->y,
		  rect->width, rect->height);
}

void daemon_canvas_draw_sprite_all(Canvas *canvas) {
  Sprite* sprite;
  gint i;

  g_assert(canvas != NULL);

  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_canvas_get_sprite(canvas, i);

    if (sprite == NULL) {
      continue;
    }
    if (daemon_sprite_isvisible(sprite) == FALSE) {
      continue;
    }

    daemon_canvas_draw_sprite(canvas, i);
  }
}

void daemon_canvas_draw_sprite_update_all(Canvas *canvas, GdkRectangle* rect) {
  GdkRectangle s_rect;
  Sprite* sprite;
  gint    i;

  g_assert(canvas != NULL);

  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_canvas_get_sprite(canvas, i);

    if (sprite == NULL) {
      continue;
    }
    if (daemon_sprite_isvisible(sprite) == FALSE) {
      continue;
    }
    s_rect.x      = daemon_sprite_get_x(sprite);
    s_rect.y      = daemon_sprite_get_y(sprite);
    s_rect.width  = daemon_sprite_get_width(sprite);
    s_rect.height = daemon_sprite_get_height(sprite);
    if (is_on_rectangle(rect, &s_rect)) {
      daemon_canvas_draw_sprite(canvas, i);
    }
  } 
}

void daemon_canvas_draw_sprite(Canvas *canvas, gint no) {
  GdkGC*     gc;
  GdkPixmap* pixmap;
  GdkPixmap* sprite_pixmap;
  GdkBitmap* mask;
  Sprite*    sprite;
  gint       width, height;
  gint       x, y;
  gint       src_x, src_y;

  g_assert(canvas != NULL);
  g_assert(0 <= no && no < SPRITEMAX);

  sprite = daemon_canvas_get_sprite(canvas, no);
  
  if (daemon_sprite_isvisible(sprite) == FALSE) {
    /* ɽʤ */
    return;
  }
  
  gc            = canvas->gc;
  pixmap        = daemon_canvas_get_pixmap(canvas);
  sprite_pixmap = daemon_sprite_get_pixmap(sprite);
  mask          = daemon_sprite_get_mask(sprite);
  width         = daemon_sprite_get_width(sprite);
  height        = daemon_sprite_get_height(sprite);
  x             = daemon_sprite_get_x(sprite);
  y             = daemon_sprite_get_y(sprite);
  src_x         = daemon_sprite_get_src_x(sprite);
  src_y         = daemon_sprite_get_src_y(sprite);

  if (sprite->type == KOMA_SPRITE) {
    /* ξ */
    gdk_gc_set_clip_mask(gc, mask);
    gdk_gc_set_clip_origin(gc, x - src_x, y - src_y);
  } else {
    /* ʳ */
    GdkRectangle rect;
    
    rect.x = 0;
    rect.y = 0;
    rect.width  = sprite->width;
    rect.height = sprite->height;
  
    gdk_gc_set_clip_rectangle(gc, &rect);
    gdk_gc_set_clip_origin(gc, x, y);
  }
  gdk_draw_pixmap(pixmap, gc, sprite_pixmap, src_x, src_y, x, y,
		  width, height);
}

/** ֤ */
void daemon_canvas_draw_time(Canvas *canvas) {
  GdkGC*     gc;
  GdkPixmap* pixmap;
  GdkPixmap* back;
  GdkFont *font;
  GdkRectangle rect;
  gint x, y, width, height, step;
  gchar buf[BUFSIZ];
  extern BOARD g_board;
  
  g_assert(canvas != NULL); 
  
  gc          = canvas->gc;
  pixmap      = daemon_canvas_get_pixmap(canvas);
  back        = daemon_canvas_get_back(canvas);
  rect.x      = 0;
  rect.y      = 0;
  rect.width  = canvas->width;
  rect.height = canvas->height;
  x = y = width = height = step = 0;

  /*  */
  x      = daemon_canvas_rate_xy(canvas, 516);
  y      = daemon_canvas_rate_xy(canvas, 171);
  width  = daemon_canvas_rate_xy(canvas, 129);
  height = daemon_canvas_rate_xy(canvas,  38);
  step   = daemon_canvas_rate_xy(canvas,  14);

  font = canvas->window->style->font;
  set_color(gc, 0xFFFF, 0xFFFF, 0XFFFF);

  gdk_gc_set_clip_rectangle(gc, &rect);
  gdk_gc_set_clip_origin(gc, 0, 0);
  gdk_draw_pixmap(pixmap, gc, back, x, y, x, y, width, height);
  if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME) {
    int n;
    n = daemon_dboard_get_front(&(canvas->board));
    time2string(buf, g_board.total_time[n]);
    buf[8] = '/';
    time2string(buf +9, g_board.limit_time[n]);
    gdk_draw_string(pixmap, font, gc, x, y + step, buf);
  }
  if (canvas->record != NULL) {
    if (daemon_dboard_get_front(&(canvas->board)) == D_SENTE) {
      sprintf(buf, "%-20s", canvas->record->player[0]);
      buf[16] = '\0';
      gdk_draw_string(pixmap, font, gc, x, y + step * 2 + 3, buf);
    } else {
      sprintf(buf, "%-20s", canvas->record->player[1]);
      buf[16] = '\0';
      gdk_draw_string(pixmap, font, gc, x, y + step * 2 + 3, buf);
    }
  }

  /*  */
  x      = daemon_canvas_rate_xy(canvas,   6);
  y      = daemon_canvas_rate_xy(canvas, 172);
  width  = daemon_canvas_rate_xy(canvas, 129);
  height = daemon_canvas_rate_xy(canvas,  38);
  step   = daemon_canvas_rate_xy(canvas,  14);

  font = canvas->window->style->font;
  set_color(gc, 0xFFFF, 0xFFFF, 0XFFFF);

  gdk_gc_set_clip_rectangle(gc, &rect);
  gdk_gc_set_clip_origin(gc, 0, 0);
  gdk_draw_pixmap(pixmap, gc, back, x, y, x, y, width, height);
  if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME) {
    int n;
    n = daemon_dboard_get_front(&(canvas->board)) ^ 1;
    time2string(buf, g_board.total_time[n]);
    buf[8] = '/';
    time2string(buf +9, g_board.limit_time[n]);
    gdk_draw_string(pixmap, font, gc, x, y + step, buf);
  }
  if (canvas->record != NULL) {
    if (daemon_dboard_get_front(&(canvas->board)) == D_SENTE) {
      sprintf(buf, "%-20s", canvas->record->player[1]);
      buf[16] = '\0';
      gdk_draw_string(pixmap, font, gc, x, y + step * 2 + 3, buf);
    } else {
      sprintf(buf, "%-20s", canvas->record->player[0]);
      buf[16] = '\0';
      gdk_draw_string(pixmap, font, gc, x, y + step * 2 + 3, buf);
    }
  }

  /* ֡μɽ */
  x      = daemon_canvas_rate_xy(canvas, 516);
  y      = daemon_canvas_rate_xy(canvas, 131);
  width  = daemon_canvas_rate_xy(canvas, 129);
  height = daemon_canvas_rate_xy(canvas,  38);
  step   = daemon_canvas_rate_xy(canvas,  16);

  font = canvas->window->style->font;
  set_color(gc, 0xFFFF, 0xFFFF, 0XFFFF);
  
  gdk_gc_set_clip_rectangle(gc, &rect);
  gdk_gc_set_clip_origin(gc, 0, 0);
  gdk_draw_pixmap(pixmap, gc, back, x, y, x, y, width, height);
  if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME) {
    gint count;
    DTe *te;
    sprintf(buf, "%s", canvas->board.next == D_SENTE ? _("next SENTE(WHITE)") : _("next GOTE(BLACK)"));
    y += step;
    gdk_draw_string(pixmap, font, gc, x, y, buf);

    y += step;

    count = canvas->board.mi.count;
    te = &(canvas->board.mi.te[count - 1]);
    create_te_string2(&(canvas->board), count, te, buf);
    gdk_draw_string(pixmap, font, gc, x, y, buf);
  }

  daemon_canvas_draw_all_update(canvas, &rect);
}

void daemon_canvas_update_statusbar(Canvas *canvas) {
  /* ơС򹹿 */
  GtkStatusbar* statusbar;
  static guint context_id = -1;
  char *buf;

  if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_BOOK) {
    buf = _("BOOK MODE");
  } else if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME) {
    buf = _("GAME MODE");
  } else if (daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_EDIT) {
    buf = _("EDIT MODE");
  } else {
    buf = "uknown mode";
  }

  statusbar = (GtkStatusbar*)get_widget(GTK_WIDGET(canvas->window), "statusbar");
  if (context_id != -1) {
    gtk_statusbar_pop(statusbar, context_id);
  }
  context_id = gtk_statusbar_get_context_id(statusbar, "status");
  gtk_statusbar_push(statusbar, context_id, buf);
}

void daemon_canvas_update_statusbar_te(Canvas *canvas, DTe *te) {
  /* ơС򹹿 */
  GtkStatusbar* statusbar;
  static guint context_id = -1;
  char buf[BUFSIZ];
  char buf2[BUFSIZ];
  gint count;

  if (daemon_canvas_get_mode(canvas) != D_CANVAS_MODE_GAME) {
    return;
  }

  count = canvas->record->mi.count;
  create_te_string(&(canvas->board), count, te, buf);

  sprintf(buf2,
	  "%s %s --- %s", _("GAME MODE"),
	  buf, 
	  canvas->board.next == D_GOTE ? _("SENTE(WHITE)") : _("GOTE(BLACK)"));

  statusbar = (GtkStatusbar*)get_widget(GTK_WIDGET(canvas->window), "statusbar");
  if (context_id != -1) {
    gtk_statusbar_pop(statusbar, context_id);
  }
  context_id = gtk_statusbar_get_context_id(statusbar, "status");
  gtk_statusbar_push(statusbar, context_id, buf2);
}

/*************************************************************/
/* etc */
/*************************************************************/

/**
 * canvas->mode ͤ򸫤 back Ƥ񤭴롣
 *
 */
void daemon_canvas_create_back_pixmap(Canvas *canvas) {
  GdkRectangle rect;

  /* κɸ */
  static const gint x1 = 1, y1 = 210, x2 = 129, y2 = 378;

  g_assert(canvas != NULL);

  if (canvas->back[0] != NULL)
    gdk_imlib_free_pixmap(canvas->back[0]);
  if (canvas->back[1] != NULL)
    gdk_imlib_free_pixmap(canvas->back[1]);
  if (canvas->back[2] != NULL)
    gdk_imlib_free_pixmap(canvas->back[2]);

  canvas->back[0] = load_pixmap_from_imlib_image_scaled(canvas->window,
							canvas->back_src,
							D_CANCAS_SIZE_SMALL_X,
							D_CANCAS_SIZE_SMALL_Y);
  canvas->back[1] = load_pixmap_from_imlib_image_scaled(canvas->window,
							canvas->back_src,
							D_CANCAS_SIZE_MIDDLE_X,
							D_CANCAS_SIZE_MIDDLE_Y);
  canvas->back[2] = load_pixmap_from_imlib_image_scaled(canvas->window,
							canvas->back_src,
							D_CANCAS_SIZE_BIG_X,
							D_CANCAS_SIZE_BIG_Y);

  if (canvas->mode == D_CANVAS_MODE_EDIT) {
    return;
  }

  rect.x      = 0;
  rect.y      = 0;
  rect.width  = canvas->width;
  rect.height = canvas->height;

  gdk_gc_set_clip_rectangle(canvas->gc, &rect);
  gdk_gc_set_clip_origin(canvas->gc, 0, 0);

  /* gc ˹򥻥å */
  set_color(canvas->gc, 0, 0, 0);

  gdk_draw_rectangle(canvas->back[0], canvas->gc, TRUE, x1, y1, x2, y2);
  gdk_draw_rectangle(canvas->back[1],canvas->gc, TRUE,
		     x1 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X,
		     y1 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X,
		     x2 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X + 1,
		     y2 * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X);
  gdk_draw_rectangle(canvas->back[2], canvas->gc, TRUE,
		     x1 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X,
		     y1 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X,
		     x2 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X + 1,
		     y2 * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X);
}

/**
 * ޥ뤬ץ饤ȤξˤФΥץ饤Ȥֹ֤
 * ʳ -1 ֤
 * @param cavnas оݤDragTestCanvas
 * @param event оݤGdkEventButton
 */
gint daemon_canvas_on_sprite(Canvas* canvas, GdkEventButton  *event) {
  Sprite* sprite;
  gint x, y, width, height;
  gint i;

  for (i=0; i<SPRITEMAX; i++) {
    sprite = daemon_canvas_get_sprite(canvas, i);

    if (sprite == NULL) {
      continue;
    }

    if (daemon_sprite_ismove(sprite) == FALSE) {
      continue;
    }

    if (daemon_sprite_isvisible(sprite) == FALSE) {
      continue;
    }

    x      = daemon_sprite_get_x(sprite);
    y      = daemon_sprite_get_y(sprite);
    width  = daemon_sprite_get_width(sprite);
    height = daemon_sprite_get_height(sprite);

    if (x <= event->x && event->x <= x + width &&
	y <= event->y && event->y <= y + height) {
      /* ץ饤Ⱦ˥ޥ뤬 */
      return i;
    }
  }

  return -1;
}

/**
 * ץ饤Ȥ׾ˤ TREU ֤ʳ FALSE ֤
 */
gboolean daemon_canvas_sprite_on_board(Canvas *canvas, Sprite *sprite) {
  GdkRectangle board;
  GdkPoint point;
  
  point.x = daemon_sprite_get_x(sprite);
  point.y = daemon_sprite_get_y(sprite);
  point.x = daemon_canvas_rate_xy(canvas, point.x);
  point.y = daemon_canvas_rate_xy(canvas, point.y);

  board.x = 140;
  board.y = 11;
  board.width  = 360;
  board.height = 360;

  return daemon_on_rectangle(&board, &point);
}

/**
 * ֥륯åå
 * @param cavnas оݤDragTestCanvas
 * @return ֥륯åʤ TRUE֤
 */
gboolean daemon_canvas_is_doubleclick(Canvas* canvas) {
  struct timeval  tv;
  struct timezone tz;

  gettimeofday(&tv, &tz);

  if (check_timeval(&(canvas->tv), &tv) == TRUE) {
    canvas->tv.tv_sec  = 0;
    canvas->tv.tv_usec = 0;
    return TRUE;
  }

  gettimeofday(&(canvas->tv), &tz);
  canvas->tv.tv_sec  = tv.tv_sec;
  canvas->tv.tv_usec = tv.tv_usec;

  return FALSE;
}

/** ɥå˻ȤѤ pixmap  */
void daemon_canvas_create_tmp_pixmap(Canvas* canvas) {
  Sprite *sprite;
  GdkRectangle rect;
  GdkPixmap *pixmap;
  GdkGC *gc;
  gint no;

  g_assert(canvas != NULL);
  g_assert(daemon_canvas_isdrag(canvas) == 1);

  canvas->tmp_pixmap = gdk_pixmap_new(canvas->window->window, 
				      canvas->width,
				      canvas->height,
				      -1);
  if (canvas->tmp_pixmap == NULL) {
    daemon_abort("In daemon_canvas_create_tmp_pixmap(). tmp_pixmap is NULL.\n");
  }

  pixmap = daemon_canvas_get_pixmap(canvas);
  gc     = daemon_canvas_get_gc(canvas);
  
  no = daemon_canvas_get_dragno(canvas);
  sprite = daemon_canvas_get_sprite(canvas, no);
  rect.x      = daemon_sprite_get_x(sprite);
  rect.y      = daemon_sprite_get_y(sprite);
  rect.width  = daemon_sprite_get_width(sprite);
  rect.height = daemon_sprite_get_height(sprite);

  daemon_sprite_set_visible(sprite, FALSE);
  /* طʤ */
  daemon_canvas_draw_back_update(canvas, &rect);
  /*  */
  daemon_canvas_draw_sprite_update_all(canvas, &rect);
  daemon_sprite_set_visible(sprite, TRUE);

  rect.x = 0;
  rect.y = 0;
  rect.width  = canvas->width;
  rect.height = canvas->height;
  gdk_gc_set_clip_rectangle(gc, &rect);
  gdk_draw_pixmap(canvas->tmp_pixmap, gc, pixmap, 0, 0, 0, 0,
		  canvas->width, canvas->height);

  daemon_canvas_draw_sprite(canvas, no);
}

GtkWidget *daemon_create_fileselection(void) {
  GtkWidget *fileselection;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *vbox;
  GtkWidget *optionmenu;
  GtkWidget *optionmenu_menu;
  GtkWidget *glade_menuitem;

  fileselection = gtk_file_selection_new (_("load book file"));
  gtk_object_set_data (GTK_OBJECT (fileselection), "fileselection", fileselection);
  gtk_container_set_border_width (GTK_CONTAINER (fileselection), 10);
  gtk_window_set_modal (GTK_WINDOW (fileselection), TRUE);

  ok_button = GTK_FILE_SELECTION (fileselection)->ok_button;
  gtk_object_set_data (GTK_OBJECT (fileselection), "ok_button", ok_button);
  gtk_widget_show (ok_button);
  GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT);

  cancel_button = GTK_FILE_SELECTION (fileselection)->cancel_button;
  gtk_object_set_data (GTK_OBJECT (fileselection), "cancel_button", cancel_button);
  gtk_widget_show (cancel_button);
  GTK_WIDGET_SET_FLAGS (cancel_button, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (ok_button), "pressed",
                      GTK_SIGNAL_FUNC (daemon_kiffile_load_ok_button_pressed),
                      fileselection);
  gtk_signal_connect (GTK_OBJECT (cancel_button), "pressed",
                      GTK_SIGNAL_FUNC (daemon_kiffile_load_cancel_button_pressed),
                      fileselection);


  /* ľѥå󥰥ܥå */
  vbox = GTK_FILE_SELECTION (fileselection)->main_vbox;

  /* ץ˥塼 */
  optionmenu = gtk_option_menu_new ();
  gtk_widget_ref (optionmenu);
  gtk_object_set_data_full (GTK_OBJECT (fileselection), "optionmenu_load", optionmenu,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (optionmenu);
  gtk_box_pack_start (GTK_BOX (vbox), optionmenu, FALSE, FALSE, 0);
  optionmenu_menu = gtk_menu_new ();
  glade_menuitem = gtk_menu_item_new_with_label (_("AUTO"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_auto");
  glade_menuitem = gtk_menu_item_new_with_label (_("CSA format(SJIS)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_csa_sjis");
  glade_menuitem = gtk_menu_item_new_with_label (_("CSA format(EUC)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_csa_euc");
  glade_menuitem = gtk_menu_item_new_with_label (_("KIF format(SJIS)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_kif_sjis");
  glade_menuitem = gtk_menu_item_new_with_label (_("KIF format(EUC)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_kif_euc");
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu), optionmenu_menu);

  return fileselection;
}

GtkWidget *daemon_create_fileselection_save(void) {
  GtkWidget *fileselection;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *vbox;
  GtkWidget *optionmenu;
  GtkWidget *optionmenu_menu;
  GtkWidget *glade_menuitem;

  fileselection = gtk_file_selection_new (_("save book file"));
  gtk_object_set_data (GTK_OBJECT (fileselection), "fileselection", fileselection);
  gtk_container_set_border_width (GTK_CONTAINER (fileselection), 10);
  gtk_window_set_modal (GTK_WINDOW (fileselection), TRUE);

  ok_button = GTK_FILE_SELECTION (fileselection)->ok_button;
   
  gtk_object_set_data (GTK_OBJECT (fileselection), "ok_button", ok_button);
  gtk_widget_show (ok_button);
  GTK_WIDGET_SET_FLAGS (ok_button, GTK_CAN_DEFAULT);

  cancel_button = GTK_FILE_SELECTION (fileselection)->cancel_button;
  gtk_object_set_data (GTK_OBJECT (fileselection), "cancel_button", cancel_button);
  gtk_widget_show (cancel_button);
  GTK_WIDGET_SET_FLAGS (cancel_button, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (ok_button), "pressed",
                      GTK_SIGNAL_FUNC (daemon_kiffile_save_ok_button_pressed),
                      fileselection);
  gtk_signal_connect (GTK_OBJECT (cancel_button), "pressed",
                      GTK_SIGNAL_FUNC (daemon_kiffile_save_cancel_button_pressed),
                      fileselection);


  /* ľѥå󥰥ܥå */
  vbox = GTK_FILE_SELECTION (fileselection)->main_vbox;

  /* ץ˥塼 */
  optionmenu = gtk_option_menu_new ();
  gtk_widget_ref (optionmenu);
  gtk_object_set_data_full (GTK_OBJECT (fileselection), "optionmenu_save", optionmenu,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_set_name(GTK_WIDGET(optionmenu), "optionmenu_save");
  gtk_widget_show (optionmenu);
  GTK_WIDGET_SET_FLAGS (optionmenu, GTK_CAN_DEFAULT);

  gtk_box_pack_start (GTK_BOX (vbox), optionmenu, FALSE, FALSE, 0);
  optionmenu_menu = gtk_menu_new ();
  glade_menuitem = gtk_menu_item_new_with_label (_("CSA format(SJIS)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_csa_sjis");
  glade_menuitem = gtk_menu_item_new_with_label (_("CSA format(EUC)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_csa_euc");
  glade_menuitem = gtk_menu_item_new_with_label (_("KIF format(SJIS)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_kif_sjis");
  glade_menuitem = gtk_menu_item_new_with_label (_("KIF format(EUC)"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "menu_item_kif_euc");
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu), optionmenu_menu);

  return fileselection;
}

void daemon_messagebox(char *s) {
  GtkWidget *messagebox;

  messagebox = daemon_create_messagebox(s);

  gnome_dialog_run_and_close(GNOME_DIALOG(messagebox));
}

GtkWidget *daemon_create_messagebox(char *s) {
  GtkWidget *messagebox;
  GtkWidget *dialog_vbox;
  GtkWidget *button;
  GtkWidget *dialog_action_area1;

  /* We create it with an OK button, and then remove the button, to work
     around a bug in gnome-libs. */
  messagebox = gnome_message_box_new (s,
                              GNOME_MESSAGE_BOX_INFO,
                              GNOME_STOCK_BUTTON_OK, NULL);
  gtk_container_remove (GTK_CONTAINER (GNOME_DIALOG (messagebox)->action_area), GNOME_DIALOG (messagebox)->buttons->data);
  GNOME_DIALOG (messagebox)->buttons = NULL;
  gtk_object_set_data (GTK_OBJECT (messagebox), "messagebox", messagebox);
  gtk_window_set_title (GTK_WINDOW (messagebox), _("information"));
  gtk_window_set_modal (GTK_WINDOW (messagebox), TRUE);
  gtk_window_set_policy (GTK_WINDOW (messagebox), FALSE, FALSE, FALSE);

  dialog_vbox = GNOME_DIALOG (messagebox)->vbox;
  gtk_object_set_data (GTK_OBJECT (messagebox), "dialog_vbox", dialog_vbox);
  gtk_widget_show (dialog_vbox);

  gnome_dialog_append_button (GNOME_DIALOG (messagebox), GNOME_STOCK_BUTTON_OK);
  button = GTK_WIDGET (g_list_last (GNOME_DIALOG (messagebox)->buttons)->data);
  gtk_widget_ref (button);
  gtk_object_set_data_full (GTK_OBJECT (messagebox), "button", button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  dialog_action_area1 = GNOME_DIALOG (messagebox)->action_area;
  gtk_widget_ref (dialog_action_area1);
  gtk_object_set_data_full (GTK_OBJECT (messagebox), "dialog_action_area1", dialog_action_area1,
                            (GtkDestroyNotify) gtk_widget_unref);

  return messagebox;
}

gint daemon_canvas_rate_xy(Canvas* canvas, gint value) {
  if (canvas->size == D_CANVAS_SIZE_SMALL) {
  } else if (canvas->size == D_CANVAS_SIZE_MIDDLE) {
    value = value * D_CANCAS_SIZE_MIDDLE_X / D_CANCAS_SIZE_SMALL_X;
  } else if (canvas->size == D_CANVAS_SIZE_BIG) {
    value = value * D_CANCAS_SIZE_BIG_X / D_CANCAS_SIZE_SMALL_X;
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }

  return value;
}

gint daemon_canvas_rerate_xy(Canvas* canvas, gint value) {
  if (canvas->size == D_CANVAS_SIZE_SMALL) {
  } else if (canvas->size == D_CANVAS_SIZE_MIDDLE) {
    value = value * D_CANCAS_SIZE_SMALL_X / D_CANCAS_SIZE_MIDDLE_X;
  } else if (canvas->size == D_CANVAS_SIZE_BIG) {
    value = value * D_CANCAS_SIZE_SMALL_X / D_CANCAS_SIZE_BIG_X;
  } else {
    /* ;夢ꤨʤʬ */
    g_assert(1);
  }

  return value;
}

GtkWidget* daemon_create_dialog_game (void) {
  GtkWidget *dialog_game;
  GtkWidget *dialog_vbox3;
  GtkWidget *vbox5;
  GtkWidget *table1;
  GtkWidget *label1;
  GtkWidget *label2;
  GtkWidget *optionmenu_sente;
  GtkWidget *optionmenu_sente_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *optionmenu_gote;
  GtkWidget *optionmenu_gote_menu;
  GtkWidget *entry_sente;
  GtkWidget *entry_gote;
  GtkWidget *frame1;
  GtkWidget *optionmenu_teai;
  GtkWidget *optionmenu_teai_menu;
  GtkWidget *dialog_action_area3;
  GtkWidget *button_game_ok;
  GtkWidget *button_game_cancel;

  dialog_game = gnome_dialog_new (NULL, NULL);
  gtk_object_set_data (GTK_OBJECT (dialog_game), "dialog_game", dialog_game);
  gtk_window_set_modal (GTK_WINDOW (dialog_game), TRUE);
  gtk_window_set_policy (GTK_WINDOW (dialog_game), FALSE, FALSE, FALSE);

  dialog_vbox3 = GNOME_DIALOG (dialog_game)->vbox;
  gtk_object_set_data (GTK_OBJECT (dialog_game), "dialog_vbox3", dialog_vbox3);
  gtk_widget_show (dialog_vbox3);

  vbox5 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox5);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "vbox5", vbox5,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox5);
  gtk_box_pack_start (GTK_BOX (dialog_vbox3), vbox5, TRUE, TRUE, 0);

  table1 = gtk_table_new (3, 2, FALSE);
  gtk_widget_ref (table1);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "table1", table1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (table1);
  gtk_box_pack_start (GTK_BOX (vbox5), table1, TRUE, TRUE, 0);

  label1 = gtk_label_new (_("SENTE(WHITE)"));
  gtk_widget_ref (label1);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "label1", label1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label1);
  gtk_table_attach (GTK_TABLE (table1), label1, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label1), 0, 0.5);

  label2 = gtk_label_new (_("GOTE(BLACK)"));
  gtk_widget_ref (label2);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "label2", label2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label2);
  gtk_table_attach (GTK_TABLE (table1), label2, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label2), 0, 0.5);

  optionmenu_sente = gtk_option_menu_new ();
  gtk_widget_ref (optionmenu_sente);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "optionmenu_sente", optionmenu_sente,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (optionmenu_sente);
  gtk_table_attach (GTK_TABLE (table1), optionmenu_sente, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  optionmenu_sente_menu = gtk_menu_new ();
  glade_menuitem = gtk_menu_item_new_with_label (_("HUMAN"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "HUMAN");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL1"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL1");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL2"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL2");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL3"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL3");
  glade_menuitem = gtk_menu_item_new_with_label (_("SERIALPORT1"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "SERIALPORT1");
  glade_menuitem = gtk_menu_item_new_with_label (_("SERIALPORT2"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_sente_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "SERIALPORT2");
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu_sente), optionmenu_sente_menu);

  optionmenu_gote = gtk_option_menu_new ();
  gtk_widget_ref (optionmenu_gote);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "optionmenu_gote", optionmenu_gote,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (optionmenu_gote);
  gtk_table_attach (GTK_TABLE (table1), optionmenu_gote, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  optionmenu_gote_menu = gtk_menu_new ();
  glade_menuitem = gtk_menu_item_new_with_label (_("HUMAN"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "HUMAN");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL1"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL1");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL2"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL2");
  glade_menuitem = gtk_menu_item_new_with_label (_("LEVEL3"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "LEVEL3");
  glade_menuitem = gtk_menu_item_new_with_label (_("SERIALPORT1"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "SERIALPORT1");
  glade_menuitem = gtk_menu_item_new_with_label (_("SERIALPORT2"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_gote_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "SERIALPORT2");
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu_gote), optionmenu_gote_menu);
  gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu_gote), 1);

  entry_sente = gtk_entry_new_with_max_length (32);
  gtk_widget_ref (entry_sente);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "entry_sente", entry_sente,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (entry_sente);
  gtk_table_attach (GTK_TABLE (table1), entry_sente, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  entry_gote = gtk_entry_new_with_max_length (32);
  gtk_widget_ref (entry_gote);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "entry_gote", entry_gote,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (entry_gote);
  gtk_table_attach (GTK_TABLE (table1), entry_gote, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  frame1 = gtk_frame_new (_("HANDICAP"));
  gtk_widget_ref (frame1);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "frame1", frame1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame1);
  gtk_box_pack_start (GTK_BOX (vbox5), frame1, TRUE, TRUE, 0);

  optionmenu_teai = gtk_option_menu_new ();
  gtk_widget_ref (optionmenu_teai);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "optionmenu_teai", optionmenu_teai,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (optionmenu_teai);
  gtk_container_add (GTK_CONTAINER (frame1), optionmenu_teai);
  optionmenu_teai_menu = gtk_menu_new ();
  glade_menuitem = gtk_menu_item_new_with_label (_("HIRATE"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_teai_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "HIRATE");
  /*
  glade_menuitem = gtk_menu_item_new_with_label (_("NIMAIOTI"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_teai_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "NIMAIOTI");
  glade_menuitem = gtk_menu_item_new_with_label (_("YONMAIOTI"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_teai_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "YONMAIOTI");
  */
  glade_menuitem = gtk_menu_item_new_with_label (_("ETC"));
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (optionmenu_teai_menu), glade_menuitem);
  gtk_widget_set_name(GTK_WIDGET(glade_menuitem), "ETC");
  gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu_teai), optionmenu_teai_menu);

  dialog_action_area3 = GNOME_DIALOG (dialog_game)->action_area;
  gtk_object_set_data (GTK_OBJECT (dialog_game), "dialog_action_area3", dialog_action_area3);
  gtk_widget_show (dialog_action_area3);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area3), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (dialog_action_area3), 8);

  gnome_dialog_append_button (GNOME_DIALOG (dialog_game), GNOME_STOCK_BUTTON_OK);
  button_game_ok = GTK_WIDGET (g_list_last (GNOME_DIALOG (dialog_game)->buttons)->data);
  gtk_widget_ref (button_game_ok);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "button_game_ok", button_game_ok,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (button_game_ok);
  GTK_WIDGET_SET_FLAGS (button_game_ok, GTK_CAN_DEFAULT);

  gnome_dialog_append_button (GNOME_DIALOG (dialog_game), GNOME_STOCK_BUTTON_CANCEL);
  button_game_cancel = GTK_WIDGET (g_list_last (GNOME_DIALOG (dialog_game)->buttons)->data);
  gtk_widget_ref (button_game_cancel);
  gtk_object_set_data_full (GTK_OBJECT (dialog_game), "button_game_cancel", button_game_cancel,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (button_game_cancel);
  GTK_WIDGET_SET_FLAGS (button_game_cancel, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (button_game_ok), "pressed",
                      GTK_SIGNAL_FUNC (daemon_canvas_new_game_ok),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (button_game_cancel), "pressed",
                      GTK_SIGNAL_FUNC (daemon_canvas_new_game_cancel),
                      NULL);

  return dialog_game;
}

/** жɤΥᥤ롼 */
void daemon_canvas_game(Canvas *canvas) {
  BOARD *board;
  GAME *game;
  int minmax_return;
  char send_buf[BUFSIZ];
  TE te;
  DTe dte;
  INPUTSTATUS (*si_input_next[2])(BOARD *bo, TE *te);
  time_t t;
  extern int THINK_DEEP_MAX;
  extern BOARD g_board;
  extern GAME g_game;
  extern void *si_input_next_gui_human;

  g_assert(canvas != NULL);
  g_assert(daemon_canvas_get_mode(canvas) == D_CANVAS_MODE_GAME);

  si_input_next_gui_human = daemon_input_next_gui_human_impl;

  board = &g_board;
  game = &g_game;

  game->player_number[SENTE] = canvas->playertype[D_SENTE];
  game->player_number[GOTE]  = canvas->playertype[D_GOTE];

  if (canvas->record == NULL) {
    canvas->record = daemon_record_new();
  }
  daemon_dboard_copy(&(canvas->board), &(canvas->record->first_board));

  if (canvas->bookwindow != NULL) {
    gtk_widget_destroy(canvas->bookwindow);
  }

  if (canvas->flg_bookwindow == TRUE) {
    canvas->bookwindow = daemon_canvas_create_bookwindow(canvas);
    gtk_widget_show(canvas->bookwindow);
  }

  boInit(board);
  daemon_dboard_copy_board(&(canvas->board), board);
  boPlayInit(board);

  gmSetInputFunc(game, SENTE, game->player_number[SENTE]);
  gmSetInputFunc(game, GOTE, game->player_number[GOTE]);
  si_input_next[SENTE] = gmGetInputNextPlayer(game, SENTE);
  si_input_next[GOTE]  = gmGetInputNextPlayer(game, GOTE);

  canvas->record->time[D_SENTE] = 0;
  canvas->record->time[D_GOTE] = 0;

#ifdef USE_SERIAL
  if ( gmGetInputNextPlayer(game, SENTE ) == si_input_next_serialport ) {
    if (canvas->playertype[SENTE] == SI_SERIALPORT_1) {
      gmSetDeviceName(&g_game, SENTE, "/dev/cua0");
    } else {
      gmSetDeviceName(&g_game, SENTE, "/dev/cua1");
    }
    caStart(&(game->ca[SENTE]));
  }
  if ( gmGetInputNextPlayer(game, GOTE ) == si_input_next_serialport ) {
    if (canvas->playertype[GOTE] == SI_SERIALPORT_1) {
      gmSetDeviceName(&g_game, GOTE, "/dev/cua0");
    } else {
      gmSetDeviceName(&g_game, GOTE, "/dev/cua1");
    }
    caStart(&(game->ca[GOTE]));
  }
#endif /* USE_SERIAL */

  if (gmGetInputNextPlayer(game, SENTE ) != si_input_next_gui_human &&
      gmGetInputNextPlayer(game, GOTE ) != si_input_next_gui_human) {
    GtkWidget *widget;
    widget = lookup_widget(canvas->window, "give_up");
    gtk_widget_set_sensitive(widget, FALSE);
    widget = lookup_widget(canvas->window, "button_give_up");
    gtk_widget_set_sensitive(widget, FALSE);
  }

  gmSetGameStatus(game, 0);

#ifdef USE_HASH  
  /* ϥåơ֥ */
  newHASH();
#endif /* USE_HASH */

  while ( 1 ) {
    THINK_DEEP_MAX = gmGetComputerLevel(game, board->next);
    /* ߤλ֤Ͽ */
    boSetNowTime(board);
    minmax_return = (*si_input_next[board->next])(board, &te);
    /* ֤Ͽ */
    boSetUsedTime(board);

#ifdef USE_SERIAL
    if ( minmax_return == SI_NORMAL &&
        si_input_next[(board->next == 0)] == si_input_next_serialport ) {
      /* μ֤ꥢݡȤξϼ */
      getTE2SerialString(send_buf, board, &te);
      caSend(&(game->ca[(board->next == 0)]), send_buf);
    }
#endif /* USE_SERIAL */

    if (te.fm == 0 && te.to == 0) {
      break;
    }

    if (minmax_return != SI_NORMAL) {
      break;
    }

    play_move( board, &te );

    dte.fm   = te.fm;
    dte.to   = te.to;
    dte.nari = te.nari;
    dte.uti  = te.uti;

    t = board->used_time[board->mi.count - 1];
    
    daemon_record_add_dte(canvas->record, &dte, t);
    daemon_canvas_bookwindow_add_dte(canvas, &dte);
    daemon_canvas_update_statusbar_te(canvas, &dte);
    daemon_dboard_move(&(canvas->board), &dte);

    /* ץ饤Ⱥ */
    daemon_canvas_dispose_sprite(canvas);

    /* ̺ */
    daemon_canvas_draw_back(canvas);
    daemon_canvas_draw_sprite_all(canvas);
    daemon_canvas_draw_time(canvas);
    daemon_canvas_draw_all(canvas);
  }

  /* ơȽ */
  if (minmax_return == SI_TORYO) {
    if (board->next == SENTE) {
      daemon_messagebox(_("Gote(white) win."));
    } else {
      daemon_messagebox(_("Sente(black) win."));
    }
  } else if ( minmax_return == SI_WIN ) {
    if (board->next == SENTE) {
      daemon_messagebox(_("Sente(black) win."));
    } else {
      daemon_messagebox(_("Gote(white) win."));
    }
  } else if ( minmax_return == SI_CHUDAN ) {
    daemon_messagebox(_("Stoped."));
  } else if ( minmax_return == SI_CHUDAN_RECV ) {
    daemon_messagebox(_("Stoped."));
  } else if ( minmax_return == SI_NGDATA ) {
    daemon_messagebox(_("No good data."));
  } else {
      /* ;夢ꤨʤʬ */
    g_assert( 1 );
  }

#ifdef USE_HASH  
  /* ϥåơ֥ */
  freeHASH();
#endif /* USE_HASH */

  daemon_dboard_copy(&(canvas->record->first_board), &(canvas->board));
  canvas->count = 0;
  daemon_canvas_set_mode(canvas, D_CANVAS_MODE_BOOK);
  daemon_canvas_kif_last(canvas);
  daemon_canvas_change_mode(canvas, D_CANVAS_MODE_BOOK);
}

/**
 * ʹ֤GUIΰ te ˳Ǽ롣
 * @param bo оݤBOARD
 * @param te TE
 * @return GAMESTAT
 */
int daemon_input_next_gui_human_impl(BOARD *bo, TE *te) {
  TE tmp_te;
  TE *gui_te;
  extern GAME g_game;
  extern Canvas *g_canvas;

  gui_te = &(g_canvas->gui_te);

  teSetTE(&tmp_te, 0 , 0, 0, 0);
  teSetTE(gui_te, 0 , 0, 0, 0);
  
  while ( teCmpTE(&tmp_te, gui_te) == 0 ) {
    if ( gmGetGameStatus(&g_game) != 0 ) {
      /* Ǥޤϡλä */
      return gmGetGameStatus(&g_game);
    }
    processing_pending();
    usleep( 50 );
  }
  
  *te = *gui_te;

  return SI_NORMAL;
}

/** μο */
gint count_piece(DBoard *board, gint next) {
  gint count;
  gint i;

  count = 0;
  for (i=1; i <= 7; i++) {
    if (0 < daemon_dboard_get_piece(board, next, i)) {
      count++;
    }
  }

  return count;
}

/** Ȣζμο */
gint count_pbox(DBoard *board) {
  gint count;
  gint i;

  count = 0;
  for (i=1; i <= 8; i++) {
    if (0 < daemon_dboard_get_pbox(board, i)) {
      count++;
    }
  }

  return count;
}

/**
 * 襦ɥɽʸ롣
 */
void create_te_string(DBoard *board, int no, DTe *te, char *buf) {
  char *x[10] = {
    "",     _("1_"), _("2_"), _("3_"), _("4_"),
    _("5_"),_("6_"), _("7_"), _("8_"), _("9_"),
  };
  char *y[10] = {
    "",     _("1 "), _("2 "), _("3 "), _("4 "),
    _("5 "),_("6 "), _("7 "), _("8 "), _("9 "),
  };
  char *piece[16] = {
    "",           _("FU"), _("KYO"),     _("KEI"),
    _("GIN"),     _("KIN"),_("KAKU"),    _("HISYA"),
    _("OH"),      _("TO"), _("NARIKYO"), _("NARIKEI"),
    _("NARIGIN"), "",      _("UMA"),     _("RYU"),
  };
  char *nari = _(" NARI");
  char *uti = _(" UTI");
  char *game_start = _("game start");
  int p;

  if (no == 0) {
    sprintf(buf, "%3d : %s", no, game_start);
  } else if (te->uti == 0) {
    p = daemon_dboard_get_board(board, te->fm & 0xF, te->fm >> 4);
    p &= 0xF;
    sprintf(buf, "%3d : %s%s%s%s(%d,%d)",
	    no,
	    x[te->to & 0xF],
	    y[te->to >> 4],
	    piece[p],
	    te->nari ? nari : "",
	    te->fm & 0xF,
	    te->fm >> 4
	    );
  } else {
    sprintf(buf, "%3d : %s%s%s%s", no,
	    x[te->to & 0xF],
	    y[te->to >> 4],
	    piece[te->uti],
	    uti);
  }
}

void create_te_string2(DBoard *board, int no, DTe *te, char *buf) {
  char *x[10] = {
    "",     _("1_"), _("2_"), _("3_"), _("4_"),
    _("5_"),_("6_"), _("7_"), _("8_"), _("9_"),
  };
  char *y[10] = {
    "",     _("1 "), _("2 "), _("3 "), _("4 "),
    _("5 "),_("6 "), _("7 "), _("8 "), _("9 "),
  };
  char *piece[16] = {
    "",           _("FU"), _("KYO"),     _("KEI"),
    _("GIN"),     _("KIN"),_("KAKU"),    _("HISYA"),
    _("OH"),      _("TO"), _("NARIKYO"), _("NARIKEI"),
    _("NARIGIN"), "",      _("UMA"),     _("RYU"),
  };
  char *nari = _(" NARI");
  char *uti = _(" UTI");
  char *game_start = _("game start");
  char *teme = _("teme");
  int p;
  DBoard *tmp_board;

  if (no == 0) {
    sprintf(buf, "%d %s %s",
	    no,
	    teme,
	    game_start);
  } else if (te->uti == 0) {
    tmp_board = daemon_dboard_new();
    daemon_dboard_copy(board, tmp_board);
    daemon_dboard_back(tmp_board);

    p = daemon_dboard_get_board(tmp_board, te->fm & 0xF, te->fm >> 4);
    p &= 0xF;
    sprintf(buf, "%d %s %s%s%s%s(%d,%d)",
	    no,
	    teme,
	    x[te->to & 0xF],
	    y[te->to >> 4],
	    piece[p],
	    te->nari ? nari : "",
	    te->fm & 0xF,
	    te->fm >> 4
	    );
  } else {
    sprintf(buf, "%d %s %s%s%s%s",
	    no,
	    teme, 
	    x[te->to & 0xF],
	    y[te->to >> 4],
	    piece[te->uti],
	    uti
	    );
  }
}

