/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * Canvas class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/canvas.h,v $
 * $Id: canvas.h,v 1.1.1.1 2005/12/09 09:03:03 tokita Exp $
 */

#ifndef _CANVAS_H_
#define _CANVAS_H_

#include <gnome.h>
#include <sys/time.h>
#include "dte.h"
#include "dmoveinfo.h"
#include "dboard.h"
#include "sprite.h"
#include "record.h"

#include "si/si.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


typedef enum {
  /** Ե⡼ */
  D_CANVAS_MODE_WAIT,
  /** ⡼ */
  D_CANVAS_MODE_BOOK,
  /** Խ⡼ */
  D_CANVAS_MODE_EDIT,
  /** жɥ⡼ */
  D_CANVAS_MODE_GAME,
  /** ;⡼ */
  D_CANVAS_MODE_MATE,
} D_CANVAS_MODE;

/* canvas 򥵥ɽ */
typedef enum {
  /* ǾΥ */
  D_CANVAS_SIZE_SMALL  = 1,
  /* Υ */
  D_CANVAS_SIZE_MIDDLE = 2,
  /* Υ */
  D_CANVAS_SIZE_BIG    = 3,
} D_CANVAS_SIZE;

typedef enum {
  D_CANVAS_TOOLBAR_NONE,
  D_CANVAS_TOOLBAR_ICON,
  D_CANVAS_TOOLBAR_TEXT,
  D_CANVAS_TOOLBAR_BOTH,
} D_CANVAS_TOOLBAR;

typedef enum {
  /** ꥢʤ */
  D_CANVAS_AREA_NONE,
  /**  */
  D_CANVAS_AREA_LEFT_KOMADAI,
  /**  */
  D_CANVAS_AREA_RIGHT_KOMADAI,
  /** Ȣ */
  D_CANVAS_AREA_KOMABAKO,
  /**  */
  D_CANVAS_AREA_BOARD,
} D_CANVAS_AREA;

#define D_SPRIRE_KOMA_WIDTH  40
#define D_SPRIRE_KOMA_HEIGHT 40
#define D_CANCAS_SIZE_SMALL_X  640
#define D_CANCAS_SIZE_SMALL_Y  380
#define D_CANCAS_SIZE_MIDDLE_X 800
#define D_CANCAS_SIZE_MIDDLE_Y 475
#define D_CANCAS_SIZE_BIG_X    1024
#define D_CANCAS_SIZE_BIG_Y    608

#define SPRITEMAX 41

typedef struct {
  /** window */
  GtkWidget* window;
  /** canvasβΥɥåȿ */
  gint width;
  /** canvasνĤΥɥåȿ */
  gint height;
  /** 褹ɥ󥰥ꥢ */
  GtkWidget *drawingarea;
  /** ֥Хåե󥰤Τpixmap */
  GdkPixmap *pixmap[3];
  /** ɥå˻Ȥ pixmap */
  GdkPixmap *tmp_pixmap;
  /** طʤpixmap */
  GdkPixmap *back[3];
  /** β */
  GdkPixmap *koma[3];
  /** Υޥ */
  GdkBitmap *mask[3];
  /** طʤθβ */
  GdkImlibImage *back_src;
  /** θβ */
  GdkImlibImage *koma_src;
  /** pixmapgc */
  GdkGC *gc;
  /** SpriteChar */
  Sprite *sprite[SPRITEMAX];
  /** 󥯥å */
  struct timeval tv;
  /** ɥå椫Υե饰 */
  gboolean drag;
  /** ɥåΥץ饤ɤֹ */
  gint drag_no;
  /** ץ饤Ȥȥޥκɸκʬ */
  GdkPoint diff;
  /** ̾ */
  DBoard board;
  /** ⡼ */
  D_CANVAS_MODE mode;
  /** canvas Υɽ */
  D_CANVAS_SIZE size;
  /** 襦ɥo */
  GtkWidget *bookwindow;
  /** 襦ɥɽѤߤΥե饰 */
  gboolean flg_bookwindow;

  /** record */
  Record *record;
  /** ⡼ɤǻȤΥ */
  gint count;
  /** ɽ¦μ */
  D_TEBAN front;

  /** ͤᾭ */
  GtkWidget *dialog_mate;
  /** ׹椫Υե饰 */
  gboolean flg_thinking;
  /** жɥ */
  GtkWidget *dialog_game;

  /** 礤 */
  D_TEAI teai;
  /** жɼ */
  PLAYERTYPE playertype[2];
  /** GUIϤ */
  TE gui_te;
  /** status bar */
  gboolean statusbar;
} Canvas;


Canvas* daemon_canvas_new           (GtkWidget *window);
void    daemon_canvas_init          (Canvas *canvas, GtkWidget *window);
void    daemon_canvas_free          (Canvas *canvas);
void    daemon_canvas_init_sprite   (Canvas *canvas);
void    daemon_canvas_dispose_sprite(Canvas *canvas);

/* set_* , get_* */
GtkWidget*    daemon_canvas_get_window    (Canvas *canvas);
GdkGC*        daemon_canvas_get_gc        (Canvas *canvas);
GdkPixmap*    daemon_canvas_get_pixmap    (Canvas *canvas);
GdkPixmap*    daemon_canvas_get_back      (Canvas *canvas);
GtkWidget*    daemon_canvas_get_drawinarea(Canvas *canvas);
gboolean      daemon_canvas_isdrag        (Canvas *canvas);
void          daemon_canvas_set_drag      (Canvas *canvas,
					   gboolean drag);
gint          daemon_canvas_get_dragno    (Canvas *canvas);
void          daemon_canvas_set_dragno    (Canvas *canvas,
					   gint no);
GdkPoint      daemon_canvas_get_diff      (Canvas *canvas);
void          daemon_canvas_set_diff      (Canvas *canvas,
					   gint x,
					   gint y);
D_CANVAS_MODE daemon_canvas_get_mode      (Canvas *canvas);
void          daemon_canvas_set_mode      (Canvas *canvas,
					   D_CANVAS_MODE mode);
void          daemon_canvas_set_size      (Canvas* canvas,
					   D_CANVAS_SIZE size);
Sprite*       daemon_canvas_get_sprite    (Canvas* canvas, gint no);
void          daemon_canvas_set_front     (Canvas *canvas, D_TEBAN next);
D_TEBAN       daemon_canvas_get_front     (Canvas *canvas);

/* event */
void       daemon_canvas_expose                     (Canvas *canvas,
						     GtkWidget       *widget, 
						     GdkEventExpose  *event,
						     gpointer         user_data);
void       daemon_canvas_motion_notify              (Canvas *canvas, GdkEventMotion  *event);
void       daemon_canvas_button_press               (Canvas *canvas, GdkEventButton  *event);
void       daemon_canvas_button_release             (Canvas *canvas, GdkEventButton  *event);
GtkWidget* daemon_canvas_create_bookwindow          (Canvas *canvas);
void       daemon_canvas_bookwindow_add_dte         (Canvas *canvas, DTe *te);
void       daemon_canvas_bookwindow_activate        (Canvas *canvas,
						     GtkMenuItem *menuitem,
						     gpointer user_data);
void       daemon_canvas_bookwindow_click           (GtkWidget       *widget,
						     GdkEventButton  *event,
						     gpointer         user_data);
void       daemon_canvas_bookwindow_destroy         (Canvas *canvas,
						     GtkObject *object,
						     gpointer user_data);
void       daemon_canvas_change_size                (Canvas *canvas, D_CANVAS_SIZE size);
void       daemon_canvas_change_toolbar             (Canvas *canvas, D_CANVAS_TOOLBAR toolbar);
void       daemon_canvas_change_mode                (Canvas *canvas, D_CANVAS_MODE mode);
void       daemon_canvas_change_size_sprite         (Canvas *canvas, D_CANVAS_SIZE size);
void       daemon_canvas_timeout                    (gpointer data);
void       daemon_kiffile_load_ok_button_pressed    (GtkButton *button,
						     gpointer user_data);
void       daemon_kiffile_load_cancel_button_pressed(GtkButton*button,
						     gpointer user_data);
void       daemon_kiffile_save_ok_button_pressed    (GtkButton *button,
						     gpointer user_data);
void       daemon_kiffile_save_cancel_button_pressed(GtkButton*button,
						     gpointer user_data);
void       daemon_canvas_leave_notify               (Canvas *canvas,
						     GtkWidget *widget,
						     GdkEventCrossing *event,
						     gpointer user_data);
void       daemon_canvas_kif_first                  (Canvas *canvas);
void       daemon_canvas_kif_back                   (Canvas *canvas);
void       daemon_canvas_kif_next                   (Canvas *canvas);
void       daemon_canvas_kif_last                   (Canvas *canvas);
void       daemon_canvas_set_kif_sensitive          (Canvas *canvas);
void       daemon_canvas_set_sensitive              (Canvas *canvas);
void       daemon_canvas_flip_activate              (Canvas *canvas);
void       daemon_canvas_mate_activate              (Canvas *canvas);
void       daemon_canvas_mate_start_pressed         (Canvas *canvas);
void       daemon_canvas_mate_cancel_pressed        (Canvas *canvas);
void       daemon_canvas_set_hirate_activate        (Canvas *canvas);
void       daemon_canvas_set_mate_activate          (Canvas *canvas);
void       daemon_canvas_set_rl_reversal_activate   (Canvas *canvas);
void       daemon_canvas_set_order_reversal_activate(Canvas *canvas);
void       daemon_canvas_set_edit_flip_activate     (Canvas *canvas);
void       daemon_canvas_all_koma_to_pbox_activate  (Canvas *canvas);
int        daemon_canvas_ok_cancel_dialog           (Canvas *canvas,
						     char *message);
void       daemon_canvas_ok_cancel_dialog_cb        (gint reply,
						     gpointer data);
int        daemon_canvas_yes_no_dialog              (Canvas *canvas,
						     char *message);
void       daemon_canvas_yes_no_dialog_cb           (gint reply,
						     gpointer data);
void       daemon_canvas_mate_think_cancel          (Canvas *canvas);
void       daemon_canvas_new_game                   (Canvas *canvas);
void       daemon_canvas_new_game_ok                (GtkButton *button,
						     gpointer user_data);
void       daemon_canvas_new_game_cancel            (GtkButton *button,
						     gpointer user_data);
void       daemon_canvas_new_game_player_cb         (GtkButton *button,
						     gpointer user_data);
void       daemon_canvas_new_game_teai_cb           (GtkButton *button,
						     gpointer user_data);
void       daemon_canvas_stop_activate              (Canvas *canvas);
void       daemon_canvas_restart_activate           (Canvas *canvas);
void       daemon_canvas_give_up_activate           (Canvas *canvas);
void       daemon_canvas_update_statusbar           (Canvas *canvas);
void       daemon_canvas_update_statusbar_te        (Canvas *canvas,
						     DTe *te);
void       daemon_canvas_statusbar_activate         (Canvas *canvas);

/* draw */
void daemon_canvas_draw_all                        (Canvas *canvas);
void daemon_canvas_draw_all_update                 (Canvas *canvas,
						    GdkRectangle *rect);
void daemon_canvas_draw_back                       (Canvas *canvas);
void daemon_canvas_draw_back_update                (Canvas *canvas,
						    GdkRectangle *rect);
void daemon_canvas_draw_back_update_from_tmp_pixmap(Canvas *canvas,
						    GdkRectangle *rect);
void daemon_canvas_draw_sprite_all                 (Canvas *canvas);
void daemon_canvas_draw_sprite_update_all          (Canvas *canvas,
						    GdkRectangle *rect);
void daemon_canvas_draw_sprite                     (Canvas *canvas,
						    gint no);
void daemon_canvas_draw_time                       (Canvas *canvas);

/* etc */
void          daemon_canvas_create_back_pixmap(Canvas *canvas);
gint          daemon_canvas_on_sprite         (Canvas *canvas,
					       GdkEventButton *event);
gboolean      daemon_canvas_sprite_on_board   (Canvas *canvas,
					       Sprite *sprite);
gboolean      daemon_canvas_is_doubleclick    (Canvas* canvas);
void          daemon_canvas_create_tmp_pixmap (Canvas* canvas);
GtkWidget*    daemon_create_fileselection     (void);
GtkWidget*    daemon_create_fileselection_save(void);
void          daemon_messagebox               (char *s);
GtkWidget*    daemon_create_messagebox        (char *s);
gint          daemon_canvas_rate_xy           (Canvas* canvas,
					       gint value);
gint          daemon_canvas_rerate_xy         (Canvas* canvas,
					       gint value);
D_CANVAS_AREA daemon_canvas_on_drag_area      (Canvas *canvas,
					       gint x, gint y,
					       gint *dest_x,
					       gint *dest_y);
GtkWidget*    daemon_create_dialog_game       (void);
void          daemon_canvas_game              (Canvas *canvas);
int           daemon_input_next_gui_human_impl(BOARD *bo,
					       TE *te);
gint          count_piece                     (DBoard *board,
					       gint next);
gint          count_pbox                      (DBoard *board);
void          create_te_string                (DBoard *board,
					       int no,
					       DTe *te,
					       char *buf);
void          create_te_string2               (DBoard *board,
					       int no,
					       DTe *te,
					       char *buf);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* _CANVAS_H_ */

