/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * DBoard class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/dboard.c,v $
 * $Id: dboard.c,v 1.1.1.1 2005/12/09 09:03:04 tokita Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "dboard.h"

/** DBoard ƤΥݥ󥿤֤ */
DBoard* daemon_dboard_new(void) {
  DBoard* bo;
  
  bo = (DBoard *)malloc(sizeof(DBoard));
  if (bo == NULL) {
    printf("No enough memory in daemon_dboard_new().");
    abort();
  }
  
  daemon_dboard_init(bo);
  
  return bo;
}

/**  */
void daemon_dboard_init(DBoard* bo) {
  int i, x, y;
  
  for (i=0; i<256; i++) {
    bo->board[i] = DWALL;
  }
  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      bo->board[(y << 4) + x] = 0;
    }
  }
  for (i=0; i<8; i++) {
    bo->piece[D_SENTE][i] = 0;
    bo->piece[D_GOTE][i] = 0;
  }
  bo->next = D_SENTE;
  daemon_dmoveinfo_init(&(bo->mi));
  for (i=0; i<DMOVEINFOMAX; i++) {
    bo->tori[i] = 0;
  }
  bo->count = 0;
  for (i=0; i<9; i++) {
    bo->pbox[i] = 0;
  }
  bo->front = D_SENTE;
}

/** DBoard ƤΥݥ󥿤֤ */
void daemon_dboard_free(DBoard* bo) {
  assert(bo != NULL);
  
  free(bo);
}

/** (x, y)˶ p 򥻥åȤ롣 */
void daemon_dboard_set_board(DBoard* bo, int x, int y, int p) {
  assert(bo != NULL);
  assert(1 <= x && x <= 9);
  assert(1 <= y && y <= 9);
  assert(0 <= p && p <= 0x1F);
  if (bo->front == D_GOTE) {
    if (p != 0) {
      p ^= 0x10;
    }
    bo->board[((10 - y) << 4) + (10 - x)] = p;
  } else {
    bo->board[(y << 4) + x] = p;
  }
}

/** (x, y)ζ֤ */
int daemon_dboard_get_board(DBoard* bo, int x, int y) {
  assert(bo != NULL);
  assert(1 <= x && x <= 9);
  assert(1 <= y && y <= 9);
  if (bo->front == D_GOTE) {
    int p;
    p = bo->board[((10 - y) << 4) + (10 - x)];
    if (p != 0) {
      p ^= 0x10;
    }
    return p;
  } else {
    return bo->board[(y << 4) + x];
  }
}

/** next ֤λ p 򥻥åȤ롣 */
void daemon_dboard_set_piece(DBoard* bo, int next, int p, int n) {
  if (bo->front == D_GOTE) {
    next ^= 1;
  }
  bo->piece[next][p] = n;
}

/** next ֤λ p ץ饹롣 */
void daemon_dboard_add_piece(DBoard* bo, int next, int p) {
  /*
  if (bo->front == D_GOTE) {
    next ^= 1;
  }
  */
  bo->piece[next][p]++;
}

/** next ֤λ p 縺餹 */
void daemon_dboard_dec_piece(DBoard* bo, int next, int p) {
  /*
  if (bo->front == D_GOTE) {
    next ^= 1;
  }
  */
  bo->piece[next][p]--;
}

/** next ֤λ p ֤ */
int daemon_dboard_get_piece(DBoard* bo, int next, int p) {
  if (bo->front == D_GOTE) {
    next ^= 1;
  }
  return bo->piece[next][p];
}

/** ʿ̤򥻥åȤ */
void daemon_dboard_set_hirate(DBoard* bo) {
  /* ʿ֥ѥ */
  const static int hirate[] = {
    0x12, 0x13, 0x14, 0x15, 0x18, 0x15, 0x14, 0x13, 0x12,
    0x00, 0x16, 0x00, 0x00, 0x00, 0x00, 0x00, 0x17, 0x00,
    0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
    0x00, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00,
    0x02, 0x03, 0x04, 0x05, 0x08, 0x05, 0x04, 0x03, 0x02,
  };
  int x, y;
  
  daemon_dboard_init(bo);
  
  for (y = 0; y < 9; y++) {
    for (x = 0; x < 9; x++) {
      daemon_dboard_set_board(bo, x + 1, y + 1, hirate[y * 9 + x]);
    }
  }
}

/** ;Ѥ̤򥻥åȤ */
void daemon_dboard_set_mate(DBoard *board) {
  daemon_dboard_init(board);

  daemon_dboard_set_board(board, 5, 1, 0x18);

  daemon_dboard_set_piece(board, D_GOTE, 1, 18);
  daemon_dboard_set_piece(board, D_GOTE, 2, 4);
  daemon_dboard_set_piece(board, D_GOTE, 3, 4);
  daemon_dboard_set_piece(board, D_GOTE, 4, 4);
  daemon_dboard_set_piece(board, D_GOTE, 5, 4); 
  daemon_dboard_set_piece(board, D_GOTE, 6, 2);
  daemon_dboard_set_piece(board, D_GOTE, 7, 2);

  daemon_dboard_set_pbox(board, 8, 1);
}

void daemon_dboard_set_rl_reversal(DBoard *board) {
  int x, y, p;
  DBoard *tmp;

  tmp = daemon_dboard_new();

  daemon_dboard_copy(board, tmp);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      daemon_dboard_set_board(board, 10 - x, y, p);
    }
  }

  daemon_dboard_free(tmp);
}

void daemon_dboard_set_order_reversal(DBoard *board) {
  int x, y, p;
  DBoard *tmp;

  tmp = daemon_dboard_new();

  daemon_dboard_copy(board, tmp);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      if (p != 0) {
	p ^= 0x10;
      }
      daemon_dboard_set_board(board, x, 10 - y, p);
    }
  }

  daemon_dboard_free(tmp);
}

/** ̤촹 */
void daemon_dboard_flip(DBoard *board) {
  int x, y, p, i;
  DBoard *tmp;

  tmp = daemon_dboard_new();

  daemon_dboard_copy(board, tmp);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(tmp, x, y);
      if (p != 0) {
	p ^= 0x10;
      }
      daemon_dboard_set_board(board, 10 - x, 10 - y, p);
    }
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(tmp, D_SENTE, i);
    daemon_dboard_set_piece(board, D_GOTE, i, p);
    p = daemon_dboard_get_piece(tmp, D_GOTE, i);
    daemon_dboard_set_piece(board, D_SENTE, i, p);
  }

  daemon_dboard_free(tmp);
}

/** ƤζȢذư */
void daemon_dboard_set_all_to_pbox(DBoard *board) {
  int x, y, i;
  int piecemax[9] = {
    0, 18, 4, 4, 4, 4, 2, 2, 2,
  };

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      daemon_dboard_set_board(board, x, y, 0);
    }
  }

  for (i=1; i<=7; i++) {
    daemon_dboard_set_piece(board, D_GOTE, i, 0);
    daemon_dboard_set_piece(board, D_SENTE, i, 0);
  }

  for (i=1; i<=8; i++) {
    daemon_dboard_set_pbox(board, i, piecemax[i]);
  }
}

/** ԽѤ˻ȤäƤʤ pbox ˳Ǽ */
void daemon_dboard_set_for_edit(DBoard *board) {
  int piece[9];
  int x, y, p;
  int i;
  int piecemax[9] = {
    0, 18, 4, 4, 4, 4, 2, 2, 2,
  };
  
  for (i=0; i<9; i++) {
    piece[i] = 0;
  }

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(board, x, y);
      if ((p & 0xF) == 8) {
	p = 8;
      } else {
	p &= 7;
      }
      piece[p]++;
    }
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(board, D_SENTE, i);
    piece[i] += p;
    p = daemon_dboard_get_piece(board, D_GOTE, i);
    piece[i] += p;
  }

  for (i=1; i<=8; i++) {
    daemon_dboard_set_pbox(board, i, piecemax[i] - piece[i]);
  }
}

/** out  bo ƤϤ */
void daemon_dboard_output(DBoard* bo, FILE* out) {
  int x, y, i;
  int p;
  int mode = 1;

  assert(bo != NULL);
  assert(out != NULL);

  for (y = 1; y <= 9; y++) {
    for (x = 9; 0 < x; x--) {
      p = daemon_dboard_get_board(bo, x, y);
      if (p) {
	if (mode)
	  fprintf(out, "%c%X ", (p & 0x10) ? 'v' : '^', p & 0x0F);
	else
	  fprintf(out, "%02X ", p);
      } else {
	if (mode)
	  fprintf(out, " . ");
	else
	  fprintf(out, "00 ");
      }
    }
    fprintf(out, "\n");
  }

  fprintf(out, "%2d ", daemon_dboard_get_piece(bo, D_SENTE, 1));
  for (i=2; i<=7; i++) {
    fprintf(out, "%d ", daemon_dboard_get_piece(bo, D_SENTE, i));
  }
  fprintf(out, "\n");
  fprintf(out, "%2d ", daemon_dboard_get_piece(bo, D_GOTE, 1));
  for (i=2; i<=7; i++) {
    fprintf(out, "%d ", daemon_dboard_get_piece(bo, D_GOTE, i));
  }
  fprintf(out, "\n");
}

/**
 * te Ƥǰʤ롣
 * @param bo оݤDBoard
 * @param te ؤ
 */
void daemon_dboard_move(DBoard *bo, DTe *te) {
  int p;

  assert(bo != NULL);
  assert(te != NULL);
  assert(bo->count < DMOVEINFOMAX);
  if (DMOVEINFOMAX <= bo->count) {
    return;
  }

  daemon_dmoveinfo_add(&(bo->mi), te);

  if (te->uti) {
    /* Ǥ */
    bo->board[te->to] = te->uti + (bo->next == D_GOTE ? 0x10 : 0);
    daemon_dboard_dec_piece(bo, bo->next, te->uti);
  } else {
    /* ư */
    p = bo->board[te->fm];
    bo->board[te->fm] = 0;
    if (bo->board[te->to] != 0) {
      bo->tori[bo->count] = bo->board[te->to];
      daemon_dboard_add_piece(bo, bo->next, bo->tori[bo->count] & 7);
    } else {
      bo->tori[bo->count] = 0;
    }
    if (te->nari) {
      p |= 0x08;
    }
    bo->board[te->to] = p;
  }

  daemon_dboard_toggle_next(bo);
  bo->count++;
}

/**
 * ᤹
 * @param bo оݤDBoard
 */
void daemon_dboard_back(DBoard *bo) {
  DTe *te;
  int p;

  assert(bo != NULL);
  assert(0 < bo->count);

  if (bo->count == 0) {
    return;
  }

  bo->count--;

  daemon_dboard_toggle_next(bo);
  te = daemon_dmoveinfo_pop(&(bo->mi));

  if (te->uti) {
    /* Ǥ */
    bo->board[te->to] = 0;
    daemon_dboard_add_piece(bo, bo->next, te->uti);
  } else {
    /* ư */
    p = bo->board[te->to];
    bo->board[te->to] = bo->tori[bo->count];
    if (bo->tori[bo->count] != 0) {
      daemon_dboard_dec_piece(bo, bo->next, bo->tori[bo->count] & 7);
    }
    if (te->nari && p != 0x08 && p != 0x18) {
      p &= 0x7;
      p += (bo->next == D_GOTE) ? 0x10 : 0;
    }
    bo->board[te->fm] = p;
  }
}

/**
 * ᤹ȤǤϡ1 ֤ʳ 0 ֤
 */
int daemon_dboard_can_back(DBoard *bo) {
  if (0 < bo->count) {
    return 1;
  } else {
    return 0;
  }
}

/** DBoard 渫 src  dest ˥ԡ */
void daemon_dboard_copy(DBoard *src, DBoard *dest) {
  D_TEBAN next;

  next = dest->front;

  *dest = *src;

  dest->front = next;
}

/** ꡢ촹 */
void daemon_dboard_toggle_next(DBoard *bo) {
  assert(bo->next == D_SENTE || bo->next == D_GOTE);

  if (bo->next == D_SENTE) {
    bo->next = D_GOTE;
  } else {
    bo->next = D_SENTE;
  }
}

void daemon_dboard_set_pbox(DBoard *bo, int p, int n) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);
  assert(0 <= n && n <= 18);

  bo->pbox[p] = n;
}

int daemon_dboard_get_pbox(DBoard *bo, int p) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  return bo->pbox[p];
}

void daemon_dboard_add_pbox(DBoard *bo, int p) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  bo->pbox[p]++;
}

void daemon_dboard_dec_pbox(DBoard *bo, int p) {
  assert(bo != NULL);
  assert(1 <= p && p <= 8);

  bo->pbox[p]--;
}

/**
 * ¦μ֤ next ˤ
 * get_board, set_board, set_piece, add_piece, dec_piece, get_piece 
 * Ѥ롣
 */
void daemon_dboard_set_front(DBoard *bo, D_TEBAN next) {
  bo->front = next;
}

D_TEBAN daemon_dboard_get_front(DBoard *bo) {
  return bo->front;
}

void daemon_dboard_toggle_front(DBoard *bo) {
  if (bo->front == D_SENTE) {
    bo->front = D_GOTE;
  } else {
    bo->front = D_SENTE;
  }
}

/**
 * DBoard Ƥ BOARD إԡ롣
 */
void daemon_dboard_copy_board(DBoard *dbo, BOARD *bo) {
  int x, y, p, i;

  assert(dbo != NULL);
  assert(bo != NULL);

  boInit(bo);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = daemon_dboard_get_board(dbo, x, y);
      boSetPiece(bo, (y << 4) + x, p); 
    }
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(dbo, 0, i);
    boSetPiece3(bo, i, p, 0);
  }

  for (i=1; i<=7; i++) {
    p = daemon_dboard_get_piece(dbo, 1, i);
    boSetPiece3(bo, i, p, 1);
  }

  bo->next = dbo->next;
}


/**
 * BOARD Ƥ DBoard إԡ롣
 */
void daemon_dboard_copy_from_board(BOARD *src, DBoard *dest) {
  int x, y, p, i;

  assert(src != NULL);
  assert(dest != NULL);

  daemon_dboard_init(dest);

  for (y=1; y<=9; y++) {
    for (x=1; x<=9; x++) {
      p = boGetPiece(src, (y << 4) + x); 
      daemon_dboard_set_board(dest, x, y, p);
    }
  }

  for (i=1; i<=7; i++) {
    p = src->piece[0][i];
    daemon_dboard_set_piece(dest, 0, i, p);

    p = src->piece[1][i];
    daemon_dboard_set_piece(dest, 1, i, p);
  }

  dest->next = src->next;
}

