/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * DMoveInfo class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/dmoveinfo.c,v $
 * $Id: dmoveinfo.c,v 1.1.1.1 2005/12/09 09:03:04 tokita Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "dte.h"
#include "dmoveinfo.h"

/** 󥹥ȥ饯 */
DMoveInfo* daemon_dmoveinfo_new(void) {
  DMoveInfo* mi;

  mi = malloc(sizeof(DMoveInfo));
  if (mi == NULL) {
    printf("No enough memory in daemon_dmovenifo_new().");
    abort();
  }

  daemon_dmoveinfo_init(mi);

  return mi;
}

void daemon_dmoveinfo_init(DMoveInfo* mi) {
  assert(mi != NULL);
  mi->count = 0;
}

/** ǥȥ饯 */
void daemon_dmoveinfo_free(DMoveInfo* mi) {
  assert(mi != NULL);
}

/** DTe ɲ */
void daemon_dmoveinfo_add(DMoveInfo* mi, DTe* te) {
  assert(mi != NULL);
  assert(te != NULL);
  assert(mi->count + 1 < DMOVEINFOMAX);
  if (DMOVEINFOMAX <= mi->count + 1) {
    return;
  }
  mi->te[mi->count++] = *te;
}

/** index ܤ DTe ؤΥݥ󥿤 */
DTe* daemon_dmoveinfo_get(DMoveInfo* mi, int index) {
  assert(mi != NULL);
  assert(index < DMOVEINFOMAX);
  if (DMOVEINFOMAX <= index) {
    return NULL;
  }
  return &(mi->te[index]);
}

/** ֺǸϿ֤ */
DTe* daemon_dmoveinfo_pop(DMoveInfo* mi) {
  assert(mi != NULL);

  if (mi->count == 0) {
    return NULL;
  }

  return &(mi->te[--mi->count]);
}

/** index ܤ DTe ˺ǸǤͤ */
void daemon_dmoveinfo_remove(DMoveInfo* mi, int index) {
  assert(mi != NULL);
  assert(index < DMOVEINFOMAX);
  if (DMOVEINFOMAX <= index) {
    return;
  }
  assert(index <= mi->count);
  if (mi->count < index) {
    return;
  }
  if (index == mi->count) {
    mi->count--;
    return;
  }
  mi->te[mi->count] = mi->te[index];
}

void daemon_dmoveinfo_output(DMoveInfo* mi, FILE* out) {
  int i;

  assert(mi != NULL);
  assert(out != NULL);

  for (i=0; i<mi->count; i++) {
    daemon_dte_output(&(mi->te[i]), out);
  }
}

int daemon_dmoveinfo_size(DMoveInfo* mi) {
  return mi->count;
}

