/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * FileReader class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/filereader.c,v $
 * $Id: filereader.c,v 1.1.1.1 2005/12/09 09:03:04 tokita Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include <ctype.h>
#include "filereader.h"


char* daemon_filereader_convert(FileReader* reader, char* s);
void daemon_filereader_return2zero(char *s);


/** 󥹥ȥ饯 */
FileReader* daemon_filereader_new(void) {
  FileReader* reader;

  reader = (FileReader *)malloc(sizeof(FileReader));
  if (reader == NULL) {
    printf("No enough memory in daemon_filereader_new().\n");
    abort();
  }

  daemon_filereader_init(reader);

  return reader;
}

void daemon_filereader_init(FileReader* reader) {
  memset(reader, 0, sizeof(FileReader));
  strncpy(reader->incode, "shift_jis", 32);
  strncpy(reader->outcode, "euc-jp", 32);
  reader->pushed = 0;
  reader->stat   = D_FILEREADER_STAT_SUCCESSFUL;
}

/** ǥȥ饯 */
void daemon_filereader_free(FileReader* reader) {
  assert(reader != NULL);
  free(reader->filename);
  free(reader);
}

char* daemon_filereader_convert(FileReader* reader, char *s) {
  static char buf[FILEREADER_MAXCOLUMN * 2];
  iconv_t cd;
  char *inbuf;
  char *outbuf;
  size_t isize;
  size_t osize;

  memset(buf, 0, FILEREADER_MAXCOLUMN * 2);
  inbuf = s;
  outbuf = buf;
  isize = strlen(s);
  osize = isize * 2;

  cd = iconv_open(reader->outcode, reader->incode);
  if ((int)cd == -1) {
    return NULL;
  }

  iconv(cd, &inbuf, &isize, &outbuf, &osize);

  iconv_close(cd);

  return buf;
}

/**
 * ե filename 򳫤
 * ξ 0 ֤Ԥ -1 ֤
 */
int daemon_filereader_open(FileReader* reader, char *filename) {
  int len;

  assert(reader != NULL);
  assert(filename != NULL);

  len = strlen(filename);
  reader->filename = (char *)malloc(len + 1);
  strncpy(reader->filename, filename, len);
  reader->filename[len] = '\0';

  reader->in = fopen(reader->filename, "r");
  if (reader->in == NULL) {
    reader->stat = D_FILEREADER_STAT_ERROR;
    return -1;
  }

  reader->cd = iconv_open(reader->outcode, reader->incode);
  if ((int)reader->cd == -1) {
    fclose(reader->in);
    reader->stat = D_FILEREADER_STAT_ERROR;
    return -1;
  }

  return 0;
}

void daemon_filereader_close(FileReader* reader) {
  assert(reader != NULL);

  iconv_close(reader->cd);
  fclose(reader->in);
}

/** ιԤ֤ */
char* daemon_filereader_next(FileReader* reader) {
  size_t isize;
  size_t osize;
  char *inbuf, *outbuf, *result;
  char buf[BUFSIZ];

  assert(reader != NULL);

  if (reader->pushed) {
    reader->pushed = 0;
    return reader->pushed_string;
  }

  if (fgets(buf, BUFSIZ, reader->in) == NULL) {
    return NULL;
  }
  
  isize = BUFSIZ;
  osize = isize * 2;
  inbuf  = buf;
  outbuf = result = (char *)reader->buf;
  memset(result, 0, osize);

  iconv(reader->cd, &inbuf, &isize, &outbuf, &osize);

  return result;
}

/** ϥɤ򥻥åȤ롣 */
void daemon_filereader_set_incode(FileReader* reader, char *incode) {
  assert(reader != NULL);
  assert(incode != NULL);
  strncpy(reader->incode, incode, 32);
}

/** ϥɤ򥻥åȤ롣 */
void daemon_filereader_set_outcode(FileReader* reader, char *outcode) {
  assert(reader != NULL);
  assert(outcode != NULL);
  strncpy(reader->outcode, outcode, 32);
}

/**
 * ƬΥڡ򥹥åפʸ֤
 * isspace()  1 ˤʤʸ򥹥åפ롣
 */
char* daemon_filereader_skip_space(char *s) {
  for ( ; *s != '\0' && isspace(*s); s++) {
    /* ʤˤ⤷ʤ */
  }

  return s;
}

/**
 * Ƭ c ǻϤޤϥȤȤɤФ
 * ȰʳιԤդ餽ιԤ֤
 * @param reader 
 * @param c 
 */
char* daemon_filereader_skip_comment(FileReader* reader, char c) {
  char *s;

  assert(reader != NULL);

  while ((s=daemon_filereader_next(reader)) != NULL) {
    s = daemon_filereader_skip_space(s);
    if (*s != c) {
      break;
    }
  }

  return s;
}

void daemon_filereader_push(FileReader* reader, char *s) {
  assert(reader != NULL);

  reader->pushed = 1;
  strncpy(reader->pushed_string, s, FILEREADER_MAXCOLUMN);
  reader->pushed_string[strlen(s)] = '\0';
}

/**
 * s Ĵ٤ 0x0D  0x0A 򸫤Ĥ 0 ֤롣
 * @param s оݤʸ
 */
void daemon_filereader_return2zero(char *s) {
  for ( ; *s != '\0'; s++) {
    if (*s == 0x0D || *s == 0x0A) {
      *s = '\0';
      return;
    }
  }
}

D_FILEREADER_STAT daemon_filereader_get_stat(FileReader* reader) {
  return reader->stat;
}

