/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * FileWriter class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/filewriter.c,v $
 * $Id: filewriter.c,v 1.1.1.1 2005/12/09 09:03:04 tokita Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iconv.h>
#include "filewriter.h"

/** 󥹥ȥ饯 */
FileWriter* daemon_filewriter_new(void) {
  FileWriter* writer;

  writer = (FileWriter *)malloc(sizeof(FileWriter));
  if (writer == NULL) {
    printf("No enough memory in daemon_filewriter_new().\n");
    abort();
  }

  daemon_filewriter_init(writer);

  return writer;
}

/**  */
void daemon_filewriter_init(FileWriter* writer) {
  assert(writer != NULL);

  daemon_filewriter_set_incode(writer, "euc-jp");
  daemon_filewriter_set_outcode(writer, "shift_jis");
  writer->out      = NULL;
  writer->filename = NULL;
  writer->stat     = D_FILEWRITER_SUCCESSFUL;
}

/** ǥȥ饯 */
void daemon_filewriter_free(FileWriter* writer) {
  assert(writer != NULL);
  free(writer->filename);
  free(writer);
}

/** 񤭹Ѥ˥ե򳫤 */
int daemon_filewriter_open(FileWriter* writer, char* filename) {
  int len;

  assert(writer != NULL);
  assert(filename != NULL);

  len = strlen(filename);
  writer->filename = (char *)malloc(len + 1);
  strncpy(writer->filename, filename, len);
  writer->filename[len] = '\0';

  writer->out = fopen(writer->filename, "w");
  if (writer->out == NULL) {
    writer->stat = D_FILEWRITER_ERROR;
    return -1;
  }

  writer->cd = iconv_open(writer->outcode, writer->incode);
  if ((int)writer->cd == -1) {
    fclose(writer->out);
    writer->stat = D_FILEWRITER_ERROR;
    return -1;
  }

  return 0;
}

/** եĤ */
void daemon_filewriter_close(FileWriter* writer) {
  assert(writer != NULL);

  iconv_close(writer->cd);
  fflush(writer->out);
  fclose(writer->out);
}

/** Խ */
void daemon_filewriter_put(FileWriter* writer, char* src) {
  size_t isize;
  size_t osize;
  char *outbuf, *result;

  assert(writer != NULL);
  assert(src != NULL);
  
  isize = strlen(src);
  osize = isize * 2;
  result = outbuf = (char *)malloc(osize);
  memset(result, 0, osize);

  iconv(writer->cd, &src, &isize, &outbuf, &osize);

  fprintf(writer->out, "%s", result);
}

/** ϥɤ򥻥åȤ롣*/
void daemon_filewriter_set_incode(FileWriter* writer, char *incode) {
  strncpy(writer->incode, incode, 32);
  writer->incode[strlen(incode)] = '\0';
}

/** ϥɤ򥻥åȤ롣 */
void daemon_filewriter_set_outcode(FileWriter* writer, char *outcode) {
  strncpy(writer->outcode, outcode, 32);
  writer->outcode[strlen(outcode)] = '\0';
}

/** եɤ߹߾֤ */
D_FILEWRITER_STAT daemon_filewriter_get_stat(FileWriter* writer) {
  return writer->stat;
}
