/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/misc.c,v $
 * $Id: misc.c,v 1.1.1.1 2005/12/09 09:03:06 tokita Exp $
 */

#include <gnome.h>
#include "sys/time.h"
#include "misc.h"

void daemon_abort(char *s) {
  printf("%s\n", s);
  abort();
}

/**
 * XPM Υǡɤ߹ GdkPixmap ˤ֤
 */
GdkPixmap* load_pixmap_from_data(GtkWidget* window, char **data) {
  GdkImlibImage* im;
  GdkPixmap *pixmap;

  /* ɤ */
  im = gdk_imlib_create_image_from_xpm_data(data);
  if (im == NULL) {
    daemon_abort("No enough memory in load_pixmap_from_data(). im is NULL\n");
  }
  gdk_imlib_render(im, im->rgb_width, im->rgb_height);
  pixmap = gdk_imlib_move_image(im);
  gdk_imlib_destroy_image(im);

  return pixmap;
}

GdkPixmap* load_pixmap_from_data_scaled(GtkWidget *window, char **data,
					gint w,
					gint h) {
  GdkImlibImage *im;
  GdkImlibImage *im2;
  GdkPixmap *pixmap;

  /* ɤ */
  im = gdk_imlib_create_image_from_xpm_data(data);
  if (im == NULL) {
    daemon_abort("No enough memory in load_pixmap_from_data_scaled(). im is NULL\n");
  }
  im2 = gdk_imlib_clone_scaled_image(im, w, h);
  if (im2 == NULL) {
    daemon_abort("No enough memory in load_pixmap_from_data_scaled(). im2 is NULL\n");
  }
  gdk_imlib_render(im2, im2->rgb_width, im2->rgb_height);
  pixmap = gdk_imlib_move_image(im2);
  gdk_imlib_destroy_image(im);
  gdk_imlib_destroy_image(im2);

  return pixmap;
}

GdkPixmap* load_pixmap_from_imlib_image_scaled(GtkWidget *window, 
					       GdkImlibImage *im,
					       gint w,
					       gint h) {
  GdkImlibImage *im2;
  GdkPixmap *pixmap;

  /* ɤ */
  im2 = gdk_imlib_clone_scaled_image(im, w, h);
  if (im2 == NULL) {
    daemon_abort("No enough memory in load_pixmap_from_data_scaled(). im2 is NULL\n");
  }
  gdk_imlib_render(im2, im2->rgb_width, im2->rgb_height);
  pixmap = gdk_imlib_move_image(im2);
  gdk_imlib_destroy_image(im2);

  return pixmap;
}

void set_color(GdkGC *gc, gushort r, gushort g, gushort b) {
  GdkColor color;

  color.red = r, color.green = g, color.blue = b;
  gdk_color_alloc(gdk_colormap_get_system(), &color);
  gdk_gc_set_background(gc, &color);
  gdk_gc_set_foreground(gc, &color);
}

/**
 * tv  now λֺ DOUBLECLICKTIMEOUT ξ TRUE ֤
 * ʳ FALSE ֤
 * @param tv  ˥å줿
 * @param now λ
 * @return ֺ DOUBLECLICKTIMEOUT ξ TRUE
 */
gboolean check_timeval(struct timeval *tv, struct timeval *now) {
  if (2 <= now->tv_sec - tv->tv_sec) {
    /* ðʾ庹 */
    return FALSE;
  }
  
  if (0 < now->tv_sec - tv->tv_sec) {
    if ((now->tv_usec + 1000000) - tv->tv_usec > DOUBLECLICKTIMEOUT * 1000) {
      return FALSE;
    } else {
      tv->tv_sec = 0L;
      tv->tv_usec = 0L;
      return TRUE;
    }
  } else {
    if (now->tv_usec - tv->tv_usec > DOUBLECLICKTIMEOUT * 1000) {
      return FALSE;
    } else {
      tv->tv_sec = 0L;
      tv->tv_usec = 0L;
      return TRUE;
    }
  }
}

/**
 * a  b ŤʤꤢäƤ뤫Ĵ٤롣
 * @param a 
 * @param b
 * @return ŤʤäƤ TRUE Ǥʤ FALSE ֤
 */
gboolean is_on_rectangle(GdkRectangle* a, GdkRectangle* b) {
  if (a->x <= b->x && b->x <= a->x + a->width &&
      a->y <= b->y && b->y <= a->y + a->height) {
    return TRUE;
  }
  
  if (b->x <= a->x && a->x <= b->x + b->width &&
      a->y <= b->y && b->y <= a->y + a->height) {
    return TRUE;
  }
  
  if (b->x <= a->x && a->x <= b->x + b->width &&
      b->y <= a->y && a->y <= b->y + b->height) {
    return TRUE;
  }

  if (a->x <= b->x && b->x <= a->x + a->width &&
      b->y <= a->y && a->y <= b->y + b->height) {
    return TRUE;
  }

  return FALSE;
}

/* timeout <= 0 ξϻ¤ʤ  */
void pending_loop(void) {
  while(gtk_events_pending()){ 
    gtk_main_iteration();
  }
}

gboolean daemon_on_rectangle(GdkRectangle *rect, GdkPoint *point) {
  g_assert(rect != NULL);
  g_assert(point != NULL);

  if (rect->x <= point->x &&
      rect->y <= point->y &&
      point->x <= rect->x + rect->width &&
      point->y <= rect->y + rect->height) {
    return TRUE;
  }

  return FALSE;
}

