/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * Record class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/record.c,v $
 * $Id: record.c,v 1.3 2005/12/12 08:49:36 tokita Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "dte.h"
#include "dmoveinfo.h"
#include "dboard.h"
#include "record.h"
#include "filereader.h"
#include "filewriter.h"

int daemon_record_output_csa_p(Record* record, FileWriter *writer);
int daemon_record_output_kif_p(Record* record, FileWriter *writer);
Record* daemon_record_load_csa_p(FileReader *reader);
Record* daemon_record_load_kif_p(FileReader *reader);

void daemon_record_output_csa_piece(Record* record, FileWriter* writer);
void daemon_record_output_csa_board(Record* record, FileWriter* writer);
void daemon_record_output_csa_te(Record* record, FileWriter* writer);

void daemon_record_output_kif_piece(DBoard *board, FileWriter *writer, D_TEBAN next);

void daemon_record_load_csa_board(Record* record, FileReader* reader);
void daemon_record_load_csa_te(Record* record, FileReader* reader);

void daemon_record_load_kif_header(Record* record, FileReader* reader);
void daemon_record_load_kif_board(Record* record, FileReader* reader);
void daemon_record_load_kif_te(Record* record, FileReader* reader);
void daemon_record_load_kif_piece(Record* record, char *s, D_TEBAN next);

void daemon_record_create_time_string(char *buf, time_t t, time_t total);

/** 󥹥ȥ饯 */
Record* daemon_record_new(void) {
  Record* record;

  record = malloc(sizeof(Record));
  if (record == NULL) {
    printf("No enough memory in daemon_record_new().");
    abort();
  }

  daemon_record_init(record);

  return record;
}

/** Record  */
void daemon_record_init(Record* record) {
  assert(record != NULL);

  memset(record, 0, sizeof(Record));

  record->teai = D_HIRATE;
  daemon_dboard_init(&(record->first_board));
  record->start_time = 0;
  daemon_dmoveinfo_init(&(record->mi));
  record->time[0] = (time_t)0;
  record->time[1] = (time_t)0;
  memset(&(record->player[0]), 0, MAXPLAYERNAME);
  memset(&(record->player[1]), 0, MAXPLAYERNAME);
  record->loadstat = D_LOAD_SUCCESS;
  record->matchstat = D_UNKNOWN;
}

/** ǥȥ饯 */
void daemon_record_free(Record* record) {
  assert(record != NULL);
  free(record);
}

/** ɲ */
void daemon_record_add_dte(Record* record, DTe* te, time_t t) {
  assert(record != NULL);
  assert(te != NULL);
  assert(0 <= t);

  record->time[record->mi.count] = t;
  daemon_dmoveinfo_add(&(record->mi), te);
}

/** CSAǥե filename ˽Ϥ롣 */
int daemon_record_output_csa(Record* record, char *filename) {
  FileWriter* writer;
  int r;
  
  assert(record != NULL);
  assert(filename != NULL);

  writer = daemon_filewriter_new();

  r = daemon_filewriter_open(writer, filename);
  if (r == -1) {
    return -1;
  }

  r = daemon_record_output_csa_p(record, writer);

  daemon_filewriter_close(writer);

  return r;
}

/** CSAǥե filename ˽Ϥ롣 */
int daemon_record_output_csa_code(Record* record, char *filename, char *code) {
  FileWriter* writer;
  int r;
  
  assert(record != NULL);
  assert(filename != NULL);

  writer = daemon_filewriter_new();
  daemon_filewriter_set_outcode(writer, code);

  r = daemon_filewriter_open(writer, filename);
  if (r == -1) {
    return -1;
  }

  r = daemon_record_output_csa_p(record, writer);

  daemon_filewriter_close(writer);

  return r;
}

/** CSAǥե filename ˽Ϥ롣 */
int daemon_record_output_csa_p(Record* record, FileWriter *writer) {
  assert(record != NULL);
  assert(writer != NULL);

  daemon_filewriter_put(writer, "'  жɼ̾\n");
  daemon_filewriter_put(writer, "N+");
  if (record->player[0][0] != '\0') {
    daemon_filewriter_put(writer, record->player[0]);
  }
  daemon_filewriter_put(writer, "\n");
  daemon_filewriter_put(writer, "N-");
  if (record->player[1][0] != '\0') {
    daemon_filewriter_put(writer, record->player[1]);
  }
  daemon_filewriter_put(writer, "\n");
  daemon_filewriter_put(writer, "'  \n");
  daemon_record_output_csa_board(record, writer);
  daemon_record_output_csa_piece(record, writer);
  daemon_filewriter_put(writer, "'  \n");
  daemon_filewriter_put(writer, record->first_board.next ? "-\n" : "+\n");
  daemon_record_output_csa_te(record, writer);

  return 0;
}

/**  */
void daemon_record_output_csa_piece(Record* record, FileWriter* writer) {
  DBoard *board;
  int p, j, i;
  char buf[BUFSIZ];
  static const char koma[][3] = {
    "",   "FU", "KY", "KE", "GI", "KI", "KA", "HI",
  };

  board = &(record->first_board);

  daemon_filewriter_put(writer, "'  \n");
  for (j=7; 1 <= j; j--) {
    p = daemon_dboard_get_piece(board, D_SENTE, j);
    for (i=0; i<p; i++) {
      sprintf(buf, "P+00%s\n", koma[j]);
      daemon_filewriter_put(writer, buf);
    }    
  }

  daemon_filewriter_put(writer, "'  \n");
  for (j=7; 1 <= j; j--) {
    p = daemon_dboard_get_piece(board, D_GOTE, j);
    for (i=0; i<p; i++) {
      sprintf(buf, "P-00%s\n", koma[j]);
      daemon_filewriter_put(writer, buf);
    }
  }
}

void daemon_record_output_csa_board(Record* record, FileWriter* writer) {
  DBoard *bo;
  int p;
  int x, y;
  char buf[BUFSIZ];
  static const char koma[][3] = {
    "",   "FU", "KY", "KE", "GI", "KI", "KA", "HI",
    "OU", "TO", "NY", "NK", "NG", "",   "UM", "RY",
  };

  bo = &(record->first_board);

  for (y=1; y <= 9; y++) {
    sprintf(buf, "P%d", y);
    daemon_filewriter_put(writer, buf);
    for (x=9; 1 <= x; x--) {
      p = daemon_dboard_get_board(bo, x, y);
      if (p) {
	sprintf(buf, "%c%s", (p & 0x10) ? '-' : '+', koma[p & 0x0F]);
	daemon_filewriter_put(writer, buf);
      } else {
	daemon_filewriter_put(writer, " * ");
      }
    }
    daemon_filewriter_put(writer, "\n");
  }
}

void daemon_record_output_csa_te(Record* record, FileWriter* writer) {
  static char buf[80];
  DTe *te;
  DBoard *bo;
  int next, p;
  int i;
  static char const koma_str[][3] = {
    "",   "FU", "KY", "KE", "GI", "KI", "KA", "HI",
    "OU", "TO", "NY", "NK", "NG", "",   "UM", "RY",
  };

  next = record->first_board.next;
  bo = daemon_dboard_new();
  daemon_dboard_copy(&(record->first_board), bo);

  daemon_filewriter_put(writer, "'  ؤ\n");

  for (i=0; i<record->mi.count; i++) {
    te = &(record->mi.te[i]);
    memset(buf, 0, 80);
    if (te->uti) {
      /* Ǥ */
      sprintf(buf, "%c%d%d%d%d%s\n",
	      (next == D_SENTE) ? '+' : '-',
	      0, 0,
	      te->to & 0xF, te->to >> 4,
	      koma_str[te->uti]);
    } else {
      /* ư */
      p = daemon_dboard_get_board(bo, te->fm & 0xF, te->fm >> 4);
      p += te->nari ? 0x08 : 0;
      sprintf(buf, "%c%d%d%d%d%s\n",
	      (next == D_SENTE) ? '+' : '-',
	      te->fm & 0xF, te->fm >> 4,
	      te->to & 0xF, te->to >> 4,
	      koma_str[p & 0xF]);
    }
    daemon_filewriter_put(writer, buf);
    memset(buf, 0, 80);
    sprintf(buf, "T%d\n", record->time[i]);
    daemon_filewriter_put(writer, buf);
    daemon_dboard_move(bo, te);

    /* ָ */
    next ^= 1;
  }

  if (record->matchstat == D_TORYO) {
    daemon_filewriter_put(writer, "%TORYO");
  } else if (record->matchstat == D_MATTA) {
    daemon_filewriter_put(writer, "%MATTA");
  } else if (record->matchstat == D_SENNICHITE) {
    daemon_filewriter_put(writer, "%SENNICHITE");
  } else if (record->matchstat == D_JISHOGI) {
    daemon_filewriter_put(writer, "%JISHOGI");
  } else if (record->matchstat == D_TSUMI) {
    daemon_filewriter_put(writer, "%TSUMI");
  } else if (record->matchstat == D_FUZUMI) {
    daemon_filewriter_put(writer, "%FUZUMI");
  } else if (record->matchstat == D_ERROR) {
    daemon_filewriter_put(writer, "%ERROR");
  }

  daemon_dboard_free(bo);
}

void daemon_record_create_time_string(char *buf, time_t t, time_t total) {
  time_t min, sec;
  time_t total_hour, total_min, total_sec;

  min = t / 60;
  sec = t % 60;
  total_hour = total / 3600;
  total = total % 3600;
  total_min  = total / 60;
  total_sec  = total % 60;

  sprintf(buf, "(%2d:%02d/%02d:%02d:%02d)",
	  (int)min,
	  (int)sec,
	  (int)total_hour,
	  (int)total_min,
	  (int)total_sec);
}

void daemon_record_output_kif_piece(DBoard *board, FileWriter *writer, D_TEBAN next) {
  static char *koma_data[] = {
    "",   "", "", "", "", "", "", "",
  };
  static char *knum_data[] = {
    "",  "", "", "", "", "", "ϻ", "", "Ȭ", "",
  };
  int count;
  int i;

  if (next == D_SENTE) {
    daemon_filewriter_put(writer, "λ");
  } else {
    daemon_filewriter_put(writer, "λ");
  }

  for (i=1; i<=7; i++) {
    if (daemon_dboard_get_piece(board, next, i) != 0) {
      break;
    }
  }

  if (i == 8) {
    /* ʤ */
    daemon_filewriter_put(writer, "ʤ\n");
  }

  for (i=7; 1 <= i; i--) {
    count = daemon_dboard_get_piece(board, next, i);
    if (count == 0) {
      continue;
    }
    daemon_filewriter_put(writer, koma_data[i]);
    if (count == 1) {
      daemon_filewriter_put(writer, "");
      continue;
    }
    if (10 <= count) {
      daemon_filewriter_put(writer, "");
      count -= 10;
      if (count == 0) {
	daemon_filewriter_put(writer, "");
	continue;
      }
    }
    daemon_filewriter_put(writer, knum_data[count]);
    daemon_filewriter_put(writer, "");
  }

  daemon_filewriter_put(writer, "\n");
}

int daemon_record_output_kif_p(Record* record, FileWriter *writer) {
  static char *koma_data[] = {
    "",   "", "", "", "", "", "", "",
    "", "", "", "", "", "",   "", "ζ",
  };
  static char *koma_data2[] = {
    "",   "", "", "", "", "", "", "",
    "", "", "", "", "", "",   "", "ζ",
  };
  static char *knum_data[] = {
    "",  "", "", "", "", "", "ϻ", "", "Ȭ", "",
  };
  static char *num_data[] = {
    "",  "", "", "", "", "", "", "", "", "", 
  };
  DTe *te;
  static char buf[BUFSIZ];
  static char buf2[BUFSIZ];
  static char time_buf[BUFSIZ];
  DBoard *board;
  int last_xy;
  int x, y;
  int p;
  int i;
  time_t total[2];

  board = daemon_dboard_new();
  daemon_dboard_copy(&(record->first_board), board);

  daemon_filewriter_put(writer, "# ---- daemonshogi kif file ----\n");
  daemon_filewriter_put(writer, "# ե̾");
  daemon_filewriter_put(writer, writer->filename);
  daemon_filewriter_put(writer, "\n");
  daemon_filewriter_put(writer, "ж\n");
  daemon_filewriter_put(writer, "䡧ʿ\n");
  daemon_record_output_kif_piece(board, writer, D_GOTE);
  daemon_filewriter_put(writer, "          \n");
  daemon_filewriter_put(writer, "+---------------------------+\n");
  for (y=1; y<=9; y++) {
    daemon_filewriter_put(writer, "|");
    for (x=9; 1 <= x; x--) {
      p = daemon_dboard_get_board(board, x, y);
      if (p == 0) {
	daemon_filewriter_put(writer, " ");
      } else {
	if (p < 0x10) {
	  daemon_filewriter_put(writer, " ");
	} else {
	  daemon_filewriter_put(writer, "v");
	}
	daemon_filewriter_put(writer, koma_data2[p & 0xF]);
      }
    }
    daemon_filewriter_put(writer, "|");
    daemon_filewriter_put(writer, knum_data[y]);
    daemon_filewriter_put(writer, "\n");
  }
  daemon_filewriter_put(writer, "+---------------------------+\n");
  daemon_record_output_kif_piece(board, writer, D_SENTE);
  if (record->first_board.next == D_SENTE) {
    daemon_filewriter_put(writer, "\n");
  } else {
    daemon_filewriter_put(writer, "\n");
  }
  daemon_filewriter_put(writer, "ꡧ");
  if (record->player[0][0] != '\0') {
    daemon_filewriter_put(writer, record->player[0]);
  }
  daemon_filewriter_put(writer, "\n");
  daemon_filewriter_put(writer, "ꡧ");
  if (record->player[1][0] != '\0') {
    daemon_filewriter_put(writer, record->player[1]);
  }
  daemon_filewriter_put(writer, "\n");
  daemon_filewriter_put(writer, "----ؼ-----------\n");

  total[0] = 0;
  total[1] = 0;
  last_xy = 0;
  for (i=0; i<record->mi.count; i++) {
    te = daemon_dmoveinfo_get(&(record->mi), i);
    if (te->uti) {
      sprintf(buf, "%4d %s%s%s",
	      i + 1,
	      num_data[te->to & 0xF],
	      knum_data[te->to >> 4],
	      koma_data[te->uti]);
      last_xy = te->to;
    } else {
      p = daemon_dboard_get_board(board, te->fm & 0xF, te->fm >> 4);
      p &= 0xF;
      if (last_xy == te->to) {
	sprintf(buf, "%4d Ʊ%s(%d%d)",
		i + 1,
		koma_data[p],
		te->fm & 0xF,
		te->fm >> 4);
      } else {
	sprintf(buf, "%4d %s%s%s%s(%d%d)",
		i + 1,
		num_data[te->to & 0xF],
		knum_data[te->to >> 4],
		koma_data[p],
		te->nari ? "" : "",
		te->fm & 0xF, 
		te->fm >> 4);
	      }
      last_xy = te->to;
    }
    total[board->next] += record->time[i];
    daemon_record_create_time_string(time_buf, record->time[i], total[board->next]);
    sprintf(buf2, "%-17s %s\n", buf, time_buf);
    daemon_filewriter_put(writer, buf2);
    daemon_dboard_move(board, te);
    daemon_dboard_toggle_next(board);
  }

  if (record->matchstat == D_TORYO) {
    total[board->next] += record->time[i];
    daemon_record_create_time_string(time_buf, record->time[i], total[board->next]);
    sprintf(buf, "%4d λ         %s\n", i + 1, time_buf);
    daemon_filewriter_put(writer, buf);
  }

  daemon_dboard_free(board);

  return 0;
}

/**  KIFǥե filename ˽Ϥ롣 */
int daemon_record_output_kif(Record* record, char *filename) {
  FileWriter* writer;
  int r;
  
  assert(record != NULL);
  assert(filename != NULL);

  writer = daemon_filewriter_new();

  r = daemon_filewriter_open(writer, filename);
  if (r == -1) {
    return -1;
  }

  r = daemon_record_output_kif_p(record, writer);

  daemon_filewriter_close(writer);

  return r;
}

/**  KIFǥե filename ˽Ϥ롣 */
int daemon_record_output_kif_code(Record* record, char *filename, char *code) {
  FileWriter* writer;
  int r;
  
  assert(record != NULL);
  assert(filename != NULL);

  writer = daemon_filewriter_new();
  daemon_filewriter_set_outcode(writer, code);

  r = daemon_filewriter_open(writer, filename);
  if (r == -1) {
    return -1;
  }

  r = daemon_record_output_kif_p(record, writer);

  daemon_filewriter_close(writer);

  return r;
}

/**
 * ե filename ɤ߹ Record 롣
 * եϼưǧ롣
 */
Record* daemon_record_load(char *filename) {
  Record *record;
  D_RECFORMAT recformat;

  assert(filename != NULL);

  record = NULL;
  recformat = daemon_record_reserach_format(filename);

  switch (recformat) {
  case D_FORMAT_ERROR:
  case D_FORMAT_UNKNOWN:
    record = NULL;
    break;
  case D_FORMAT_CSA:
    record = daemon_record_load_csa(filename);
    break;
  case D_FORMAT_KIF:
    record = daemon_record_load_kif(filename);
    break;
  default:
    break;
  }

  return record;
}

/**
 * ե filename ɤ߹ Record 롣
 * եϼưǧ롣
 */
Record* daemon_record_load_code(char *filename, char *code) {
  Record *record;
  D_RECFORMAT recformat;

  assert(filename != NULL);

  record = NULL;
  recformat = daemon_record_reserach_format(filename);

  switch (recformat) {
  case D_FORMAT_ERROR:
  case D_FORMAT_UNKNOWN:
    record = NULL;
    break;
  case D_FORMAT_CSA:
    record = daemon_record_load_csa_code(filename, code);
    break;
  case D_FORMAT_KIF:
    record = daemon_record_load_kif_code(filename, code);
    break;
  default:
    break;
  }

  return record;
}

void daemon_record_load_csa_board(Record* record, FileReader* reader) {
  int x, y;
  char *s;
  int next;
  int i;
  static char const koma_data[][3] = {
    "",   "FU", "KY", "KE", "GI", "KI", "KA", "HI",
    "OU", "TO", "NY", "NK", "NG", "",   "UM", "RY",
  };

  next = D_SENTE;

  while ((s=daemon_filereader_skip_comment(reader, '\'')) != NULL) {
    if (*s != 'P') {
      daemon_filereader_push(reader, s);
      break;
    }

    s++;

    if (*s == '+' || *s == '-') {
      /*  */
      int n = (*s == '+') ? 0 : 1;

      if (*(s + 1) != '0' && *(s + 2) != '0') {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }

      s += 3;

      for (i=1; i <= 7; i++) {
	if (strncmp(s, koma_data[i], 2) == 0) {
	  break;
	}
      }

      if (7 < i) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }

      daemon_dboard_add_piece(&(record->first_board), n, i);

      continue;
    }

    y = *s - 0x30;
    x = 9;
    
    if (y < 1 || 9 < y) {
      /* ;夢ꤨʤ */
      record->loadstat = D_LOAD_ERROR;
      return;
    }

    for ( ; *s != '\0'; s++) {
      if (*s == ' ') {
	continue;
      }
      if (*s == '+') {
	next = D_SENTE;
	continue;
      }
      if (*s == '-') {
	next = D_GOTE;
	continue;
      }
      if (isupper(*s)) {
	for (i=1; i<=0xF; i++) {
	  if (strncmp(s, koma_data[i], 2) == 0) {
	    break;
	  }
	}
	if (0xF < i) {
	  /* ;夢ꤨʤ */
	  record->loadstat = D_LOAD_ERROR;
	  return;
	}
	i = i + (next==D_GOTE) * 0x10;
	daemon_dboard_set_board(&(record->first_board), x, y, i);
	x--;
	s++;
      } else if (*s == '*') {
	daemon_dboard_set_board(&(record->first_board), x, y, 0);
	x--;
      }
      if (x < 0) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }
    }
  }

  if (s != NULL) {
    daemon_filereader_push(reader, s);
  }
}

void daemon_record_load_csa_te(Record* record, FileReader* reader) {
  int p, p2;
  char *s;
  DTe te;
  int t;
  DBoard *board;
  static char const koma_data[][3] = {
    "",   "FU", "KY", "KE", "GI", "KI", "KA", "HI",
    "OU", "TO", "NY", "NK", "NG", "",   "UM", "RY",
  };

  record->matchstat = D_UNKNOWN;
  board = daemon_dboard_new();
  daemon_dboard_copy(&(record->first_board), board);

  while ((s=daemon_filereader_skip_comment(reader, '\'')) != NULL) {
    if (*s == '+' || *s == '-') {
      /* μ */
      te.fm = ((s[2] - 0x30) << 4) + s[1] - 0x30;
      te.to = ((s[4] - 0x30) << 4) + s[3] - 0x30;
      s += 5;
      for (p=1; p<=0xF; p++) {
	if (strncmp(s, koma_data[p], 2) == 0) {
	  break;
	}
      }
      if (0xF < p) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	daemon_dboard_free(board);
	return;
      }
      if (te.fm == 0) {
	/* Ǥξ */
	if (te.to < 0x11 || 0x99 < te.to) {
	  /* ;夢ꤨʤ */
	  record->loadstat = D_LOAD_ERROR;
	  daemon_dboard_free(board);
	  return;
	}
	te.nari = 0;
	te.uti = p;
      } else {
	/* ư */
	if (te.fm < 0x11 || 0x99 < te.fm ||
	    te.to < 0x11 || 0x99 < te.to) {
	  /* ;夢ꤨʤ */
	  record->loadstat = D_LOAD_ERROR;
	  daemon_dboard_free(board);
	  return;
	}
	te.uti = 0;
	if (p == 0x05 || p == 0x8 || p == 0x15 || p == 0x18) {
	  /* ⡢ξ礤Ϥʤ */
	  te.nari = 0;
	} else {
	  /* ⡢ʳǰưꡢư褬ʤ顢Ȥ */
	  p2 = daemon_dboard_get_board(board, te.fm & 0xF, te.fm >> 4);
	  if (0x9 <= p && (p2 & 0xF) < 0x9) {
	    te.nari = 1;
	  } else {
	    te.nari = 0;
	  }
	}
      }
      s = daemon_filereader_skip_comment(reader, '\'');
      if (s == NULL) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	daemon_dboard_free(board);
	return;
      }
      (void)sscanf(s, "T%d", &t);
      daemon_record_add_dte(record, &te, (time_t)t);

      daemon_dboard_move(board, &te);
    } else if (*s == '%') {
      if (strncmp(s, "%TORYO", 6) == 0) {
	record->matchstat = D_TORYO;
      } else if (strncmp(s, "%MATTA", 6) == 0) {
	record->matchstat = D_MATTA;
      } else if (strncmp(s, "%SENNICHITE", 11) == 0) {
	record->matchstat = D_SENNICHITE;
      } else if (strncmp(s, "%JISHOGI", 8) == 0) {
	record->matchstat = D_JISHOGI;
      } else if (strncmp(s, "%TSUMI", 6) == 0) {
	record->matchstat = D_TSUMI;
      } else if (strncmp(s, "%FUZUMI", 7) == 0) {
	record->matchstat = D_FUZUMI;
      } else if (strncmp(s, "%ERROR", 6) == 0) {
	record->matchstat = D_ERROR;
      } 
    } else {
      /* ꤨʤʬ */
    }
  }

  daemon_dboard_free(board);
}

/** CSAΥեɤ߹ Record ֤ */
Record* daemon_record_load_csa_p(FileReader *reader) {
  Record* record;
  char *s;
  
  assert(reader != NULL);

  record = daemon_record_new();

  while ((s=daemon_filereader_skip_comment(reader, '\'')) != NULL) {
    if (strncmp(s, "N+", 2) == 0) {
      /* жɼ̾ */
      s += 2;
      daemon_filereader_return2zero(s);
      strncpy(record->player[0], s, MAXPLAYERNAME);
      record->player[0][strlen(s)] = '\0';
      daemon_filereader_return2zero(record->player[0]);
    } else if (strncmp(s, "N-", 2) == 0) {
      /* жɼ̾ */
      s += 2;
      daemon_filereader_return2zero(s);
      strncpy(record->player[1], s, MAXPLAYERNAME);
      record->player[1][strlen(s)] = '\n';
      daemon_filereader_return2zero(record->player[1]);
    } else if (strncmp(s, "PI", 2) == 0) {
      /*  */
    } else if (strncmp(s, "P1", 2) == 0) {
      /*   */
      daemon_filereader_push(reader, s);
      daemon_record_load_csa_board(record, reader);
    } else if (strncmp(s, "+", 1) == 0) {
      /*   */
      record->first_board.next = D_SENTE;
      break;
    } else if (strncmp(s, "-", 1) == 0) {
      /*   ָ*/
      record->first_board.next = D_GOTE;
      break;
    } else {
      break;
    }
  }

  daemon_record_load_csa_te(record, reader);

  return record;
}

/** CSAΥեɤ߹ Record ֤ */
Record* daemon_record_load_csa(char *filename) {
  Record* record;
  FileReader* reader;
  
  assert(filename != NULL);

  reader = daemon_filereader_new();
  daemon_filereader_open(reader, filename);

  if (daemon_filereader_get_stat(reader) == D_FILEREADER_STAT_ERROR) {
    /* ɤ߹ߤ˼Ԥ */
    daemon_filereader_free(reader);
    return NULL;
  }

  record = daemon_record_load_csa_p(reader);

  daemon_filereader_close(reader);

  return record;
}

/** CSAΥեɤ߹ Record ֤ */
Record* daemon_record_load_csa_code(char *filename, char *code) {
  Record* record;
  FileReader* reader;
  
  assert(filename != NULL);

  reader = daemon_filereader_new();
  daemon_filereader_set_incode(reader, code);
  daemon_filereader_open(reader, filename);

  if (daemon_filereader_get_stat(reader) == D_FILEREADER_STAT_ERROR) {
    /* ɤ߹ߤ˼Ԥ */
    daemon_filereader_free(reader);
    return NULL;
  }

  record = daemon_record_load_csa_p(reader);

  daemon_filereader_close(reader);

  return record;
}

/** β */
void daemon_record_load_kif_piece(Record* record, char *s, D_TEBAN next) {
  int p, i, num;
  static char *koma_data[] = {
    "",   "", "", "", "", "", "", "",
  };
  static char *num_data[] = {
    "",  "", "", "", "", "", "ϻ", "", "Ȭ", "",
  };

  s += 12;

  if (strncmp("ʤ", s, 4) == 0) {
    /* ʤ */
    return;
  }

  while (*s != '\0' && *s != 0xd && *s != 0xa) {
    for (p=1; p<=7; p++) {
      if (strncmp(koma_data[p], s, 2) == 0) {
	break;
      }
    }
    if (7 < p) {
      /* ;夢ꤨʤ */
      record->loadstat = D_LOAD_ERROR;
      return;
    }
    s += 2;
    num = 0;
    if (strncmp(s, "", 2) == 0) {
      /*  */
      s += 2;
      num = 10;
    }
    if (strncmp("", s, 2) == 0) {
      s += 2;
      num++;
    } else {
      for (i=1; i<=9; i++) {
	if (strncmp(num_data[i], s, 2) == 0) {
	  s += 4;
	  break;
	}
      }
      if (9 < i) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }
      num += i;
    }
    daemon_dboard_set_piece(&(record->first_board), next, p, num);
  }
}

/** KIFΥեΥإåʬɤ߹ Record ֤ */
void daemon_record_load_kif_header(Record* record, FileReader* reader) {
  char *s;

  while ((s=daemon_filereader_skip_comment(reader, '#')) != NULL) {
    if (strncmp("ж", s, 8) == 0) {
    } else if (strncmp("䡧", s, 8) == 0) {
    } else if (strncmp("λ", s, 12) == 0) {
      daemon_record_load_kif_piece(record, s, D_GOTE);
    } else if (strncmp("+---------------------------+", s, 29) == 0) {
      daemon_record_load_kif_board(record, reader);
    } else if (strncmp("λ", s, 12) == 0) {
      daemon_record_load_kif_piece(record, s, D_SENTE);
    } else if (strncmp("", s, 6) == 0) {
      record->first_board.next = D_SENTE;
    } else if (strncmp("", s, 6) == 0) {
      record->first_board.next = D_GOTE;
    } else if (strncmp("ꡧ", s, 6) == 0) {
      daemon_filereader_return2zero(s + 6);
      strncpy(record->player[0], s + 6, sizeof(record->player[0]));
    } else if (strncmp("ꡧ", s, 6) == 0) {
      daemon_filereader_return2zero(s + 6);
      strncpy(record->player[1], s + 6, sizeof(record->player[0]));
    } else if (strncmp("----ؼ-----------", s, 6) == 0) {
      break;
    }
  }  
}

/** KIFΥեΥإåʬɤ߹ Record ֤ */
void daemon_record_load_kif_board(Record* record, FileReader* reader) {
  char *s;
  int x, y, p;
  int next;
  static char *koma_data[] = {
    "",   "", "", "", "", "", "", "",
    "", "", "", "", "", "",   "", "ζ",
  };

  for (y=1; (s=daemon_filereader_skip_comment(reader, '#')) != NULL && y<=9; y++) {
    s++;
    for (x=9; 1 <= x; x--) {
      next = (*s == 'v') ? D_GOTE : D_SENTE;
      s++;
      if (strncmp("", s, 2) == 0) {
	s += 2;
	continue;
      }
      for (p=1; p<=0xF; p++) {
	if (strncmp(koma_data[p], s, 2) == 0) {
	  break;
	}
      }
      if (0xF < p) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }
      if (next == D_GOTE) {
	p += 0x10;
      }
      daemon_dboard_set_board(&(record->first_board), x, y, p);
      s += 2;
    }
  }  
}

void daemon_record_load_kif_te(Record* record, FileReader* reader) {
  static char *koma_data[] = {
    "",   "", "", "", "", "", "", "",
    "", "", "", "", "", "",   "", "ζ",
  };
  static char *knum_data[] = {
    "",  "", "", "", "", "", "ϻ", "", "Ȭ", "",
  };
  static char *num_data[] = {
    "",  "", "", "", "", "", "", "", "", "", 
  };
  char *s, *ss;
  int p, x, y, last_x, last_y;
  int nari_koma;
  int next;
  DTe te;
  time_t t;

  p = x = y = last_x = last_y = 0;
  next = record->first_board.next;

  while ((s=daemon_filereader_next(reader)) != NULL) {
    if (strncmp("ޤ", s, 4) == 0) {
      break;
    }
    if (*s == '*') {
      continue;
    }
    ss = s;
    s += 5;
    if (strncmp("Ʊ", s, 4) == 0) {
      x = last_x;
      y = last_y;
      s += 4;
    } else if (strncmp("ͤ", s, 4) == 0) {
      record->matchstat = D_TSUMI;
      break;
    } else if (strncmp("λ", s, 4) == 0) {
      record->matchstat = D_TORYO;
      break;
    } else {
      for (x=1; x <= 9; x++) {
	if (strncmp(num_data[x], s, 2) == 0) {
	  break;
	}
      }
      if (9 < x) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }
      s += 2;
      for (y=1; y <= 9; y++) {
	if (strncmp(knum_data[y], s, 2) == 0) {
	  break;
	}
      }
      if (9 < y) {
	/* ;夢ꤨʤ */
	record->loadstat = D_LOAD_ERROR;
	return;
      }
      s += 2;
    }

    if (strncmp("", s, 2) == 0) {
      nari_koma = 1;
      s += 2;
    } else {
      nari_koma = 0;
    }
    for (p=1; p <= 0xF; p++) {
      if (strncmp(koma_data[p], s, 2) == 0) {
	break;
      }
    }
    if (0xF < p) {
      /* ;夢ꤨʤ */
      record->loadstat = D_LOAD_ERROR;
      return;
    }
    s += 2;
    if (nari_koma) {
      p |= 0x08;
    }
    if (strncmp("", s, 2) == 0) {
      te.uti = p;
      te.fm = 0;
      s += 2;
    } else {
      te.uti = 0;
    }
    if (strncmp("", s, 2) == 0) {
      te.nari = 1;
      te.uti = 0;
      s += 2;
    } else {
      te.nari = 0;
    }
    s++;
    te.fm = ((*(s + 1) - 0x30) << 4) + (*s - 0x30);
    te.to = (y << 4) + x;
    t = (ss[20] - 0x30) * 60 + (ss[22] - 0x30) * 10 + (ss[23] - 0x30);
    daemon_record_add_dte(record, &te, t);
    last_x = x;
    last_y = y;
    
    next ^= 1;
  }
}

/** KIFΥեɤ߹ Record ֤ */
Record* daemon_record_load_kif_p(FileReader *reader) {
  Record *record;

  assert(reader != NULL);

  record = daemon_record_new();
  daemon_record_load_kif_header(record, reader);
  daemon_record_load_kif_te(record, reader);

  return record;
}

/** KIFΥեɤ߹ Record ֤ */
Record* daemon_record_load_kif(char *filename) {
  Record* record;
  FileReader* reader;
  
  assert(filename != NULL);

  reader = daemon_filereader_new();
  daemon_filereader_open(reader, filename);

  if (daemon_filereader_get_stat(reader) == D_FILEREADER_STAT_ERROR) {
    /* ɤ߹ߤ˼Ԥ */
    daemon_filereader_free(reader);
    return NULL;
  }

  record = daemon_record_load_kif_p(reader);

  daemon_filereader_close(reader);

  return record;
}

/** KIFΥեɤ߹ Record ֤ */
Record* daemon_record_load_kif_code(char *filename, char *code) {
  Record* record;
  FileReader* reader;
  
  assert(filename != NULL);

  reader = daemon_filereader_new();
  daemon_filereader_set_incode(reader, code);
  daemon_filereader_open(reader, filename);

  if (daemon_filereader_get_stat(reader) == D_FILEREADER_STAT_ERROR) {
    /* ɤ߹ߤ˼Ԥ */
    daemon_filereader_free(reader);
    return NULL;
  }

  record = daemon_record_load_kif_p(reader);

  daemon_filereader_close(reader);

  return record;
}

/** ե filename ΥեޥåȤĴ֤ */
D_RECFORMAT daemon_record_reserach_format(char *filename) {
  char *s;
  FileReader *reader;
  D_RECFORMAT format;
  int i;
  
  assert(filename != NULL);

  reader = daemon_filereader_new();
  if (reader == NULL) {
    return D_FORMAT_ERROR;
  }
  if (daemon_filereader_open(reader, filename) != 0) {
    return D_FORMAT_ERROR;
  }
  if (reader == NULL) {
    printf("No enough memory in daemon_record_reserach_format().\n");
    abort();
  }

  format = D_FORMAT_ERROR;

  for (i=0; i<256 && (s=daemon_filereader_next(reader)) != NULL; i++) {
    if (strncmp(s, "N+", 2) == 0) {
      format = D_FORMAT_CSA;
      break;
    } else if (strncmp(s, "", 10) == 0) {
      format = D_FORMAT_KIF;
      break;
    } else if (strncmp(s, "λ", 10) == 0) {
      format = D_FORMAT_KIF;
      break;
    } else if (strncmp(s, "ж", 8) == 0) {
      format = D_FORMAT_KIF;
      break;
    }
  }

  daemon_filereader_close(reader);
  daemon_filereader_free(reader);

#ifdef NOREAD  
  for (i=0; i<4; i++) {
    /* s = daemon_filereader_next(reader); */
    s = daemon_filereader_skip_comment(reader, '\'');
    if (s == NULL) {
      break;
    }
    strncpy((char *)buf[i], s, FILEREADER_MAXCOLUMN);
  }

  daemon_filereader_close(reader);
  daemon_filereader_free(reader);

  if (strncmp((char *)buf[0], "N+", 2) == 0 ||
      strncmp((char *)buf[1], "N+", 2) == 0 ||
      strncmp((char *)buf[2], "N+", 2) == 0 ||
      strncmp((char *)buf[3], "N+", 2) == 0) {
    /* 餯CSAեޥå */
    return D_FORMAT_CSA;
  }

  if (strncmp((char *)buf[0], "ж", 8) == 0 ||
      strncmp((char *)buf[1], "ж", 8) == 0 ||
      strncmp((char *)buf[2], "ж", 8) == 0 ||
      strncmp((char *)buf[3], "ж", 8) == 0) {
    /* 餯KIFեޥå */
    return D_FORMAT_KIF;
  }
#endif /* NOREAD */

  return format;
}

D_LOADSTAT daemon_record_get_loadstat(Record* record) {
  return record->loadstat;
}

D_MATCHSTAT daemon_record_get_matchstat(Record *record) {
  assert(record != NULL);

  return record->matchstat;
}

char *daemon_record_get_player(Record *record,
			      D_TEBAN next) {
  return record->player[next];
}

void daemon_record_set_player(Record *record,
			      D_TEBAN next,
			      char *s) {
  assert(record != NULL);
  assert(next == D_SENTE || next == D_GOTE);
  assert(s != NULL);

  strncpy(record->player[next], s, sizeof(record->player[next]));
}

