/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * Record class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/record.h,v $
 * $Id: record.h,v 1.1.1.1 2005/12/09 09:03:07 tokita Exp $
 */

#ifndef _RECORD_H_
#define _RECORD_H_

#include "dte.h"
#include "dmoveinfo.h"
#include "dboard.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
  

/** ǵϿ */
#define MAXRECORD DMOVEINFOMAX
/** жɼԤ̾κʸ */
#define MAXPLAYERNAME 256

/** 礤ɽ */
typedef enum {
  /** ʿ */
  D_HIRATE = 0,
  /**  */
  D_KYOOCHI = 1,
  /**  */
  D_KAKUOCHI = 2,
  /**  */
  D_HISYAOCHI = 3,
  /** ѹ */
  D_KAKUKYOOCHI = 4,
  /** ֹ */
  D_HISYAKYOOCHI = 5,
  /**  */
  D_NIMAIOCHI = 6,
  /**  */
  D_YONMAIOCHI = 7,
  /** ϻ */
  D_ROKUMAIOCHI = 8,
  /** Ȭ */
  D_HACHIMAIOCHI = 9,
} D_TEAI;

/** եեޥå dameon_record_reserach_format() ֤ */
typedef enum {
  D_FORMAT_ERROR = 0,
  D_FORMAT_UNKNOWN = 1,
  D_FORMAT_CSA = 2,
  D_FORMAT_KIF = 3,
} D_RECFORMAT;

typedef enum {
  /** ɤ߹ */
  D_LOAD_SUCCESS = 0,
  /** ɤ߹߼ */
  D_LOAD_ERROR = 1,
} D_LOADSTAT;

/** жɷ̤ */
typedef enum {
  /**  */
  D_UNKNOWN = 0,
  /** λ */
  D_TORYO = 1,
  /** Ԥä */
  D_MATTA = 2,
  /** ʬ */
  D_SENNICHITE = 3,
  /**  */
  D_JISHOGI = 4,
  /** ͤ */
  D_TSUMI = 5,
  /** ͤ */
  D_FUZUMI = 6,
  /** ؤʸʤ뤤ϥ롼ؤʤ */
  D_ERROR = 7,
} D_MATCHSTAT;

/** ɽ */
typedef struct {
  /** 礤 */
  D_TEAI teai;
  /** ֿ */
  DBoard first_board;
  /**  (00:00:00 UTC, January 1, 1970 ÿ)*/
  time_t start_time;
  /**  */
  DMoveInfo mi;
  /**  */
  int time[MAXRECORD];
  /** жɼԤ̾ */
  char player[2][MAXPLAYERNAME];
  /** ɤ߹߾ */
  D_LOADSTAT loadstat;
  /** Ծ */
  D_MATCHSTAT matchstat;
} Record;



/*********************************************************/
/* function prototypes */
/*********************************************************/

Record*     daemon_record_new            (void);
void        daemon_record_init           (Record *record);
void        daemon_record_free           (Record *record);
void        daemon_record_add_dte        (Record *record,
					  DTe *te,
					  time_t t);
int         daemon_record_output_csa     (Record* record,
					  char *filename);
int         daemon_record_output_csa_code(Record* record,
					  char *filename,
					  char *code);
int         daemon_record_output_kif     (Record* record,
					  char *filename);
int         daemon_record_output_kif_code(Record* record,
					  char *filename,
					  char *code);
Record*     daemon_record_load           (char *filename);
Record*     daemon_record_load_code      (char *filename, char *code);
Record*     daemon_record_load_csa       (char *filename);
Record*     daemon_record_load_csa_code  (char *filename, char *code);
Record*     daemon_record_load_kif       (char *filename);
Record*     daemon_record_load_kif_code  (char *filename, char *code);
int         daemon_record_read_lines     (char *filename,
					  char *outbuf[],
					  int maxline);
D_RECFORMAT daemon_record_reserach_format(char *filename);
D_LOADSTAT  daemon_record_get_loadstat   (Record *record);
D_MATCHSTAT daemon_record_get_matchstat  (Record *record);
char*       daemon_record_get_player     (Record *record,
					  D_TEBAN next);
void        daemon_record_set_player     (Record *record,
					  D_TEBAN next,
					  char *s);


#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* _RECORD_H_ */
