/* SI SHOGI Library
   Copyright (C) 2002 Masahiko Tokita
   This file is part of the SI SHOGI Library.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.w

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with The Daemonshogi; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */
/**
 * think.c
 * $Id: think.c,v 1.1.1.1 2005/12/09 09:03:10 tokita Exp $
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/si/think.c,v $
 */

#include <stdio.h>
#include <assert.h>
#include "si.h"
#include "ui.h"

void think_remove_akiohte(BOARD *bo, MOVEINFO *mi, int king_xy);
void think_remove_ng_oh_move(BOARD *bo, MOVEINFO *mi, int king_xy);
int think_is_tsumero(BOARD *bo, int search_depth);
void think_tsumero_uke(BOARD *bo, MOVEINFO *mi);

/**
 * õ
 * @param bo оݤBOARD
 * @param mi ̤ǼMOVEINFO
 * @param depth ɤߤο
 */
void think(BOARD *bo, MOVEINFO *mi, int depth) {
  int king_xy, king_xy2;
  int num_of_ohte;
  extern int THINK_DEEP_MAX;

#ifdef DEBUG
  assert( bo != NULL );
  assert( mi != NULL );
  assert( 0 <= depth );
#endif /* DEBUG */
  
  mi->count = 0;

  /* ʬβΰ */
  king_xy = bo->code_xy[1 + bo->next][0];
  /* βΰ */
  king_xy2 = bo->code_xy[1 + (bo->next == 0)][0];
  
  /* ꤵƤ뤫Ĵ٤ */
  num_of_ohte = boGetToBoardCount(king_xy, (bo->next == 0));

#ifdef DEBUG
  assert( 0 <= num_of_ohte && num_of_ohte <= 2 );
#endif /* DEBUG */
  
  if ( num_of_ohte ) {
    /* ꤵƤõ */
    uke( bo, mi, bo->next );
    return;
  }

  if ( depth == 0 ) {
    if ( think_is_tsumero(bo, 5) ) {
      /* ͤˤʤäƤ */
      /* ͤƨѥ롼Ƥ */
      think_tsumero_uke(bo, mi);
      if ( 0 < mi->count ) {
	return;
      }
      
      /* ꤬ʤϲõ */
      ohte(bo, mi);
      
      if ( 0 < mi->count ) {
	return;
      }
      
      /* ʤϹˡ򥻥åȤƽλ */
      search_053(bo, mi);
      search_068(bo, mi);
      /* ֤ä꤬ä */
      miRemoveDuplicationTE( mi );

      /* ưƶˤʤϺ */
      think_remove_akiohte(bo, mi, king_xy);

      /* ζθȤ˲ư꤬ä */
      think_remove_ng_oh_move(bo, mi, king_xy);
      
      return;
    }
  }
  
#undef NOREAD
#ifdef NOREAD
  for ( i=0; i<1; i++ ) {
    MATE_DEEP_MAX = i;
    if ( mate(bo, &root, 0) ) {
      /* ͤ */
      printTREE(root.child, 0);
      break;
    }
  }
    
  if ( root.child != NULL ) {
    /* check mate */
    miAdd( mi, &(root.child->te) );
    freeTREE( root.child );
    return;
  }
#endif /* NOREAD */

  search_011(bo, mi);
  search_013(bo, mi);
  search_014(bo, mi);
  search_012(bo, mi);
  if ( mi->count < 2 ) {
    search_024a(bo, mi);
  }
  if ( depth == 0 ) {
    search_068(bo, mi);
  }
  if ( depth <= THINK_DEEP_MAX - 3 ) {
    search_022(bo, mi);
    search_027(bo, mi);
    search_024a(bo, mi);
    search_031a(bo, mi);
    search_031b(bo, mi);
  }
  if ( depth <= 3 ) {
    search_046(bo, mi);
    search_048(bo, mi);
    search_044(bo, mi);
    search_041(bo, mi);
    search_047(bo, mi);
  }
  if ( depth <= 1 ) {
    search_053(bo, mi);
  }

  /* ֤ä꤬ä */
  miRemoveDuplicationTE( mi );

  /* ưƶˤʤϺ */
  think_remove_akiohte(bo, mi, king_xy);

  /* ζθȤ˲ư꤬ä */
  think_remove_ng_oh_move(bo, mi, king_xy);
  
  return;
}

/**
 * ưƶˤʤϺ
 * @param bo оݤBOARD
 * @param mi оݤMOVEINFO
 */
void think_remove_akiohte(BOARD *bo, MOVEINFO *mi, int king_xy) {
  int i;

  for ( i=0; i<mi->count; ) {
    if ( ukeAkiohte( bo, &(mi->te[ i ]), king_xy, bo->next ) ) {
      miRemoveTE2( mi, i );
    } else {
      i++;
    }
  }
}

/**
 * ζθȤ˲ư꤬ä
 * @param bo оݤBOARD
 * @param mi оݤMOVEINFO
 */
void think_remove_ng_oh_move(BOARD *bo, MOVEINFO *mi, int king_xy) {
  int i;

  for ( i=0; i<mi->count; ) {
    if ( mi->te[ i ].fm == king_xy &&
	0 < boGetToBoardCount(mi->te[ i ].to, (bo->next == 0)) ) {
      miRemoveTE2( mi, i );
    } else {
      i++;
    }
  }
}

/**
 * ̤ͤˤʤäƤʤĴ٤롣
 * ͤʤ 1 ֤ʤ 0 ֤
 * @param bo оݤBOARD
 * @param search_depth ͤߤĴ٤뿼
 * @reutrn ͤʤ 1 ֤ʤ 0 ֤
 */
int think_is_tsumero(BOARD *bo, int search_depth) {
  int flg_mate;
  int i;
  TREE root;
  extern int MATE_DEEP_MAX;

  initTREE( &root );
  
  flg_mate = 0;
  /* ̤ͤˤʤäƤʤĴ٤ */
  boToggleNext(bo);
  for ( i=0; i<search_depth; i++ ) {
    MATE_DEEP_MAX = i;
    if ( mate(bo, &root, 0) ) {
      /* ͤ */
      flg_mate = 1;
      break;
    }
  }
  boToggleNext(bo);
  if ( root.child != NULL ) {
    freeTREE( root.child );
  }
  
  return flg_mate;
}

/**
 * ͤ򸡺롣
 * @param bo оݤBOARD
 * @param mi ϿMOVEINFO
 */
void think_tsumero_uke(BOARD *bo, MOVEINFO *mi) {
  MOVEINFO mi1;
  int king_xy;
  TE te;
  int i;
  
  king_xy = bo->code_xy[1 + bo->next][0];

  mi1.count = 0;
  /* ľưζ */
  search_011(bo, &mi1);
  /* βؤβ */
  search_013(bo, &mi1);
  /* ư */
  search_049(bo, &mi1);
  /* ʬβդư */
  search_pressure(bo, &mi1, king_xy, bo->next);
  /* ľ˻ؤؤǤ */ 
  search_047(bo, &mi1);
  /* ʬβդ˶Ǥ */
  search_031b(bo, &mi1);
  
  /* ֤ä꤬ä */
  miRemoveDuplicationTE( &mi1 );

  /* ưƶˤʤϺ */
  think_remove_akiohte(bo, &mi1, king_xy);

  /* ζθȤ˲ư꤬ä */
  think_remove_ng_oh_move(bo, &mi1, king_xy);

  for (i=0; i<mi1.count; i++) {
    te = mi1.te[i];
    boMove_mate(bo, &te, bo->next);
    if ( think_is_tsumero(bo, 5) == 0 ) {
      /* ͤߤɤ˲ä */
      miAdd(mi, &te);
    }
    boBack_mate(bo, bo->next);
  }
}
