/*
 * Daemonshogi -- a GTK+ based, Simple shogi(japanese chess) program.
 * Copyright (C) 2002 Masahiko Tokita
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/**
 * Sprite class
 * $Source: /cvsroot/daemonshogi/daemonshogi/src/sprite.c,v $
 * $Id: sprite.c,v 1.1.1.1 2005/12/09 09:03:07 tokita Exp $
 */

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>
#include "sprite.h"
#include "misc.h"

gint src_xy[][2] = {
    { -1,  -1}, /* ʤ */
    {  0,   0}, /*    0x01 */
    { 40,   0}, /*  0x02 */
    { 80,   0}, /*  0x03 */
    {120,   0}, /*    0x04 */
    {160,   0}, /*    0x05 */
    {200,   0}, /*    0x06 */
    {240,   0}, /*  0x07 */
    {280,   0}, /* 겦   0x08 */
    {  0,  40}, /*    0x09 */
    { 40,  40}, /*  0x0A */
    { 80,  40}, /*  0x0B */
    {120,  40}, /*  0x0C */
    {280,  40}, /*    0x0D */
    {200,  40}, /*    0x0E */
    {240,  40}, /* ε   0x0F */
    { -1,  -1}, /* ε   0x10 */
    {  0,  80}, /*    0x11 */
    { 40,  80}, /*  0x12 */
    { 80,  80}, /*  0x13 */
    {120,  80}, /*    0x14 */
    {160,  80}, /*    0x15 */
    {200,  80}, /*    0x16 */
    {240,  80}, /*  0x17 */
    {280,  80}, /* 겦   0x18 */
    {  0, 120}, /*    0x19 */
    { 40, 120}, /*  0x1A */
    { 80, 120}, /*  0x1B */
    {120, 120}, /*  0x1C */
    {280, 120}, /*    0x1D */
    {200, 120}, /*    0x1E */
    {240, 120}, /* ε   0x1F */
};

/**
 * Sprite ƤΥݥ󥿤֤
 * @param pixmap ɽpixmap
 * @param width  pixmapβ
 * @param height pixmapνĥ
 */
Sprite* daemon_sprite_new(GdkPixmap* pixmap,
			  GdkBitmap* mask,
			  gint width,
			  gint height) {
  Sprite* sprite;

  sprite = (Sprite*)malloc(sizeof(Sprite));
  if (sprite == NULL) {
    daemon_abort("No enough memory in daemon_sprite_new()");
  }

  sprite->pixmap  = pixmap;
  sprite->mask    = mask;
  sprite->width   = width;
  sprite->height  = height;
  sprite->x       = 0;
  sprite->y       = 0;
  sprite->visible = FALSE;
  sprite->koma    = 0;

  return sprite;
}

/**
 * width֤
 * @param оݤsprite
 * @return width
 */
gint daemon_sprite_get_width(Sprite* sprite) {
  return sprite->width;
}

/**
 * height֤
 * @param оݤsprite
 * @return height
 */
gint daemon_sprite_get_height(Sprite* sprite) {
  return sprite->height;
}

/**
 * x֤
 * @param оݤsprite
 * @return x
 */
gint daemon_sprite_get_x(Sprite* sprite) {
  return sprite->x;
}

/**
 * y֤
 * @param оݤsprite
 * @return y
 */
gint daemon_sprite_get_y(Sprite* sprite) {
  return sprite->y;
}

/**
 * src_x֤
 * @param оݤsprite
 * @return x
 */
gint daemon_sprite_get_src_x(Sprite* sprite) {
  gint koma;

  g_assert(sprite != NULL);
  
  if (sprite->type == KOMA_SPRITE) {
    koma = daemon_sprite_get_koma(sprite);
    g_assert(1 <= koma && koma <= 0x1F);
    return src_xy[koma][0] * sprite->width / 40;
  } else {
    return 0;
  }
}

/**
 * src_y֤
 * @param оݤsprite
 * @return y
 */
gint daemon_sprite_get_src_y(Sprite* sprite) {
  gint koma;
  
  g_assert(sprite != NULL);
  
  if (sprite->type == KOMA_SPRITE) {
    koma = daemon_sprite_get_koma(sprite);
    g_assert(1 <= koma && koma <= 0x1F);
    return src_xy[koma][1] * sprite->height / 40;
  } else {
    return 0;
  }
}

/**
 * x򥻥åȤ롣
 * @param sprite оݤsprite
 * @param x åȤ x
 */
void daemon_sprite_set_x(Sprite* sprite, gint x) {
  sprite->x = x;
}

/**
 * y򥻥åȤ롣
 * @param sprite оݤsprite
 * @param y åȤ y
 */
void daemon_sprite_set_y(Sprite* sprite, gint y) {
  sprite->y = y;
}

/**
 * src_x򥻥åȤ롣
 * @param sprite оݤsprite
 * @param src_x åȤ src_x
 */
void daemon_sprite_set_src_x(Sprite* sprite, gint src_x) {
  g_assert(sprite != NULL);
  sprite->src_x = src_x;
}

/**
 * src_y򥻥åȤ롣
 * @param sprite оݤsprite
 * @param src_y åȤ src_y
 */
void daemon_sprite_set_src_y(Sprite* sprite, gint src_y) {
  g_assert(sprite != NULL);
  sprite->src_y = src_y;
}

/**
 * width 򥻥åȤ롣
 * @param sprite оݤsprite
 * @param width åȤ width
 */
void daemon_sprite_set_width(Sprite* sprite, gint width) {
  sprite->width = width;
}

/**
 * height 򥻥åȤ롣
 * @param sprite оݤsprite
 * @param height åȤ height
 */
void daemon_sprite_set_height(Sprite* sprite, gint height) {
  sprite->height = height;
}

/**
 * pixmap򥻥åȤ롣
 * @param sprite оݤsprite
 * @param pixmap åȤpixmap
 */
void daemon_sprite_set_pixmap(Sprite* sprite, GdkPixmap *pixmap) {
  sprite->pixmap = pixmap;
}

/**
 * pixmap֤
 * @param sprite оݤsprite
 * @return pixmap
 */
GdkPixmap* daemon_sprite_get_pixmap(Sprite* sprite) {
  return sprite->pixmap;
}

/**
 * ɽ뤫Υե饰֤TRUEʤɽ롣
 * @param оݤsprite
 * @return ɽ뤫Υե饰
 */
gboolean daemon_sprite_isvisible(Sprite* sprite) {
  return sprite->visible;
}

/**
 * ɽ뤫Υե饰򥻥åȤ롣
 * @param sprite оݤsprite
 * @param bool ɽ뤫Υե饰
 */
void daemon_sprite_set_visible(Sprite* sprite, gboolean bool) {
  sprite->visible = bool;
}

/**
 * pixmap֤
 * @param оݤsprite
 * @return pixmap
 */
GdkBitmap* daemon_sprite_get_mask(Sprite* sprite) {
  return sprite->mask;
}

/**
 * pixmap򥻥åȤ롣
 * @param оݤsprite
 * @param åȤBitmap
 */
void daemon_sprite_set_mask(Sprite* sprite, GdkBitmap* mask) {
  sprite->mask = mask;
}

/** 򥻥åȤ */
void daemon_sprite_set_koma(Sprite* sprite, gint koma) {
  g_assert(sprite != NULL);
  g_assert(1 <= koma && koma <= 0x1F);
  sprite->koma = koma;
}

/** ֤ */
gint daemon_sprite_get_koma(Sprite* sprite) {
  g_assert(sprite != NULL);
  return sprite->koma;
}

/** ץ饤Ȥμ֤ */
SPRITE_TYPE daemon_sprite_get_type(Sprite* sprite) {
  return sprite->type;
}

/** ץ饤Ȥμ򥻥åȤ */
void daemon_sprite_set_type(Sprite* sprite, SPRITE_TYPE type) {
  sprite->type = type;
}

/** ưǤ뤫Υե饰֤*/
gboolean daemon_sprite_ismove(Sprite* sprite) {
  return sprite->move;
}

/** ưǤ뤫Υե饰򥻥åȤ */
void daemon_sprite_set_move(Sprite* sprite, gboolean move) {
  sprite->move = move;
}

/** ɥåX֤򥻥åȤ */
void daemon_sprite_set_from_x(Sprite* sprite, gint x) {
  sprite->from_x = x;
}

/** ɥåY֤򥻥åȤ */
void daemon_sprite_set_from_y(Sprite* sprite, gint y) {
  sprite->from_y = y;
}

/** ɥåX֤֤ */
gint daemon_sprite_get_from_x(Sprite* sprite) {
  return sprite->from_x;
}

/** ɥåY֤֤ */
gint daemon_sprite_get_from_y(Sprite* sprite) {
  return sprite->from_y;
}
